
//
// For inheritence in our system,
// with this function, base class can access super class.
// Added by Bruce (Xiangdong) Li  2012/02/27
//
define([], function() {
SAS = window.SAS || {}; //define SAS namespace.

/**--- UA ----**/
SAS.Browser = {
    ua: navigator.userAgent.toLowerCase(),
    msie: function() {
        if (this._msie === undefined) {
            var pos = this.ua.match(/msie ([\d]+).[\d].+/);
            if (!!pos) {
                this._msie = parseInt(pos[1], 10);
            } else {
                // For IE 11
                pos = this.ua.match(/trident.*rv:([\d]+)/);
                if (pos) {
                    this._msie = parseInt(pos[1], 10);
                } else {
                    this._msie = false;
                }
            }
        }
        return this._msie;
    },
    chrome: function() {
        if (this._chrome === undefined) {
            var pos = this.ua.match(/chrome\/([\d]+).[\d.]+/);
            if (!!pos) {
                this._chrome = parseInt(pos[1], 10);
            } else {
                this._chrome = false;
            }
        }
        return this._chrome;
    },
    mchrome: function() {
        if (this._mchrome === undefined) {
            var pos = this.ua.match(/android.*chrome\/([\d]+).[\d.]+/);
            if (!!pos) {
                this._mchrome = parseInt(pos[1], 10);
            } else {
                this._mchrome = false;
            }
        }
        return this._mchrome;
    },
    safari: function() {
        if (this._safari === undefined) {
            var pos = this.ua.match(/version\/([\d]+).[\d.]+.[^mobile]*safari/);
            if (!!pos) {
                this._safari = parseInt(pos[1], 10);
            } else {
                this._safari = false;
            }
        }
        return this._safari;
    },
    msafari: function() {
        if (this._msafari === undefined) {
            var pos = this.ua.match(/version\/([\d]+).[\d.]+.mobile.*safari/);
            if (!!pos) {
                this._msafari = parseInt(pos[1], 10);
            } else {
                this._msafari = false;
            }
        }
        return this._msafari;
    },
    ff: function() {
        if (this._ff === undefined) {
            var pos = this.ua.match(/firefox\/([\d]+).[\d.]+/);
            if (!!pos) {
                this._ff = parseInt(pos[1], 10);
            } else {
                this._ff = false;
            }
        }
        return this._ff;
    },
    opera: function() {
        if (this._opera === undefined) {
            var pos = this.ua.match(/opera.([\d]+).[\d.]+/);
            if (!!pos) {
                this._opera = parseInt(pos[1], 10);
            } else {
                this._opera = false;
            }
        }
        return this._opera;
    },
    webkit: function() {
        if (this._webkit === undefined) {
            var pos = this.ua.match(/webkit\/([\d.]+)/);
            if (!!pos) {
                this._webkit = pos[1];
            } else {
                this._webkit = false;
            }
        }
        return this._webkit;
    },
    mozilla: function() {
        if (this._mozilla === undefined) {
            var pos = this.ua.match(/gecko\/([\d.]+)/);
            if (!!pos && !this.webkit()) {
                this._mozilla = pos[1];
            } else {
                this._mozilla = false;
            }
        }
        return this._mozilla;
    },
    mac: function() {
        if (this._mac === undefined) {
            this._mac = (/mac/).test(this.ua) && !this.ipad();
        }
        return this._mac;
    },
    ipad: function() {
        if (this._ipad === undefined) {
            this._ipad = (/ipad/).test(this.ua);
        }
        return this._ipad;
    }
}
/*
 * Modify prototypes of builtin objects.
 */
if (!Function.prototype.bind) {
    Function.prototype.bind = function (obj) {
        var self = this;
        return function () {
            return self.apply(obj, arguments);
        };
    };
}
Array.prototype.remove = function (obj) {
    var i = this.length;
    while (i) {
        if (this[i] === obj) {
            this.splice(i, 1);
        }
        i -= 1;
    }
};

/* For String */
String.prototype.Trim = function () {
    return this.replace(/(^\s+)|(\s+$)/g, "");
}
String.prototype.TrimStart = function () {
    return this.replace(/^\s+/g, "");
}
String.prototype.TrimEnd = function () {
    return this.replace(/\s+$/g, "");
}
String.prototype.EndsWith = function (str) {
    return this.substr(this.length - str.length) === str;
}
String.prototype.StartsWith = function (str) {
    return this.substr(0, str.length) === str;
}

/*
 * Utils Class Definition.
 */
SAS.Utils = function () {};
// static methods
SAS.Utils.findWordBreak = function (text, startIndex, increment) {
    var i, startState;
    if (increment < 0) {
        startIndex += increment;
    }
    startState = SAS.Utils.isWordSeparator(text[startIndex]);
    for (i = startIndex; (i >= 0) && (i < text.length); i += increment) {
        if (SAS.Utils.isWordSeparator(text[i]) != startState) {
            return (increment < 0) ? (i -= increment) : i;
        }
    }
    return (increment < 0) ? 0 : text.length;
};
SAS.Utils.isWordSeparator = function (character) {
    var separators = ' \t\'",;.!~@#$%^&*?=<>()[]:\\+-';
    return separators.indexOf(character) !== -1;
};
SAS.Utils.getCharCodeLength = function (s) {
    var l = 0,
        i;
    if (!arguments.length || !s) { return null; }
    if ("" == s) { return 0; }
    for (i = 0; i < s.length; i++) {
        if (s.charCodeAt(i) > 255) { l += 2; } else { l++; }
    }
    return l;
};
SAS.Utils.extend = function (s, b) {    var temp = function () { };
    temp.prototype = b.prototype;
    s.prototype = new temp();
    s.prototype.constructor = s;
    s.superClass = b.prototype;
    b.prototype.constructor = Object.prototype.constructor;
    if (b.prototype.constructor) {
        b.prototype.constructor = b;
    }
};
SAS.Utils.cloneObj = function (obj) {
    var pro = null,
        o = new Object();
    for (pro in obj) {
        o[pro] = obj[pro];
    }
    return o;
};
SAS.Utils.stopEvent = function (e) {
//    e.consumed = true;
    e.preventDefault();
    e.stopPropagation();
};

SAS.Utils.getIndex = function (list, key, start) {
    if (key.length > 0) {
        var i, l = list.length, r = new RegExp(key, 'i');
        for (i = isNaN(start) ? 0 : start; i < l; i++) {
            if (list[i].search(r) === 0) return i;
        }
    }
    return -1;
}

SAS.Utils.isValidString = function (string) {
    return !!string;
};
SAS.Utils.logcat = function (type) {
    var args = [], i;
    if (!SAS.Config.debug || !console) { return; }
    for (i = 1; i < arguments.length; i++) { args.push(arguments[i]); }
    console[type].apply(console, args);
};
   /**
    * Call once at beginning to ensure your app can safely call console.log() and
    * console.dir(), even on browsers that don't support it.  You may not get useful
    * logging on those browers, but at least you won't generate errors.
    * 
    * @param  alertFallback - if 'true', all logs become alerts, if necessary. 
    *   (not usually suitable for production)
    */
SAS.Utils.fixConsole = function (alertFallback)
{    
    if (typeof console === "undefined")
    {
        console = {}; // define it if it doesn't exist already
    }
    if (typeof console.log === "undefined") 
    {
        if (alertFallback) { console.log = function(msg) { alert(msg); }; } 
        else { console.log = function() {}; }
    }
    if (typeof console.dir === "undefined") 
    {
        if (alertFallback) 
        { 
            // THIS COULD BE IMPROVED? maybe list all the object properties?
            console.dir = function(obj) { alert("DIR: "+obj); }; 
        }
        else { console.dir = function() {}; }
    }
};
SAS.Utils.fixConsole();

SAS.Utils.getResourceBundle = function(resourceBundleName, packageName) {
    if (!packageName) packageName = "sas-commons/controls/_codeEditor";
    return dojo.i18n.getLocalization(packageName, resourceBundleName);
};

SAS.Utils.getBaseUrl = function () {
    if (this._baseUrl === undefined) {
        var i = 0, len = dojoConfig.packages.length, location = null;
        for (;i<len;i++) {
            location = dojoConfig.packages[i].location;
            if(location.indexOf('sas-commons') !== -1)
            {
                this._baseUrl = dojoConfig.baseUrl + (dojoConfig.baseUrl.EndsWith("/") ? "" : "/") + location + "/controls/_codeEditor";
                break;
            }
        }
        if (i === len) { this._baseUrl = dojoConfig.baseUrl; }
    }
    return this._baseUrl;
}

SAS.Utils.arrayToMap = function(arr) {
    var map = {}, len = arr.length;
    for ( var i = 0; i < len; i++) {
        map[arr[i]] = 1;
    }
    return map;
}
SAS.Utils.indexOfArray  = function(arr,key,cs) { //case sensitive
    var i = 0, len = arr.length;
    if (cs) {
        for(;i<len;i++) {
            if (arr[i]===key)return i;
        }
    } else {
        key = key.toUpperCase();
        for(;i<len;i++) {
            if (arr[i].toUpperCase() === key) return i;
        }
    }
    return -1;
}
/*
 * Log Class Definition.
 */
SAS.Log = function () {};
// static methods
SAS.Log.log = function () {
    if (!SAS.Config.debug || !console) { return; }
    console.log.apply(console, arguments);
};
SAS.Log.err = function () {
    if (!SAS.Config.debug || !console) { return; }
    console.error.apply(console, arguments);
};
/*
 * Event Class Definition.
 */
SAS.Event = function () {
    this.handlers = [];
};
SAS.Event.prototype.dispose = function () {
    this.handlers = [];
};
SAS.Event.prototype.add = function (handler) {
    this.handlers.push(handler);
};
SAS.Event.prototype.remove = function (handler) {
    var i;
    for (i = this.handlers.length - 1; i >= 0; i--) {
        if (this.handlers[i] === handler) {
            this.handlers.splice(i, 1);
        }
    }
};
SAS.Event.prototype.raise = function () {
    var i = 0;
    for (i = 0; i < this.handlers.length; i++) {
        this.handlers[i].apply(this, arguments);
    }
};
/*
 * Point Class Definition.
 */
SAS.Point = function () {
    if (arguments.length === 0) {
        this.x = 0;
        this.y = 0;
    }
    if (arguments.length === 1) { // copy constructor
        this.x = arguments[0].x;
        this.y = arguments[0].y;
    }
    if (arguments.length === 2) { // (x, y)
        this.x = arguments[0];
        this.y = arguments[1];
    }
};
SAS.Point.prototype = {
    equals: function (point) {
        return ((this.x === point.x) && (this.y === point.y));
    },
    add: function (point) {
        return new SAS.Point(this.x + point.x, this.y + point.y);
    },
    sub: function (point) {
        return new SAS.Point(this.x - point.x, this.y - point.y);
    }
};
/*
 * Size Class Definition.
 */
SAS.Size = function (width, height) {
    this.width = width;
    this.height = height;
};
/*
 * Rectangle Class Definition.
 */
SAS.Rectangle = function (x, y, width, height) {
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
};
SAS.Rectangle.prototype.union = function (rectangle) {
    var x1 = (this.x < rectangle.x) ? this.x : rectangle.x,
        y1 = (this.y < rectangle.y) ? this.y : rectangle.y,
        x2 = ((this.x + this.width) < (rectangle.x + rectangle.width)) ? (rectangle.x + rectangle.width) : (this.x + this.width),
        y2 = ((this.y + this.height) < (rectangle.y + rectangle.height)) ? (rectangle.y + rectangle.height) : (this.y + this.height);
    return new SAS.Rectangle(x1, y1, x2 - x1, y2 - y1);
};
SAS.Rectangle.prototype.isEmpty = function () {
    return this.width === 0 || this.height === 0;
};
SAS.Rectangle.prototype.toString = function () {
    return "(" + this.x + "," + this.y + ")-(" + this.width + "," + this.height + ")";
};
/*
 * TextPosition Class Definition.
 */
SAS.TextPosition = function () {
    if (arguments.length === 0) {
        this.line = 0;
        this.column = 0;
    }
    if (arguments.length === 1) { // copy constructor
        this.line = arguments[0].line;
        this.column = arguments[0].column;
    }
    if (arguments.length === 2) { // (line, column)
        this.line = arguments[0];
        this.column = arguments[1];
    }
};
SAS.TextPosition.prototype.equals = function (position) {
    return ((this.line === position.line) && (this.column === position.column));
};
SAS.TextPosition.prototype.compareTo = function (position) {
    var line = this.line - position.line;
    return (line === 0) ? (this.column - position.column) : line;
};
SAS.TextPosition.prototype.toString = function () {
    return "(" + this.line + "," + this.column + ")";
};
/*
 * TextRange Class Definition.
 */
SAS.TextRange = function () {
    if (arguments.length === 0) {
        this.start = new SAS.TextPosition();
        this.end = new SAS.TextPosition();
    }
    if (arguments.length === 1) { // copy constructor
        this.start = new SAS.TextPosition(arguments[0].start);
        this.end = new SAS.TextPosition(arguments[0].end);
    }
    if (arguments.length === 2) { // (start, end)
        this.start = arguments[0];
        this.end = arguments[1];
    }
};
SAS.TextRange.prototype.isEmpty = function () {
    return ((this.start.line === this.end.line) && (this.start.column === this.end.column));
};
SAS.TextRange.prototype.normalize = function () {
    return (this.start.compareTo(this.end) > 0) ? new SAS.TextRange(new SAS.TextPosition(this.end), new SAS.TextPosition(this.start)) : new SAS.TextRange(this);
};
SAS.TextRange.prototype.toString = function () {
    return this.start.toString() + "-" + this.end.toString();
};
SAS.TextRange.prototype.equals = function(textRange){
    return (this.start.equals(textRange.start) && this.end.equals(textRange.end));
}

SAS.Utils.KEY = {
    LEFT: 37, UP: 38, RIGHT: 39, DOWN: 40,
    PAGE_UP: 33, PAGE_DOWN: 34, END: 35, HOME: 36, INSERT: 45,
    TAB: 9, ENTER: 13, BACKSPACE: 8, DEL: 46, SPACE: 32,
    A: 65, C: 67, D: 68, F: 70, H: 72, I: 73, J: 74, K: 75, L: 76, M: 77,
    W: 87, Y :89, Z :90, MENU: 93, U:85,
    F1: 112, F2: 113, F3: 114, F4: 115, F10: 121,
    SEMICOLON: 186, PERIOD: 190, EQUAL: 187,
    SHIFT: 16, CTRL: 17, ESC: 27, IME :229, SLASH:191
};

});