define([], function() {
    var brackets = '{[()]}';

    var BracketHighlighter = function(ctrl) {
        this.ctrl = ctrl;
        ctrl.view().addEventListener('Selection', this.onSelect.bind(this));
    };

    BracketHighlighter.prototype = {
        onSelect: function() {
            var editor = this.ctrl.editor(),
                annotationModel = editor.getAnnotationModel(),
                AnnotationType = this.ctrl.core().getAnnotations().AnnotationType;
            annotationModel.removeAnnotations(AnnotationType.ANNOTATION_CURRENT_BRACKET);
            annotationModel.removeAnnotations(AnnotationType.ANNOTATION_MATCHING_BRACKET);

            var selection = editor.getSelection();
            if (selection.start !== selection.end) return;

            var allText = editor.getText(),
                text = allText[selection.start],
                index = brackets.indexOf(text), length = brackets.length,
                result;
            if (index >= 0 && index < length / 2) {
                result = this.match(text, brackets[length - index - 1], allText.slice(selection.start + 1));
                if (result >= 0) {
                    result += selection.start + 1;
                    annotationModel.addAnnotation(new (AnnotationType.getType(AnnotationType.ANNOTATION_CURRENT_BRACKET))(selection.start, selection.start + 1));
                }
            } else {
                text = allText[selection.start - 1];
                index = brackets.indexOf(text);
                if (index >= length / 2) {
                    result = this.match(text, brackets[length - index - 1], allText.slice(0, selection.start - 1), true);
                    if (result >= 0) {
                        annotationModel.addAnnotation(new (AnnotationType.getType(AnnotationType.ANNOTATION_CURRENT_BRACKET))(selection.start - 1, selection.start));
                    }
                }
            }
            if (result >= 0) {
                annotationModel.addAnnotation(new (AnnotationType.getType(AnnotationType.ANNOTATION_MATCHING_BRACKET))(result, result + 1));
            }
        },

        match: function(current, target, text, reverse) {
            var level = 1;
            for (var i = reverse ? text.length - 1 : 0;
                    (reverse && i >= 0) || (!reverse && i < text.length);
                    reverse ? i-- : i++) {
                if (text[i] === target) {
                    level--;
                } else if (text[i] === current) {
                    level++;
                }
                if (level === 0) {
                    return i;
                }
            }
            return -1;
        }
    };

    return BracketHighlighter;
});