/**
* CodeChecker APIs
*
*  This software is protected by copyright laws and international treaties.
*  U.S. GOVERNMENT RESTRICTED RIGHTS Use, duplication, or disclosure of this
*  software and related documentation by the U.S. government is subject to the
*  Agreement with SAS Institute and the restrictions set forth in FAR 52.227-19,
*  Commercial Computer Software - Restricted Rights (June 1987).
*  Copyright (c) 2012 by SAS Institute Inc., Cary, NC, USA   All rights reserved.
**/
define(["./LanguageService", "./EditorModel"], function(LanguageService, EditorModel) {
SAS = window.SAS || {};
/**
 *  This is the constructor of CodeChecker.
 */
SAS.CodeChecker = function (bConsoleLog) {
	this._bConsoleLog = !!bConsoleLog;
//   	this.model = new SAS.EditorModel(SAS.Editor.LanguageMode.SasCode);
	this._ptn_i18nOK = new RegExp("i18nok( *: *[a-z]+)?( *: *[a-z]+)?", "ig");
	this._ptn_i18nOK2 = new RegExp("ignore +for +i18n +analyzer *\\- *ID( *: *[a-z]+)?( *: *[a-z]+)?", "ig");
	this._ptn_i18nOK_item = new RegExp(" *: *([a-z]+)", "i");
	this._ptn_externalized_string = new RegExp("%sysfunc *\\( *sasmsg *\\(", "i");
	//this._ptn_macro_only_string = new RegExp("\" *&[a-z_]+\\.? *\"", "i");
	//this._ptn_macro_only_string = new RegExp("^\"((&+[a-z_][a-z_0-9]*)|[\\p{P}\\p{Z}\\p{S}])+\"$", "i");
	this._ptn_macro_only_string = /^"((&+[a-z_][a-z_0-9]*)|[`!@#$%^&*()-+`={}|\[\]\\:;'\"<>,.?/ ])+"$/i;
	this._ptn_punct_only_string  = /^['"][`!@#$%^&*()-+`={}|\[\]\\:;'\"<>,.?/ ]+['"]$/;
	this._ptn_fmtbody = new RegExp("^(\\$?)([a-z_]+([0-9]+[a-z_]+)*)(([0-9]*)\\.([0-9]*))$", "i"); // dollarW.D  wW.D  e8601daW.
    /*
     * The information source is http://zinc.jpn.sas.com/aprdpedia/index.php/KFunctions_Compatibilty .
     */
    this._ifn = {
		anyalnum: {i18nLevel: 1},
		anyalpha: {i18nLevel: 1},
		anycntrl: {i18nLevel: 1},
		anydigit: {i18nLevel: 1},
		anyfirst: {i18nLevel: 0},
		anygraph: {i18nLevel: 0},
		anylower: {i18nLevel: 1},
		anyname: {i18nLevel: 0},
		anyprint: {i18nLevel: 0},
		anypunct: {i18nLevel: 1},
		anyspace: {i18nLevel: 1},
		anyupper: {i18nLevel: 1},
		anyxdigit: {i18nLevel: 1},
		'byte': {i18nLevel: 0},
		cat: {i18nLevel: 2, substitutions: "KSTRCAT"},
		cats: {i18nLevel: 2},
		catt: {i18nLevel: 2},
		catx: {i18nLevel: 2},
		choosec: {i18nLevel: 2},
		choosen: {i18nLevel: 2},
		collate: {i18nLevel: 0},
		compare: {i18nLevel: 0, substitutions: "KCOMPARE"},
		compress: {i18nLevel: 0, substitutions: "KCOMPRESS"},
		compbl: {i18nLevel: 2},
		compged: {i18nLevel: 0},
		complev: {i18nLevel: 0},
		count: {i18nLevel: 1},
		countc: {i18nLevel: 1},
		dequote: {i18nLevel: 2},
		find: {i18nLevel: 1, substitutions: "KINDEX"},
		findc: {i18nLevel: 1, substitutions: "KINDEXC"},
		htmldecode: {i18nLevel: 2},
		htmlcode: {i18nLevel: 2},
		ifc: {i18nLevel: 2},
		ifn: {i18nLevel: 2},
		index: {i18nLevel: 0, substitutions: "KINDEX/KINDEXB"},
		indexc: {i18nLevel: 0, substitutions: "KINDEXC/KINDEXCB"},
		indexw: {i18nLevel: 0},
		left: {i18nLevel: 1, substitutions: "KLEFT"},
		length: {i18nLevel: 1, substitutions: "KLENGTH"},
		lengthc: {i18nLevel: 2},
		lengthm: {i18nLevel: 2},
		lengthn: {i18nLevel: 1, substitutions: "KLENGTH"},
		lowcase: {i18nLevel: 2, substitutions: "KLOWCASE"},
		nliteral: {i18nLevel: 2},
		notalnum: {i18nLevel: 1},
		notalpha: {i18nLevel: 1},
		notcntrl: {i18nLevel: 1},
		notdigit: {i18nLevel: 1},
		notfirst: {i18nLevel: 0},
		notgraph: {i18nLevel: 1},
		notlower: {i18nLevel: 1},
		notname: {i18nLevel: 0},
		notprint: {i18nLevel: 0},
		notpunct: {i18nLevel: 1},
		notspace: {i18nLevel: 1},
		notupper: {i18nLevel: 1},
		notxdigit: {i18nLevel: 1},
		nvalid: {i18nLevel: 2},
		propcase: {i18nLevel: 2},
		quote: {i18nLevel: 2},
		rank: {i18nLevel: 0},
		repeat: {i18nLevel: 2},
		reverse: {i18nLevel: 0, substitutions: "KREVERSE"},
		right: {i18nLevel: 2, substitutions: "KRIGHT"},
		scan: {i18nLevel: 0, substitutions: "KSCAN"},
		scanq: {i18nLevel: 1, substitutions: "KSCAN"},
		soundex: {i18nLevel: 0},
		spedis: {i18nLevel: 0},
		strip: {i18nLevel: 2, substitutions: "KSTRIP"},
		subpad: {i18nLevel: 1},
		substr: {i18nLevel: 0, substitutions: "KSUBSTR"},
		substrn: {i18nLevel: 0, substitutions: "KSUBSTRN"},
		translate: {i18nLevel: 0, substitutions: "KTRANSLATE"},
		trantab: {i18nLevel: 0},
		tranwrd: {i18nLevel: 2},
		trim: {i18nLevel: 2, substitutions: "KTRIM"},
		trimn: {i18nLevel: 2},
		upcase: {i18nLevel: 2, substitutions: "KUPCASE"},
		urldecode: {i18nLevel: 1},
		urlencode: {i18nLevel: 1},
		verify: {i18nLevel: 0, substitutions: "KVERIFY"}
    };
	this._lfm = {
		date: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		dateampm: {i18nLevel:1, substitutions: "NLDATMAP"},
		datetime: {i18nLevel:1, substitutions: "NLDATMM/NLDATMS/NLDATML/NLDATM"},
		ddmmyy: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		downame: {i18nLevel:1, substitutions: "NLDATEWN"},
		dtdate: {i18nLevel:1, substitutions: "NLDATMDT"},
		dtmonyy: {i18nLevel:1, substitutions: "NLDATMYM"},
		dtwkdatx: {i18nLevel:1},
		dtyear: {i18nLevel:1, substitutions: "NLDATMYR"},
		dtyyqc: {i18nLevel:1, substitutions: "NLDATMYQM/NLDATMYQS/NLDATMYQL/NLDATMYQ"},
		eurdfdd: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		eurdfde: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		eurdfdn: {i18nLevel:1},
		eurdfdt: {i18nLevel:1, substitutions: "NLDATMM/NLDATMS/NLDATML/NLDATM"},
		eurdfdwn: {i18nLevel:1, substitutions: "NLDATEWN"},
		eurdfmn: {i18nLevel:1, substitutions: "NLDATEMN"},
		eurdfmy: {i18nLevel:1, substitutions: "NLDATEYMM/NLDATEYMS/NLDATEYML/NLDATEYM"},
		eurdfwdx: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		eurdfwkx: {i18nLevel:1, substitutions: "NLDATEW"},
		hdate: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		mmddyy: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		mmyy: {i18nLevel:1, substitutions: "NLDATEYMM/NLDATEYMS/NLDATEYML/NLDATEYM"},
		monname: {i18nLevel:1, substitutions: "NLDATEMN"},
		monyy: {i18nLevel:1, substitutions: "NLDATEYMM/NLDATEYMS/NLDATEYML/NLDATEYM"},
		qtr: {i18nLevel:1},
		qtrr: {i18nLevel:1},
		time: {i18nLevel:1, substitutions: "NLTIME"},
		timeampm: {i18nLevel:1, substitutions: "NLTIMAP"},
		tod: {i18nLevel:1, substitutions: "NLDATMTM"},
		weekdate: {i18nLevel:1, substitutions: "NLDATEW"},
		weekdatex: {i18nLevel:1, substitutions: "NLDATEW"},
		worddate: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		worddatex: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		yymm: {i18nLevel:1, substitutions: "NLDATEYMM/NLDATEYMS/NLDATEYML/NLDATEYM"},
		yymmdd: {i18nLevel:1, substitutions: "NLDATEM/NLDATES/NLDATEL/NLDATE"},
		yymon: {i18nLevel:1, substitutions: "NLDATEYMM/NLDATEYMS/NLDATEYML/NLDATEYM"},
		yyq: {i18nLevel:1, substitutions: "NLDATEYQM/NLDATEYQS/NLDATEYQL/NLDATEYQ"},
		yyqr: {i18nLevel:1, substitutions: "NLDATEYQM/NLDATEYQS/NLDATEYQL/NLDATEYQ"},
		comma: {i18nLevel:1, substitutions: "NLNUM"},
		commax: {i18nLevel:1, substitutions: "NLNUM"},
		best: {i18nLevel:1, substitutions: "NLBEST"},
		d: {i18nLevel:1, substitutions: "NLDVALUE"},
		e: {i18nLevel:1},
		negparen: {i18nLevel:1},
		percent: {i18nLevel:1, substitutions: "NLPCT/NLPCTI"},
		pvalue: {i18nLevel:1, substitutions: "NLPVALUE"},
		ssn: {i18nLevel:1},
		w: {i18nLevel:1},
		wordf: {i18nLevel:1},
		words: {i18nLevel:1},
		dollar: {i18nLevel:1, substitutions: "NLMNLUSD/NLMNIUSD"},
		dollarx: {i18nLevel:1, substitutions: "NLMNLUSD/NLMNIUSD"},
		euro: {i18nLevel:1, substitutions: "NLMNLEUR/NLMNIEUR"},
		eurox: {i18nLevel:1, substitutions: "NLMNLEUR/NLMNIEUR"},
		yen: {i18nLevel:1, substitutions: "NLMNLJPY/NLMNIJPY"}
	};
	this._ima = {
		"%index": {i18nLevel: 0},
		"%left": {i18nLevel: 1},
		"%length": {i18nLevel: 1},
		"%qscan": {i18nLevel: 0},
		"%qsubstr": {i18nLevel: 0},
		"%scan": {i18nLevel: 0},
		"%substr": {i18nLevel: 0}
	};
	this._con = {
		cat: {i18nLevel: 0},
		cats: {i18nLevel: 0},
		catt: {i18nLevel: 0}
	};
};

SAS.CodeChecker.ParseState = {STARTING: 1, INSTATEMENT: 2};
SAS.CodeChecker.I18nIssueID = {ALL: "ALL", EMS: "EMS", LFM: "LFM", IFN: "IFN", IMA: "IMA", CON: "CON", UPS: "UPS"};
SAS.CodeChecker.I18nIssueLevel = {OK: "OK", INFO: "INFO", WARN: "WARN", ERROR: "ERROR"};

SAS.CodeChecker.prototype = {
    dispose: function () {
    },
    toString: function() {return "CodeChecker";},
    /*
     * Parameters:
     * 		text: 	  SAS program
     * 		callback: callback function as below
     *          void callback(array of i18nIssues)
     */
    doCheck: function(text, callback) {
    	var that = this;
	   	var model = new SAS.EditorModel(SAS.Editor.LanguageMode.SasCode);
	   	model.getLanguageService().tokens = [];
		model.getLanguageService().parsingStateEvent.add(onParsingState.bind(this));
		function onParsingState (state) {
        	if (state.state === SAS.LanguageService.ParsingState.STARTING) {
	        } else if (state.state === SAS.LanguageService.ParsingState.ENDED) {
				var stmt = "";
		    	var tokens = model.getLanguageService().tokens;
		    	var pstat = SAS.CodeChecker.ParseState.STARTING;
		    	var i18nIssues = new Array();
		    	var suppressions = new Array();
				for (var i = 0; i < tokens.length; i++) {
					var token =  tokens[i];
					var text = token.text;
					if (that._bConsoleLog)
					{
						console.log(text+":"+token.type+"["+token.start.line+","+				// LOG
							token.start.column+"]-["+token.end.line+","+token.end.column+"]"); 	// LOG
					}
					switch (token.type) {
						case "comment":
							// Is suppressed?
							commentAnalysis(token, suppressions);
							continue;
						case "macro-keyword":
							if (that._ima[token.text.toLowerCase()]) {	// Is non-i18n macro function?
								i18nIssues.push({type: SAS.CodeChecker.I18nIssueID.IMA,
												 level: SAS.CodeChecker.I18nIssueLevel.INFO, text: token.text, 
												 start: {line: token.start.line, column: token.start.column},
												 end: {line: token.end.line, column: token.end.column}});  
							}
							break;
						case "string":
							if (/*stmt != "%put" &&*/ !isOKString(token.text)) {// Is externalized?
								i18nIssues.push({type: SAS.CodeChecker.I18nIssueID.EMS,
												 level: SAS.CodeChecker.I18nIssueLevel.ERROR, text: token.text, 
												 start: {line: token.start.line, column: token.start.column},
												 end: {line: token.end.line, column: token.end.column}});  
							}
							break;
						case "text":
						case "keyword":
							// SAS Function
							isFunction(tokens, i, i18nIssues);
							break;
						case "format":
							// SAS Formats, Informats
							handleFormat(tokens, i, i18nIssues);
							break;
						case "numeric":
							// The tokenizer recognizes "+10" as a numeric token even in the put or input statements.
							// It has to be recognized two tokens "+" and "10" instead.
							// So, the CodeChecker handles that.
							if (text.charAt(0) == "+" && (stmt == "input" || stmt == "put")) {	// Check UPS
								i18nIssues.push({type: SAS.CodeChecker.I18nIssueID.UPS,
												 level: SAS.CodeChecker.I18nIssueLevel.WARN, text: "+", 
												 start: {line: token.start.line, column: token.start.column},
												 end: {line: token.end.line, column: token.start.column+1}});  
							}
							break;
						case "sep":
							var ptoken = (i > 0) ? tokens[i-1]: undefined;
							var ntoken = (i+1 < tokens.length - 1) ? tokens[i+1]: undefined;
							switch (text) {
								case ";": 	// statement end?
									pstat = SAS.CodeChecker.ParseState.STARTING;
									continue;
								continue;
								case "%":	// macro function call?
								case "&":	// macro variable reference?
									if (ntoken && ntoken.type == "text" && isConsequent(token, ntoken)) {
										i++;	// skip next token
									}
									break;
								case "@":
								case "+":	// column position specifier?
									if (stmt == "input" || stmt == "put") {	// Check UPS
										if (ntoken && (ntoken.text == "@" || ntoken.text == ";")) {
											// trailing @ and @@ are excluded
											if (ntoken.text == "@") {i++};
										} else {
											i18nIssues.push({type: SAS.CodeChecker.I18nIssueID.UPS,
															 level: SAS.CodeChecker.I18nIssueLevel.WARN, text: token.text, 
															 start: {line: token.start.line, column: token.start.column},
															 end: {line: token.end.line, column: token.end.column}});  
										}
									}
									break;
								case "|":
									if (ntoken && ntoken.text == "|" && isConsequent(token, ntoken)) {
										// String concatination
										i18nIssues.push({type: SAS.CodeChecker.I18nIssueID.CON,
														 level: SAS.CodeChecker.I18nIssueLevel.WARN, text: "||", 
														 start: {line: token.start.line, column: token.start.column},
														 end: {line: token.end.line, column: token.end.column+1}});  
									}
									break;
							}
							break;
					}
					if (pstat == SAS.CodeChecker.ParseState.STARTING) {
						stmt = text.toLowerCase();
						pstat = SAS.CodeChecker.ParseState.INSTATEMENT;
						if (that._bConsoleLog)
							console.log("STATEMENT:"+text);						// LOG
					}
				}	// end of the for loop
				var ret = new Array();
				for (var i = 0; i < i18nIssues.length; i++) {
					var issue = i18nIssues[i];
					//console.log("%s (%d,%d)-(%d,%d):%s:%s", issue.type, issue.start.line, issue.start.column,
					//			issue.end.line, issue.end.column, issue.text, ((issue.substitution)?issue.substitution: ""));
					var i18nOK = false;
					for (var j = 0; j < suppressions.length; j++) {
						if (suppressions[j].isI18nOK(issue.type, issue.start.line)) {
							i18nOK = true;
							break;
						}
					}
					if (i18nOK) {
						continue;
					}
					ret.push(issue);
				}
				callback(ret);
        	}
    	};
    	// Name: isFunction(array of token, index in the array, array of i18nIssues)
    	// Return value: true or false
    	// Description: Handle tokens as below
    	// &macro        : SEP, TEXT
    	// %userMacro    : SEP, TEXT
    	// name.name     : TEXT, SEP, TEXT
    	// func(         : TEXT, SEP / KEYWORD, SEP
    	function isFunction(tokens, index, i18nIssues) {
			var token =  tokens[index];
			var text = token.text;
			var ptoken = (index > 0) ? tokens[index-1]: undefined;
			var ntoken = (index+1 < tokens.length - 1 ? tokens[index+1]: undefined);
			if (text.length > 1 && text[0] == "%") {	// a user macro call?
				return false;
			}
			if (ptoken && ptoken.text == "&" && isConsequent(ptoken, token)) {	// macro variable?
				return false;
			}
			if (!ntoken || !isConsequent(token, ntoken)) {
				return false;
			}
			var issue;
			if (ntoken.text == "(") {	// SAS Function call?
				var ifn = that._ifn[token.text.toLowerCase()];
				if (ifn) {
					issue = {type: SAS.CodeChecker.I18nIssueID.IFN,
							 level: SAS.CodeChecker.I18nIssueLevel.INFO, text: token.text, 
							 start: {line: token.start.line, column: token.start.column},
							 end: {line: token.end.line, column: token.end.column}};
					if (ifn.substitutions) {
						issue.substitution = ifn.substitutions;	// Add the propaty "substitution" to the "issue"
					}
					i18nIssues.push(issue);
				}
				var con = that._con[token.text.toLowerCase()];
				if (con) {
					issue = {type: SAS.CodeChecker.I18nIssueID.CON,
							 level: SAS.CodeChecker.I18nIssueLevel.INFO, text: token.text, 
							 start: {line: token.start.line, column: token.start.column},
							 end: {line: token.end.line, column: token.end.column}};
					i18nIssues.push(issue);
				} 					
				if (that._bConsoleLog)
					console.log("FUNCTION:"+token.text);							// LOG
				return true;
			}
		}
    	function handleFormat(tokens, index, i18nIssues) {
			var token =  tokens[index];
			var groups = token.text.match(that._ptn_fmtbody);
			if (!groups || !groups[4]) {		// is format?
				return;
			}
			var fmtbody = token.text.replace(that._ptn_fmtbody, "$2");
			var prefix = token.text.replace(that._ptn_fmtbody, "$1");
			var suffix = token.text.replace(that._ptn_fmtbody, "$4");
			/*
			 * LFM
			 */
			var lfm = that._lfm[fmtbody.toLowerCase()];
			if (!lfm) {
				return false;
			}
			var issue = {type: SAS.CodeChecker.I18nIssueID.LFM, 
						 level: SAS.CodeChecker.I18nIssueLevel.INFO, text: token.text,
					 	 start: {line: token.start.line, column: token.start.column},
					 	 end: {line: token.end.line, column: token.end.column}};
			if (lfm.substitutions) {
				var substitution = "";
				var subs = lfm.substitutions.split("/");
				for (var i = 0; i < subs.length; i++) {
					if (i > 0) {
						substitution = substitution + "/";
					}
					substitution = substitution + prefix + subs[i];
					if (subs[i].match(/^NLDAT/) || subs[i].match(/^NLTIM/))
						substitution += "."; 
					else
						substitution += suffix; 
				}
				issue.substitution = substitution;		// Add the propaty "substitution" to lfm
			}
			i18nIssues.push(issue);
			if (that._bConsoleLog)
				console.log("FORMAT("+issue.text+"):body="+fmtbody+":suffix="+suffix+	// LOG
							":col="+issue.start.column+"-"+issue.end.column);			// LOG
			return true;
    	}
    	// Return value
    	// 	true: there is no space character between the given tokens
    	//	false: there is one or more space character
    	function isConsequent(token, ntoken) {
    		return (token.start.line == ntoken.start.line && token.end.column == ntoken.start.column);
    	}
    	function isOKString(literal) {
			var groups;
    		if (literal.length <= 2)
				return true;
  			groups = literal.match(that._ptn_punct_only_string);
  			if (groups) {
  				return true;
  			}
			if (literal.charAt(0) == "'") {
    			return false;
    		}
  			groups = literal.match(that._ptn_externalized_string);
  			if (groups) {
  				return true;
  			}
  			groups = literal.match(that._ptn_macro_only_string);  			
    		return (groups != null);
    	}
    	function commentAnalysis(token, suppresses) {
    		var groups, id, block, comment = token.text;
    		while ((groups = that._ptn_i18nOK.exec(comment)) != null ||
    			   (groups = that._ptn_i18nOK2.exec(comment)) != null) {
                block = 0; // 0: Line, 1: Begin, 2: End
    			id = SAS.CodeChecker.I18nIssueID.ALL;
    			if (groups.length > 1 && groups[1]) {
    				var item = groups[1].toLowerCase().replace(that._ptn_i18nOK_item, "$1");
    				switch (item) {
    					case "ems": id = SAS.CodeChecker.I18nIssueID.EMS; break;
    					case "lfm": id = SAS.CodeChecker.I18nIssueID.LFM; break;
    					case "ima": id = SAS.CodeChecker.I18nIssueID.IMA; break;
    					case "ifn": id = SAS.CodeChecker.I18nIssueID.IFN; break;
    					case "con": id = SAS.CodeChecker.I18nIssueID.CON; break;
    					case "ups": id = SAS.CodeChecker.I18nIssueID.UPS; break;
    				}    				
    			}
    			if (groups.length > 2 && groups[2]) {
    				var item = groups[2].toLowerCase().replace(that._ptn_i18nOK_item, "$1");
    				switch (item) {
    					case "begin": block = 1; break;
    					case "end": block = 2; break;
    				} 
    			}
    			switch (block) {
    				case 0:		// line
    					suppresses.push(new SAS.CodeChecker.Suppression(id, token.start.line, token.end.line));
    					break;
    				case 1:		// begin
    					suppresses.push(new SAS.CodeChecker.Suppression(id, token.start.line));
    					break;
    				case 2:		// end
    					for (var ndx = 0; ndx < suppresses.length; ndx++) {
    						var supp = suppresses[ndx];
    						if (supp.get_id() == id && supp.get_bLine() <= token.start.line && !supp.isEndLineSet()) {
    							supp.set_eLine(token.end.line);
    						}
    					}
    					break;
    			}
    		}
    	}
//        lines = text.split('\n');
//        lastLine = lines.length - 1;
        oldRange = new SAS.TextRange(new SAS.TextPosition(0, 0), new SAS.TextPosition(0, 0));            
        model.setText(oldRange, text);
    }
};

SAS.CodeChecker.Suppression = function (id, bLine) {
	var _id = id, _bLine = bLine, _eLine = -1;
	if (arguments.length > 2) {
		_eLine = arguments[2];
	}
	function isEndLineSet() {
		return (_eLine == -1) ? false: true;
	}
	return {
		get_id: function() {return _id;},
		get_bLine: function() {return _bLine;},
		get_eLine: function() {return _eLine;},
		set_eLine: function(eLine) {_eLine = eLine;},
		isEndLineSet: isEndLineSet,
		isI18nOK: function(id, line) {
            if (_bLine <= line && 
                (line <= _eLine || !isEndLineSet()) &&
                (_id == id || _id == SAS.CodeChecker.I18nIssueID.ALL))
                return true;
            return false;
		}
	};
};
return SAS.CodeChecker;
});