/**
 * Editor's ContextMenu
 * Added by Wei Wu
 */
// using SAS namespace
define(["dijit/Menu", "dijit/MenuItem", "dijit/MenuSeparator", "dojo/_base/lang", "dojo/on", "dojo/dom", "dojo/dom-geometry", "dojo/dom-style", "dijit/popup", "dojo/sniff"],
    function(Menu, MenuItem, MenuSeparator, lang, on, dom, domGeometry, domStyle, pm, has) {
        SAS = window.SAS || {};
        SAS.ContextMenu = function(controller) {
            var menu;
            if (SAS.Browser.ff()) {
                menu = new Menu({
                    style: 'margin: -1px'
                });
            } else {
                menu = new Menu();
            }
            var adjustThePosition = false,
            resourceBundle = SAS.Utils.getResourceBundle("ContextMenu_nls"),
            menuItems = [
                [{
                    label: resourceBundle.undo,
                    onClick: lang.hitch(controller, "undo"),
                    can: lang.hitch(controller, "canUndo")
                },
                {
                    label: resourceBundle.redo,
                    onClick: lang.hitch(controller, "redo"),
                    can: lang.hitch(controller, "canRedo")
                }],
                [{
                    label: resourceBundle.cut,
                    onClick: lang.hitch(controller, "cut"),
                    can: lang.hitch(controller, "canCut")
                },
                {
                    label: resourceBundle.copy,
                    onClick: lang.hitch(controller, "copy"),
                    can: lang.hitch(controller, "canCopy")
                },
                {
                    label: resourceBundle.paste,
                    onClick: lang.hitch(controller, "paste"),
                    can: lang.hitch(controller, "canPaste")
                }],
                [{
                    label: resourceBundle.go_to_line,
                    onClick: lang.hitch(controller, "showGoToLineDialog")
                }],
                [{
                    label: resourceBundle.backward,
                    onClick: lang.hitch(controller, "prevFunction")
                },
                {
                    label: resourceBundle.forward,
                    onClick: lang.hitch(controller, "nextFunction")
                }],
                [{
                    label: resourceBundle.line_number,
                    checked: lang.hitch(controller, "enableLineNbr"),
                    onClick: lang.hitch(controller, "toggleLineNumber")
                }],
                [{
                    label: resourceBundle.syntax_help,
                    onClick: function() { // S1294185, S1314768
                        window.setTimeout(controller.loadSyntaxHelp, SAS.Browser.msie() ? 100 : 0);
                    },
                    can: lang.hitch(controller, "canSyntaxHelp")
                }]
            ];
            
            menuItems.forEach(function(group, index) {
                group.forEach(function(item) {
                    menu.addChild(new MenuItem(item));
                });
                if (index != menuItems.length - 1) {
                    menu.addChild(new MenuSeparator());
                }
            });
            
            on(menu.domNode, "contextmenu", function(e) {
                SAS.Utils.stopEvent(e);
            });
            
            on(menu, "Open", function() {
                menu.getChildren().forEach(function(item) {
                    if (typeof(item.can) === 'function') {
                        item.set('disabled', !item.can());
                    }
                    if (typeof(item.checked) === 'function') {
                        item.set('iconClass', item.checked() ? "sce-icon sce-icon-checked" : "sce-icon");
                    }
                });
            });
            
            on(menu, "Close", function() {
                controller.focus();
            });
            
            /*if (SAS.Browser.msie()) {
                on(textarea, 'keydown', function(e) {
                    if (e.keyCode === SAS.Utils.KEY.MENU) { // context menu key
                        adjustThePosition = true;
                    }
                });
            }*/
            
            menu._openMyself = function(args) {
                
                // summary:
                //      Internal function for opening myself when the user does a right-click or something similar.
                // args:
                //      This is an Object containing:
                //
                //      - target: The node that is being clicked
                //      - iframe: If an `<iframe>` is being clicked, iframe points to that iframe
                //      - coords: Put menu at specified x/y position in viewport, or if iframe is
                //        specified, then relative to iframe.
                //
                //      _openMyself() formerly took the event object, and since various code references
                //      evt.target (after connecting to _openMyself()), using an Object for parameters
                //      (so that old code still works).
        
                var target = args.target,
                    iframe = args.iframe,
                    coords = args.coords;
        
                // To be used by MenuItem event handlers to tell which node the menu was opened on
                menu.currentTarget = target;
        
                // Get coordinates to open menu, either at specified (mouse) position or (if triggered via keyboard)
                // then near the node the menu is assigned to.
                if(coords && !adjustThePosition){
                    if(iframe){
                        // Specified coordinates are on <body> node of an <iframe>, convert to match main document
                        var ifc = domGeometry.position(iframe, true),
                            window = menu._iframeContentWindow(iframe),
                            scroll = domGeometry.docScroll(window.document);
        
                        var cs = domStyle.getComputedStyle(iframe),
                            tp = domStyle.toPixelValue,
                            left = (has("ie") && has("quirks") ? 0 : tp(iframe, cs.paddingLeft)) + (has("ie") && has("quirks") ? tp(iframe, cs.borderLeftWidth) : 0),
                            top = (has("ie") && has("quirks") ? 0 : tp(iframe, cs.paddingTop)) + (has("ie") && has("quirks") ? tp(iframe, cs.borderTopWidth) : 0);
        
                        coords.x += ifc.x + left - scroll.x;
                        coords.y += ifc.y + top - scroll.y;
                    }
                }else{
                    /*var view = controller.view(),
                        screenOffset = view.getViewAttr('screenOffset'),
                        scrollLeft = (SAS.Browser.msie() || SAS.Browser.ff()) ? document.documentElement.scrollLeft : document.body.scrollLeft,
                        scrollTop = (SAS.Browser.msie() || SAS.Browser.ff()) ? document.documentElement.scrollTop : document.body.scrollTop,
                        c = view.getCursorPosition();
                    coords = {};
                    coords.x = Math.round(view.getX(c.line, c.column) + view.getViewAttr('leftOffset')) + screenOffset.left + scrollLeft;
                    coords.y = Math.round(view.getViewAttr('topOffset') + (c.line - view.getViewAttr('startLine') + 1) * view.getViewAttr('lineHei')) + screenOffset.top + scrollTop;
                    */
                    coords = controller.getCursorCoord();
                }
                adjustThePosition = false;
        
                var self=menu;
                var prevFocusNode = menu._focusManager.get("prevNode");
                var curFocusNode = menu._focusManager.get("curNode");
                var savedFocusNode = !curFocusNode || (dom.isDescendant(curFocusNode, menu.domNode)) ? prevFocusNode : curFocusNode;
        
                function closeAndRestoreFocus(){
                    // user has clicked on a menu or popup
                    if(self.refocus && savedFocusNode){
                        savedFocusNode.focus();
                    }
                    pm.close(self);
                }
                pm.open({
                    popup: menu,
                    x: coords.x,
                    y: coords.y,
                    onExecute: closeAndRestoreFocus,
                    onCancel: closeAndRestoreFocus,
                    orient: menu.isLeftToRight() ? 'L' : 'R'
                });
                menu.focus();
        
                menu._onBlur = function(){
                    menu.inherited('_onBlur', arguments);
                    // Usually the parent closes the child widget but if this is a context
                    // menu then there is no parent
                    pm.close(menu);
                    // don't try to restore focus; user has clicked another part of the screen
                    // and set focus there
                };
            };
            
            // public APIs
            this.insertItems = function(startIndex, items) {
                startIndex = parseInt(startIndex);
                if (startIndex < 0) return;
                for (var i in items) {
                    var item = items[i],
                        index = startIndex + parseInt(i);
                    if (item.type === 'entry') {
                        menu.addChild(new MenuItem(item), index);
                    } else if (item.type === 'separator') {
                        menu.addChild(new MenuSeparator(), index);
                    }
                };
            };
            
            this.removeItems = function(startIndex, howmany) {
                startIndex = parseInt(startIndex);
                if (howmany === undefined) {
                    howmany = 1;
                }
                var items = menu.getChildren();
                for (var i = 0; i < howmany; i++) {
                    var item = items[startIndex + i];
                    if (item) {
                        item.destroyRecursive();
                    }
                }
            };
            
            this.modifyItem = function(index, options) {
                var item = menu.getChildren()[index];
                if (item && options) {
                    for (var i in options) {
                        item.set(i, options[i]);
                    }
                }
            };

            this.setView = function(view) {
                menu.bindDomNode(view._rootDiv);
            };
        }; // end ContextMenu
        return SAS.ContextMenu;
}); // end define