define([
        "./EditorModel",
        "./CoreRange",
        "./TextRange",
        "./TextPosition",
        "./Clipboard",
        "./ContextMenu",
        "./AutoCompleter",
        "./BracketHighlighter",
        "dojo/dom-style", 
        "dojo/dom-class", 
        "dojo/on",
        "dojo/has"
        ],
function(EditorModel, CoreRange, TextRange, TextPosition, Clipboard, ContextMenu, AutoCompleter,
    BracketHighlighter, domStyle, domClass, on, has) {

SAS = window.SAS || {};
SAS.ReplacingContext = {key:"", isFound:0};
SAS.ClipboardPrompted = {
    cut: false,
    copy: false,
    paste: false
};
/**
* Editor Controller
* 
**/
SAS.EditorController = function (core) {
    //private attributes
    var mode = core.mode(),
        model,
        view,
        editor,
        // searchCounter,
        // TEXTAREA: capture events
        //canvasId = id + '_canvas',
        //taId = id + '_ta',
        hint,
        findDlg,
        contextMenu,
        autoCompleter,
        sasAutoCompleter,
        wordInfo,
        CRLF_REGEXP = /\r\n|\r/g,
        pendingList = ['activate', 'setAutoCompleter', 'setLibService', 'setText', 'focus', 'cut', 'paste', 'selectAll', 'clear',
            'gotoline', 'replace', 'collapse', 'expand', 'showGoToLineDialog', 'showFindReplaceDialog', 'regCommands', 'selection',
            'bind', 'unbind'],
        funcBag = {}, callQueue = [],
        KEY = SAS.Utils.KEY,
        self = this;
    this.settings = core.settings();
    findDlg = SAS.FindAndReplaceDialog.getInstance();
    goToLineDlg = SAS.GoToLineDialog.getInstance();
    this.resourceBundle = SAS.Utils.getResourceBundle("EditorController_nls");
    //Privileged getter
    this.model = function () {
        return model;
    }
    this.mode = function() {
        return mode;
    };
    this.core = function() {
        return core;
    };
    this.view = function () {
        return view;
    };
    this.editor = function() {
        return editor;
    };

    this.getContextMenu = function() {
        return contextMenu;
    }
    this.setLibService = function(fn) {
        if (sasAutoCompleter) {
            sasAutoCompleter.setLibService(fn);
        }
    }

    this._cleanText = function(text) { //TODO:
        var result = text.replace(CRLF_REGEXP,'\n');
        if(SAS.Settings.getInstance().tabAsSpaces()) {
            var selection = this.modelTA().getSelectionRange();
            result = this.modelTA()._translateTab(result, selection.start);
        }
        return result;
    }

    this._getAnnotations = function(type) {
        var list = [];
        if (this.editor()) {
            var iter = this.editor().getAnnotationModel().getAnnotations();
            while (iter.hasNext()) {
                var anno = iter.next();
                if (anno.type === type) {
                    list.push(anno);
                }
            }
        }
        return list;
    };

    //Private methods
    //Event process
    function onFindReplace(e) {
        //SAS.Utils.stopEvent(e);
        self.showFindReplaceDialog();
        return true;
    }
    
    function onGoToLine(e) {
        // TODO: disable this code because webdms will not use this feature.
        SAS.Utils.stopEvent(e);
        self.showGoToLineDialog();
        return false;
    }

    function selectSearch() {
        window.setTimeout(function(){ // S1242235
            var searchConfig = {
                    direction: "forward",
                    wrapped: false,
                    all: true,
                    matchCase: true
                },
                selectContent = self.view().getSelectionText(),
                result = [],
                allText = self.model().getAllText();
            if (selectContent.search(/([\x21-\x7e]|[^\x00-\xff])+/) >= 0) {
                result = self.doSearchAll(selectContent, searchConfig, allText, true);
                var annotationModel = self.editor().getAnnotationModel(), range = null;
                for(var count=0;count<result.length;count++){
                    range = self.model().toCoreRange(result[count]);
                    annotationModel.addAnnotation(new (core.getAnnotations().AnnotationType.getType(core.getAnnotations().AnnotationType.ANNOTATION_MATCHING_SEARCH))(range.start, range.end));
                }
                //self.view().setFound(result);
            }
        }, 0);
    }
    
    function restoreSelection () {
        /*if (SAS.Browser.msie()) {
            self.modelTA().setSelectionRange(self.modelTA().getSelectionRange());
        }*/
    }


    function handleFind(event) {
        if (SAS.EditorController.activeCtrl === self) {
            if (event.type !== 'hide') {
                if (!self.isCursor() && self.getText('selected').toLowerCase() != event.findValue.toLowerCase()) {
                    if (event.config) event.config.inSelection = true;
                } else {
                    if (event.config) event.config.inSelection = false;
                }
            }
            switch (event.type) {
            case 'find':
                restoreSelection();
            case 'findall':
                self.search(event.findValue, event.config);
                break;
            case 'replace':
            case 'replaceall':
                restoreSelection();
                self.replace(event.findValue, event.replaceValue, event.config);
                break;
            case 'hide':
                restoreSelection();
                window.setTimeout(function(){self.focus();},400);
                break;
            }
        }
    }
    findDlg.attach(handleFind);
    
    function handleGoLine(event) {
        if(SAS.EditorController.activeCtrl === self){
            if(event.type == 'gotoline') {
                restoreSelection();
                self.gotoline(event.lineNumber);
            }
            if(event.type == 'hide') {
                restoreSelection();
                window.setTimeout(function(){self.focus();},400);
            }
        }
    }
    goToLineDlg.attach(handleGoLine);
    
    this.showFindReplaceDialog = function () {
        findDlg.show({
            findValue: self.getText('selected').split('\n')[0],
            readOnly: this.readOnly()
        });
    };
    
    this.hideFindReplaceDialog = function () {
        findDlg.hide();
    };
    
    this.showGoToLineDialog = function () {
        goToLineDlg.show(self.lineCount());
    };
    
    this.hideGoToLineDialog = function () {
        goToLineDlg.hide();
    };

    this.regCommands = function (config) {
        if (config && view) {
            $.each(config, function() {
                var keyCode = this.keyCode, ctrlKey = this.ctrlKey || this.metaKey, shiftKey = this.shiftKey, altKey = this.altKey,
                actionName = keyCode + '_' + ctrlKey+ '_' + shiftKey + '_' + altKey;
                view.setKeyBinding(new (core.getKeyBinding()).KeyBinding(keyCode, ctrlKey, shiftKey, altKey), actionName);
                view.setAction(actionName, this.handle, {name: actionName});
            });
        }
    };
    //Bin,Zhou: comment when replacing canvas with div+css
    //this.adjustTextAreaPos = function () {}; //Reserved, implemented in require.

    //this.remove = function () {}; //Reserved, implemented in require.
    this.canSyntaxHelp = function () {
        return sasAutoCompleter ? sasAutoCompleter.canSyntaxHelp() : false;
    };
    this.loadSyntaxHelp = function () {
        if (sasAutoCompleter) {
            sasAutoCompleter.tryLoadHint();
        }
    };
    this.getCursorCoord = function() {
        var x, y, caretLocation = view.getLocationAtOffset(view.getCaretOffset());
        view.convert(caretLocation, "document", "page");
        x = caretLocation.x;
        y = caretLocation.y + view.getLineHeight();
        if (SAS.Browser.msie() || SAS.Browser.ff()) {
            x += document.documentElement.scrollLeft;
            y += document.documentElement.scrollTop;
        } else {
            x += document.body.scrollLeft;
            y += document.body.scrollTop;
        }
        return {x:x,y:y};
    };
    function showContextMenu (e) {
        var element = view._clientDiv,//document.activeElement
            evt = element.ownerDocument.createEvent('MouseEvents'),
            pos = self.getCursorCoord();
            RIGHT_CLICK_BUTTON_CODE = 2; // the same for FF and IE
        evt.initMouseEvent('contextmenu', true, true,
             element.ownerDocument.defaultView, 1,
             0, 0, pos.x, pos.y, false,
             false, false, false, RIGHT_CLICK_BUTTON_CODE, null);
        if (document.createEventObject){
           // dispatch for IE
           return element.fireEvent('onclick', evt)
        }
        else{
           // dispatch for firefox + others
           return !element.dispatchEvent(evt);
        }
    }

    this._fontSize = function() {
        core.form().find('.textviewLeftRuler,.textviewContent').css('fontSize', this.settings.getFontSize() + 'px');
        core.updateFoldingBar();
        this.view().update(true);
        core.updateRowHighlight();

        //var themeClass = "myTheme";
        //var settings = {
        //    "className": "myTheme",
        //    "name": "myTheme",
        //    "styles": {
        //        "fontSize": "16px"
        //    }
        //};
        //var theme = this.view().getOptions("theme");
        //settings["styles"]["fontSize"] = ctrl.settings.getFontSize() + "px";
        //theme.setThemeClass(themeClass, theme.buildStyleSheet(themeClass, settings));
    };

    this._syntaxHighlighting = function() {
        this.view().update(true);
    };

    this.getLengthWithTabAsSpaces = function (text) {
        var tabLength = 0,
            bufferLength = text.length,
            i;
        for (i = 0; i < bufferLength; i++) {
            tabLength += (text[i] === '\t') ? this.settings.getTabSize() : 1;
        }
        return tabLength;
    };

    this.tabSize = function() {
        return this.settings.getTabSize();
    };

    this.getLanguageService = function() {
        return core.langService();
    };
    
    function onSettingsChanged (e) {
        for (var p in e) {
            switch (p) {
                case 'theme':
                    break;
                case 'lineWrapped':
                    break;
                /*case 'foldingIcons':
                 view.clpsIconController.setIcons(e[p]);
                 break;*/
                case 'lineNumberVisible':
                    view.focus();
                    editor.setLineNumberRulerVisible(self.settings.getLineNumberVisible());
                    if(self.settings.getLineNumberVisible()){
                        core.sidebarSelection();
                    }
                    break;
                case 'foldingEnabled':
                    editor.setFoldingRulerVisible(self.settings.getFoldingEnabled());
                    core.updateFoldingBar();
                    break;
                case 'readOnly':
                    self.readOnly(self.settings.getReadOnly());
                    break;
                case 'lineHighlightingVisible':
                    /*var lineHightlightEnable = self.settings.getLineHighlightingVisible();
                    editor.getStyleAccessor()._styler._highlightCaretLine = lineHightlightEnable;
                    var annotationModel = editor.getAnnotationModel(), annos = null;
                    if (core) {
                        annos = self._getAnnotations(core.getAnnotations().AnnotationType.ANNOTATION_CURRENT_LINE);
                    }
                    if (lineHightlightEnable) {
                        if (!annos) {
                            $.each(annos, function() {
                                annotationModel.addAnnotation(this);
                            });
                        }
                    } else {
                        if (annos) {
                            $.each(annos, function() {
                                annotationModel.removeAnnotation(this);
                            });
                        }
                    }
                    view.redrawLines();*/
                    core.updateRowHighlight();
                    break;
                case 'fontSize':
                    self._fontSize();
                    break;
                case 'tabSize':
                    view._setTabSize(self.settings.getTabSize());
                    break;
                case 'tabAsSpace':
                    view.setOptions({expandTab: self.settings.getTabAsSpace()});
                    break;
                case 'syntaxHighlighting':
                    view.focus();
                    self._syntaxHighlighting();
                    //view.enableHighlighting(e[p]);//Bin,Zhou: add when replacing canvas with div+css
                    break;
                case 'hintEnabled':
                    if (sasAutoCompleter) {
                        sasAutoCompleter.enableHint(self.settings.getHintEnabled());
                    }
                    break;
            }
        }
        //view.update(); //Bin,Zhou: comment when replacing canvas with div+css
    }

    function init () {
        model = core.model();
        //langService = mode.getHighlight(model);
        view = core.view();
        editor = core.editor();
        self.foldingMgr = core.foldingMgr();

        jQuery(self.settings).on("settingChanged", onSettingsChanged);
        self.regCommands([
            {ctrlKey:true, altKey:false, keyCode: KEY.F, handle:onFindReplace},    //Ctrl+F
            {ctrlKey:true, altKey:false, keyCode: KEY.SLASH, handle:self.comment.bind(self)},    //Ctrl+/
            {ctrlKey:true, altKey:false, shiftKey: true, keyCode: KEY.C, handle:self.comment.bind(self)},    //S1250057 Ctrl+Shift+C
            {ctrlKey:true, altKey:false, keyCode: KEY.U, handle:self.toUpperCase.bind(self)},    //Ctrl+U
            {ctrlKey:true, altKey:false, keyCode: KEY.L, handle:self.toLowerCase.bind(self)},     //Ctrl+L
            {ctrlKey:true, altKey:false, shiftKey: true, keyCode: KEY.U, handle:self.cycleCases.bind(self)} // Ctrl+Shift+U
        ]);
        if(SAS.Browser.mac()) {
            self.regCommands([
                //{ctrlKey:false, altKey:false, metaKey:true, keyCode: KEY.F, handle:onFindReplace}, //Cmd+F
                {ctrlKey:false, altKey:true, keyCode: KEY.F10, handle:showContextMenu} //Alt + F10
                //{ctrlKey:false, altKey:true, keyCode: KEY.UP, handle:scrollUp},
                //{ctrlKey:false, altKey:true, keyCode: KEY.DOWN, handle:scrollDown}
            ]); //defect : S0890541
        }
        /*
         * Fix defect: Mislocation when paste multi-lines to a native blank line of CE in a native-hidden tab.
         * Impl: Re-init text as soon as the tab is activated at the first time.
         * Fixed By: Boyce.
         */
        /*self.initFlag = sap.ui.Device.browser.chrome;
        self.initText = self._cleanText(content);
        if (self.model().isEmpty() && !!self.initText) {
            //modelTA.setText(self.initText);
            model.setText(model.getAllRange(), self.initText);
        }
        // self.adjustTextAreaPos();
        /*if (langMode === sas.hc.ui.commons.codeEditor.LanguageMode.Log ||
            langMode === sas.hc.ui.commons.codeEditor.LanguageMode.Lst) {
            self.readOnly(true);
        }*/

        //self.foldingMgr = new FoldingMgr(mode, model, self.settings);

        if (mode.getLanguage() === SAS.codeEditor.LanguageMode.SasCode) {
            sasAutoCompleter = new AutoCompleter(self, view, model);
            hint = sasAutoCompleter.getHint();
            jQuery(hint).on('rollBackFocus', self.focus.bind(self));
            if (self.settings.getHintEnabled()) {
                sasAutoCompleter.enableHint(true);
            }
        }

        new BracketHighlighter(self);

        self.initSettings();

        contextMenu.setView(view);

        //formatter = mode.getFormatter(self);

        jQuery(view._clientDiv).on('dblclick', selectSearch);
        pendingList.forEach(function(funcName) {
            self[funcName] = funcBag[funcName];
        });
        callQueue.forEach(function(func) {
            (self[func.name]).apply(self, func.args);
        });

        view.addEventListener('Focus', self.activate.bind(self));

        jQuery(core).triggerHandler({type: 'ready'});
    }
    this.initSettings = function() {
        var view = this.view(), editor = this.editor(), settings = this.settings;
        view.setOptions({expandTab: settings.getTabAsSpace(), tabSize: settings.getTabSize()});

        core.form().find('.textviewLeftRuler,.textviewContent').css('fontSize', settings.getFontSize() + 'px');

        editor.setLineNumberRulerVisible(settings.getLineNumberVisible());

        editor.setFoldingRulerVisible(settings.getFoldingEnabled());
        core.updateFoldingBar();

        core.updateRowHighlight();

        window.setTimeout(function(){
            self.readOnly(settings.getReadOnly());
        }, 0);
    };
    //TODO: 
    //modelTA.attr('readOnly', model.readOnly()).on(['scroll', 'dblclick'], [onScroll, selectSearch]);
    //attr('readOnly', model.readOnly());

    // sidebar selection
    /*var sidebar = view.getSidebar();
    if (has("MSPointer")){
        if (has("pointer-events")){ // IE10
            on(sidebar, "pointerdown", function(e) {
                transControl.dispatch('onsidebarselect', e);
            });		
        } else { // IE11
            on(sidebar, "MSPointerDown", function(e) {
                transControl.dispatch('onsidebarselect', e);
            });
        }
    } else {
        on(view.getSidebar(), 'mousedown', function(e){
            if (e.button === 0) {
                transControl.dispatch('onsidebarselect', e);
            }
        });
    }

    transDriver.regListeners(['dblclick'], [ selectSearch]);

    self.resize = function (wid, hei) {
        view.resize(wid, hei);
    };
    self.remove = function () {
        modelTA.dispose();
        findDlg.detach(handleFind);
    };
    if (SAS.Browser.mac()) {
        on(window, 'focus', function(e){
            if (SAS.EditorController.activeCtrl === self) {
                 //view.update();
                    return true;
            }
        });
    }*/
    contextMenu = new ContextMenu(self);
    
    jQuery(core).on('coreReady', init);

    //formatter = mode.getFormatter(this);

    pendingList.forEach(function(funcName) {
        funcBag[funcName] = self[funcName];
        (function(name) {
            self[name] = function() {
                callQueue.push({name: name, args: arguments});
            };
        })(funcName);
    });
};
SAS.EditorController.prototype = {
    /** Properties and public methods **/
    render: function(parentDomRef) {
        jQuery(parentDomRef).append(this.core().getDomRef());
    },
    isActive: function() {
        return SAS.EditorController.activeCtrl === this;
    },
    getHTML : function(option) {
        return this.core().getHTML(option);
    },
    _doAction: function(name) {
        var editor = this.editor();
        if (editor && editor.getUndoStack && editor.getUndoStack()) {
            return editor.getUndoStack()[name]();
        }
        return false;
    },
    isCursor: function() {
        var selection = this.view().getSelection();
        if (selection) {
            return (selection.start === selection.end);
        }
        return false;
    },
    resize: function(wid, hei) {
        this.core().resize(wid, hei);
    },
    enableLineNbr: function(enable) {
        if (undefined !== enable) {
            this.settings.setLineNumberVisible(enable);
            this.editor().setLineNumberRulerVisible(enable);
        }
        return this.settings.getLineNumberVisible();
    },
    bind: function (events, listener){
        this.core().evtMgr().bind(events, listener);
    },
    unbind: function (events, listener){
        this.core().evtMgr().unbind(events, listener);
    },
    canRedo: function() {
        return !this.readOnly() && this._doAction('canRedo');
    },
    canUndo: function() {
        return !this.readOnly() && this._doAction('canUndo');
    },
    canCut: function() {
        return !this.readOnly() && !this.isCursor();
    },
    canCopy: function() {
        return !this.isCursor();
    },
    canPaste: function() {
        return !this.readOnly();
    },
    undo: function () {
        this._doAction('undo');
        //var editor = this.editor;
        //if (editor && editor.getUndoStack && editor.getUndoStack()) {
        //    editor.getUndoStack().undo();
        //}
        //this.transControl().undo();
    },
    redo: function () {
        this._doAction('redo');
        //this.transControl().redo();
    },

    readOnly: function (readOnly) {
        if (undefined !== readOnly) {
            this.core().readOnly(readOnly);
            SAS.FindAndReplaceDialog.getInstance().readOnly(readOnly);
        }
        return this.settings.getReadOnly();
        /*if (typeof readOnly === 'boolean') {
            this.modelTA().readOnly(readOnly);
            FindAndReplaceDialog.getInstance().readOnly(readOnly);
            this.view().readOnly(readOnly);
        }
        return this.modelTA().readOnly();*/
    },
    //destory
    dispose: function () {
    /*    this.view().dispose();
        this.model().dispose();
        this.remove();
    */},

    //task
    activate: function() {
        if (!this.isActive()) {
            SAS.EditorController.activeCtrl = this;
            SAS.FindAndReplaceDialog.getInstance().readOnly(this.readOnly());
        }
        this.core().refresh();
    },
    deactivate: function() {
        SAS.EditorController.activeCtrl = null;
    },
    focus: function () {
        //this.modelTA().focus();
        this.view().focus();
    },

    setText: function (text) {
        text = this.core().cleanText(text);
        /*if(this.readOnly() && SAS.Browser.msie()) {
            this.readOnly(false);
            this.editor().setText(text);
            this.readOnly(true);
        } else {
            /*if (this.initFlag && !!this.initText) {
                this.initText = text;
            }*/
            this.editor().setText(text);
        //}
        //this.undoMgr().clear();
    },

    insertText: function (text) {
        if (!this.readOnly()) {
            var tmp = this.core().cleanText(text);
            var selection = this.editor().getSelection();
            this.editor().setText(tmp, selection.start, selection.end);
            //var tmp = this._cleanText(text),
            //    selection = this.modelTA().getSelectionRange(),
            //    oldText = this.modelTA().getSelectedText(),
            //    newRange = new CoreRange(selection.start, selection.start + tmp.length);
            //this.undoMgr().add(tmp, newRange, newRange, oldText, selection, selection);
            //this.modelTA().insertText(tmp, selection, 'api');
        }
    },
    replaceText: function (text, range) {
        if (!this.readOnly()) {
            var tmp = this.core().cleanText(text);
                /*oldRange = range ? range : this.modelTA().getSelectionRange(),
                oldCursor = new CoreRange(oldRange.end, oldRange.end),
                oldText = this.modelTA().getSelectedText(oldRange),
                newRange = new CoreRange(oldRange.start, oldRange.start + tmp.length);*/
            //this.undoMgr().add(tmp, newRange, newRange, oldText, oldRange, oldCursor);
            //this.modelTA().insertText(tmp, oldRange, 'api');
            this.editor().setText(tmp, range.start, range.end);
        }
    },
    selectAll: function () {
        this.editor().setSelection(0,  this.core().baseModel().getCharCount());
    },
    
    selection: function(start, end) {
        var editor = this.editor();
        if(!isNaN(start) && !isNaN(end)) {
            return editor.setSelection(start, end);
        } else if(!isNaN(start) && isNaN(end)) {
            return editor.setSelection(start, start);
        } else if(isNaN(start) && !isNaN(end)) {
            return editor.setSelection(end, end);
        } else if(isNaN(start) && isNaN(end)) {
            return editor.getSelection();
        }
    },

    caretPosition: function() {
        var result = this.model().toTextRange(this.editor().getSelection());
        return {line : result.start.line, column : result.start.column};
    },

    clear: function () {
        if(!this.readOnly()) {
            //this.expandAll();
            this.view().setText('');
            //var range = new SAS.TextRange(new SAS.TextPosition(0,0), new SAS.TextPosition(this.model().getLineCount(), 0));
            //this.modelTA().select(range);
            //this.deleteText();
            //var oldRange = new CoreRange(0, this.modelTA().getLength());
            //var oldText = this.modelTA().getAllText();
            //var newRange = new CoreRange(0, 0);
            //this.undoMgr().add('', newRange, newRange, oldText, oldRange, oldRange);
            //this.modelTA().insertText('', oldRange, 'clear');
            //this.modelTA().setCaretPosition(0);
        }
    },

    cut: function () {
        if (this.readOnly()) return;
        /*if (!SAS.ClipboardPrompted.cut && !Clipboard.canAccessSystemClipboard()) {
            SAS.PromptDialog.show(document.title, this.resourceBundle.cut_prompt);
            SAS.ClipboardPrompted.cut = true;
        }
        this.transControl().cut();*/
        if (!Clipboard.canAccessSystemClipboard()) {
            Clipboard.data(this.getText('selected'));
        }
        /*if (SAS.Browser.chrome()) {
            // It needs focus to cut correctly on Chrome
            var previous = document.activeElement;
            this.focus();
            this.view().cut();
            jQuery.sap.focus(previous);
        } else {*/
            this.view().cut();
        //}
    },

    copy: function () {
        /*if (!SAS.ClipboardPrompted.copy && !Clipboard.canAccessSystemClipboard()) {
            SAS.PromptDialog.show(document.title, this.resourceBundle.copy_prompt);
            SAS.ClipboardPrompted.copy = true;
        }
        this.transControl().copy();*/
        if (!Clipboard.canAccessSystemClipboard()) {
            Clipboard.data(this.getText('selected'));
        }
        /*if (sap.ui.Device.browser.chrome) {
            // It needs focus to copy correctly on Chrome
            var previous = document.activeElement;
            this.focus();
            this.view().copy();
            jQuery.sap.focus(previous);
        } else {*/
            this.view().copy();
        //}
    },

    paste: function () {
        if (this.readOnly()) return;
        if (!SAS.ClipboardPrompted.paste && !Clipboard.canAccessSystemClipboard()) {
            SAS.PromptDialog.show(document.title, this.resourceBundle.paste_prompt);
            SAS.ClipboardPrompted.paste = true;
        }
        if (SAS.Browser.msie()) {
            this.view().paste();
        } else {
            var selection = this.selection();
            this.editor().setText(this.core().cleanText(Clipboard.data()), selection.start, selection.end);
        }
    },
  /*collapse: function () {
        var context,
            line,
            block;
        if (!this.view()) { return; }
        context = this.view().getEditorContext();
        line = context.getSelectionRange().start.line;
        if (context.isCollapsed(line)) { return; }
        block = this.model().getCollapsibleBlock(line);
        context.collapse(block.startLine, block.endLine);
    },

    collapseAll: function () {
        if (!this.view()) { return; }
        this.view().getEditorContext().collapseAll();
    },

    expand: function () {
        var context,
            line;
        if (!this.view()) { return; }
        context = this.view().getEditorContext();
        line = this.modelTA().select().start.line;
        context.expand(line);
    },

    expandAll: function () {
        if (!this.view()) { return; }
        this.view().getEditorContext().expandAll();
    },*/
    _tryToExpandLine: function(line) {
        /*var block = this.foldingMgr.getFoldingBlock(line);
        while (block && this.foldingMgr.isCollapsed(block) && line > block.startLine) {
            this._expandBlock(true, block.anno);
            //this.transControl().expand(block);
            block = this.foldingMgr.getFoldingBlock(line);
        }*/
    },
    prevFunction: function() {
        this.view()._clientDiv.blur();
    },
    
    nextFunction: function() {
        this.view()._clientDiv.blur();
    },

    gotoline: function (line) {
        if(line >= 0) {
            line = line-1; //defect S0890739
        }
        if(line === undefined || line < 0 || line > this.model().getLineCount()) {
            //this.focus();
        } else {
            var selection, range, newRange, len;
            // get current range
            //that.focus();
            selection = this.selection();
            selection.start = selection.end;
            range = this.model().toTextRange(selection);
            // calculate target range
            newRange = new TextRange(range);
            newRange.start.line = newRange.end.line = line;
            len = this.model().getColumnCountWithoutDelimiter(line);
            if (newRange.end.column > len) {
                newRange.start.column = newRange.end.column = len;
            }
            this._tryToExpandLine(line);
            // go to
            this.selectText(this.model().toCoreRange(newRange));
        }
    },
    
    toggleLineNumber: function() {
        //this.settings.lineNumber(!this.settings.lineNumber());
        var enable = !this.settings.getLineNumberVisible();
        this.settings.setLineNumberVisible(enable);
    },

    getText: function (type) {
        var editor = this.editor();
        if (!editor) {
            return this.core().getInitContent();
        }
        var result = '';
        if (!type) {
            type = 'all';
        }
        switch (type) {
            case 'all':
                result = editor.getText();
                break;
            case 'selected':
                result = editor.getSelectionText();
                break;
        }
        return result;
    },

    lineCount: function(){
        var model = this.model();
        if (!model) {
            return this.core().getInitContent().split('\n').length;
        }
        return model.getLineCount();
    },

    comment: function() {
        if (!this.core().isSasMode() || this.readOnly()) return; // only support SAS

        var editor = this.editor(),
            selection = editor.getSelection(),
            allText = editor.getText(),
            start, end, text, commented, endWithEnter = false, RETURN_CHAR = '\n';

        start = selection.start === 0 ? 0 : (allText.lastIndexOf(RETURN_CHAR, selection.start - 1) + 1);
        if (selection.start < selection.end && editor.getText(selection.end - 1, selection.end) === RETURN_CHAR) {
            endWithEnter = true;
        }
        end = allText.indexOf(RETURN_CHAR, selection.end - (endWithEnter ? 1 : 0)); // Omit return character.
        if (end < 0) {
            end = allText.length;
        }
        text = editor.getText(start, end);
        commented = (/^\n*\s*\/\*.*(\*\/\s*(\n)*\/\*.*)*.*\*\/\s*\n*$/).test(text);
        //commented = (/^\/\*.*\*\/(\n.*)*$|^(\s)*\/\*.*(\*\/\n\/\*.*)*.*\*\/$/).test(text);
        //commented = (/^\/\*.*\*\/\n+(\n.*)*$|^\/\*.*(\*\/\n\/\*.*)*.*\*\/$/).test(text);
        //sas.log.info(commented)
        var newText = null;
        if (!commented) {
            newText = text.replace(/^(\s*)\/\* ?| ?\*\/(\s*)$/gm, '$1$2');
            newText = newText.replace(/^(.*)$/gm, '/* $1 */');
        } else {
            newText = text.replace(/^(\s*)\/\* ?| ?\*\/(\s*)$/gm, '$1$2');
        }
        editor.setText(newText, start, end);
        editor.setSelection(start, end + newText.length - text.length);
        return true;
    },

    toUpperCase: function() {
        if (this.readOnly()) return;
        var selection = this.selection();
        this.insertText(this.getText('selected').toUpperCase());
        this.selectText(selection);
        return true;
    },

    toLowerCase: function() {
        if (this.readOnly()) return;
        var selection = this.selection();
        this.insertText(this.getText('selected').toLowerCase());
        this.selectText(selection);
        return true;
    },

    toCamelCase: function() {
        if (this.readOnly()) return;
        var selection = this.selection(),
            text = this.getText('selected'),
            reg = /\b(\w)|\s(\w)/g;

        text = text.replace(reg, function(m){
            return m.toUpperCase();
        });
        this.insertText(text);
        this.selectText(selection);
        return true;
    },

    cycleCases: function() {
        var selectedText = this.getText('selected'),
            lowercaseExisted = /[a-z]/.test(selectedText),
            uppercaseExisted = /[A-Z]/.test(selectedText);
        if (lowercaseExisted && !uppercaseExisted){
            this.toCamelCase();
        } else if (!lowercaseExisted && uppercaseExisted){
            this.toLowerCase();
        } else {
            this.toUpperCase();
        }
    },

    search: function (key, config) {
        var allText = this.model().getAllText(),
            result = [];
        if (!config.matchCase) {
            key = key.toLowerCase(); allText = allText.toLowerCase();
        }
        if (config.all === true) {
            result = this.doSearchAll(key, config, allText);
            if (result.length !== 0) {
                var range = this.model().toCoreRange(result[0]);
                //this.view().setSelection(range.start, range.start);
                this.editor().setSelection(range.start, range.start);
            }
        } else {
            var selection = this.model().toTextRange(this.editor().getSelection());
            //var selection = this.model().toTextRange(this.view().getSelection());
            result = this.doSearch(key, config.inSelection ? selection.start : selection.end, allText, config.wholeWord);
            if (result.length !== 0){
                var range = this.model().toCoreRange(result[0]);
                //this.view().setSelection(range.start,range.end);
                this.editor().setSelection(range.start,range.end);
            }
        }

        if (result.length === 0) {
            SAS.FindAndReplaceDialog.getInstance().setFeedBack(this.resourceBundle.not_found);
        } else {
            var annotationModel = this.editor().getAnnotationModel(), range = null;
            annotationModel.removeAnnotations(this.core().getAnnotations().AnnotationType.ANNOTATION_MATCHING_SEARCH);
            for(var count=0;count<result.length;count++){
                range = this.model().toCoreRange(result[count]);
                annotationModel.addAnnotation(new (this.core().getAnnotations().AnnotationType.getType(this.core().getAnnotations().AnnotationType.ANNOTATION_MATCHING_SEARCH))(range.start, range.end));
            }

            //annos = getAnnotations(core.getAnnotations().AnnotationType.ANNOTATION_CURRENT_LINE);
            if (config.all) {
                SAS.FindAndReplaceDialog.getInstance().setFeedBack(this.resourceBundle.found_num.replace("${0}", result.length.toString()));
            }
        }
        return result;
    },
    
    // please not that we preserve the code below because it achieve the 'direction option'
    
    // doSearch: function (key, config) {
        // //Todo: may improve this function in the future.
        // var modelTA = this.modelTA(),
            // offset, lastPosition,
            // range = new SAS.TextRange(),
            // result = [];
        // if (config.all) {
            // range = modelTA.search(key, (new SAS.CoreRange()), 'forward', config.caseIgnored);
            // while (range) {
                // result.push(modelTA.toTextRange(range));
                // range = modelTA.search(key, range, 'forward', config.caseIgnored);
            // }
        // } else {
            // range = modelTA.search(key, modelTA.getSelectionRange(), config.direction, config.caseIgnored);
            // if (range) {
                // result.push(modelTA.toTextRange(range));
                // modelTA.moveCursor(range);
            // } else if (config.wrapped) {
                // if (config.direction === 'forward') {
                    // offset = new SAS.CoreRange();
                // } else {
                    // offset = new SAS.CoreRange(modelTA.getAllText().length - 1, modelTA.getAllText().length - 1);
                // }
                // range = modelTA.search(key, offset, config.direction, config.caseIgnored);
                // if (range) {
                    // result.push(modelTA.toTextRange(range));
                    // modelTA.moveCursor(modelTA.toCoreRange(range));
                // } else {
                    // modelTA.moveCursor(modelTA.getSelectionRange());
                // }
            // } else {
                // if (config.caseIgnored ? modelTA.getSelectedText().toLowerCase() === key.toLowerCase() : modelTA.getSelectedText() === key) {
                    // result.push(modelTA.toTextRange(modelTA.getSelectionRange()));
                // }
                // modelTA.moveCursor(modelTA.getSelectionRange().start, modelTA.getSelectionRange().end);
            // }
        // }
        // //this.view().getEditorContext().setFound(result); //Bin,Zhou: comment when replacing canvas with div+css
        // this.view().setFound(result);
        // return result;
    // },
    atomicSearch: function(key, keyStart, text, wholeWord) {
        if (wholeWord) {
            var escapedKey = key.replace(/[\\\.\*\+\?\|\(\)\{\}\^\$\[\]]/g, '\\$&'),
                searchResult = text.substring(keyStart).search('(\[^a-zA-Z0-9_\u00ff-\uffff]|^)' + escapedKey + '(\[^a-zA-Z0-9_\u00ff-\uffff]|$)');
            if (searchResult != -1) {
                var firstChar = text.substring(keyStart,keyStart+1);
                if (searchResult != 0 || (firstChar != key && !(/\w|[^\x00-\xff]/).test(firstChar))) ++searchResult;
                return keyStart + searchResult;
            }
            return -1;
        }
        return text.indexOf(key, keyStart);
    },
    
    doSearch: function(key, start, text, wholeWord) {
        var result = [],
            keyStart = 0,
            line = 0,
            model = this.model(),
            lineCount = model.getLineCount(),
            col = 0,
            offsetCount = 0,
            phase = 0;
        keyStart = this.atomicSearch(key, keyStart, text, wholeWord);
        while (keyStart !== -1 && line <= lineCount) {
            if (keyStart > offsetCount) {
                col = model.getColumnCount(line);
                offsetCount += col;
                line++;
            } else {
                var resultLine = line, resultCol = 0;
                if (keyStart < offsetCount) {
                    resultLine = line - 1;
                    resultCol = keyStart - (offsetCount - col);
                }
                if (start.line < resultLine ||
                    (start.line === resultLine && start.column <= resultCol)) {
                    result = [];
                    phase = 1;
                }
                if (phase || result.length === 0) {
                    result.push(new TextRange(new TextPosition(resultLine, resultCol),
                        new TextPosition(resultLine, resultCol + key.length)));
                    if (phase) break;
                }
                keyStart = this.atomicSearch(key, keyStart + key.length, text, wholeWord);
            }
        }
        if (result.length > 0) {
            this._tryToExpandLine(result[0].start.line);
        }
        return result;
    },
    
    doSearchAll: function(key, config, text, preventExpand) {
        var result = [],
            keyStart = 0,
            line = 0,
            model = this.model(),
            lineCount = model.getLineCount(),
            col = 0,
            offsetCount = 0,
            range = model.toTextRange(this.editor().getSelection());
        keyStart = this.atomicSearch(key, keyStart, text, config.wholeWord);
        while (keyStart !== -1 && line <= lineCount) {
            if (keyStart > offsetCount) {
                col = model.getColumnCount(line);
                offsetCount += col;
                line++;
            } else {
                var resultLine = line, resultCol = 0;
                if (keyStart < offsetCount) {
                    resultLine = line - 1;
                    resultCol = keyStart - (offsetCount - col);
                }
                if (config.inSelection && (range.end.line < resultLine ||
                    (range.end.line === resultLine && range.end.column < resultCol + key.length))) {
                    break;
                }
                if (!config.inSelection || range.start.line < resultLine ||
                    (range.start.line === resultLine && range.start.column <= resultCol)) {
                    if (preventExpand) {
                        var block = this.foldingMgr.getFoldingBlock(resultLine);
                        if (!(block && this.foldingMgr.isCollapsed(block) && resultLine > block.startLine)) {
                            result.push(new TextRange(new TextPosition(resultLine, resultCol),
                                new TextPosition(resultLine, resultCol + key.length)));
                        }
                    } else {
                        result.push(new TextRange(new TextPosition(resultLine, resultCol),
                            new TextPosition(resultLine, resultCol + key.length)));
                        this._tryToExpandLine(resultLine);
                    }
                }
                keyStart = this.atomicSearch(key, keyStart + key.length, text, config.wholeWord);
            }
        }
        return result;
    },
    
    replace: function(key, value, config) {
        var replacedWordCount,
            allText = this.model().getAllText(),
            view = this.view(),
            result = [],
            replaceType;
        if (!config.matchCase) {
            key = key.toLowerCase(); allText = allText.toLowerCase();
        }
        if (SAS.ReplacingContext.key !== key) {
            SAS.ReplacingContext.key = key;
            SAS.ReplacingContext.isFound = 0;
        }

        if (config.all) {
            replaceType = this.resourceBundle.replace_all;
            /*if (config.inSelection) {
                this.transControl().expandForEditing();
            } else {
                this.expandAll();
            }*/
            replacedWordCount = this.doReplaceAll(key, value, config);
            if (!config.inSelection) {
                view.setCaretOffset(view.getSelection().start);
            }
        } else {
            replaceType = this.resourceBundle.replace;
            result = this.doReplace(key, value, config, allText);
            replacedWordCount = result.length;
            if(replacedWordCount !== 0) {
                allText = this.model().getAllText();
                if (!config.matchCase) {
                    allText = allText.toLowerCase();
                }
                result = this.doSearch(key, this.model().toTextPosition(result[0].start + value.length), allText, config.wholeWord);
                var message = this.resourceBundle.no_occur_found;
                if (result.length) {
                    this.selectText(this.model().toCoreRange(result[0]));
                    message = this.resourceBundle.next_occur_found;
                }
                SAS.FindAndReplaceDialog.getInstance().setFeedBack(message);
                //modelTA.setCaretPosition(result[0].start + value.length);
            }
        }
        if(replacedWordCount !== 0) {
            SAS.ReplacingContext.isFound = 1;
        }
        this.showReplaceDialog(replacedWordCount, replaceType);
    },
    
    showReplaceDialog: function(replacedWordCount, replaceType) {
        var message, that = this;
        if(replacedWordCount == 0) {
            if(SAS.ReplacingContext.isFound != 0){
                message = this.resourceBundle.replaced_num.replace("${0}", '0');
            } else {
                message = this.resourceBundle.not_replaced;
            }
            //SAS.PromptDialog.show(replaceType, message);
            SAS.FindAndReplaceDialog.getInstance().setFeedBack(message);
        } else {
            if(replaceType == this.resourceBundle.replace_all) {
                message = this.resourceBundle.replaced_num.replace("${0}", replacedWordCount);
                //SAS.PromptDialog.show(replaceType, message);
                SAS.FindAndReplaceDialog.getInstance().setFeedBack(message);
            } else if(SAS.Browser.ff()) {
                window.setTimeout(function(){that.focus();},400);
            }
        }
    },
    
    doReplaceAll: function(key, value, config, allText) {
        var selectRange = this.editor().getSelection();
        key = key.replace(/[\\\.\*\+\?\|\(\)\{\}\^\$\[\]]/g, '\\$&');
        var oldText = this.editor().getModel().getText();
        if (config.wholeWord && !(/[^a-zA-Z0-9_\u00ff-\uffff]/).test(key)) { // whole word is meaningless for non-word characters
            var beginSign = '([^a-zA-Z0-9_\\u00ff-\\uffff]|\\b|^)(',  endSign = ')(?=[^a-zA-Z0-9_\\u00ff-\\uffff]|\\b|$)';
            if (config.inSelection) {
                if (selectRange.start !== 0 && oldText.slice(selectRange.start - 1, selectRange.start).search(/[^a-zA-Z0-9_\u00ff-\uffff]/) === -1) {
                    beginSign = '([^a-zA-Z0-9_\\u00ff-\\uffff])(';
                }
                if (selectRange.end !== oldText.length && oldText.slice(selectRange.end, selectRange.end + 1).search(/[^a-zA-Z0-9_\u00ff-\uffff]/) === -1) {
                    endSign = ')(?=[^a-zA-Z0-9_\\u00ff-\\uffff])';
                }
            }
            key = beginSign + key + endSign;
            value = '$1' + value;
        }
        var regExp = new RegExp(key, config.matchCase ? 'g' : 'gi');
        //this.modelTA().allText(this.modelTA().getAllText().replace(regExp, value), true , true);
        if (config.inSelection) oldText = oldText.slice(selectRange.start, selectRange.end);
        var replacedWordCount = oldText.match(regExp);
        if (replacedWordCount) {
            var newText = oldText.replace(regExp, value);
            var oldRange = config.inSelection ? selectRange : new CoreRange(0, oldText.length);
            var oldSelection = this.editor().getSelection();
            var newRange = config.inSelection ? new CoreRange(selectRange.start, selectRange.start + newText.length) : new CoreRange(0, newText.length);
            //this.undoMgr().add(newText, newRange, newRange, oldText, oldRange, oldSelection);
            this.editor().setText(newText, oldRange.start, oldRange.end);
            if (config.inSelection) {
                //this.view().setSelection(newRange.start,newRange.end);
                this.editor().setSelection(newRange.start,newRange.end);
                //this.modelTA().setSelection(newRange);
            }
            return replacedWordCount.length;
        }
        return 0;
    },
    
    doReplace: function(key, value, config, allText) {
        var view = this.view(),
            result,
            start = this.model().toTextPosition(view.getSelection().start);
        result = this.doSearch(key, start, allText, config.wholeWord);
        if (result.length !== 0) {
            //this.transControl().expandForEditing();
            result[0] = this.model().toCoreRange(result[0]);
            this.selectText(result[0]);
            key = this.getText('selected'); // deal with case ignored
            var newRange = new CoreRange(result[0].start, result[0].start + value.length);
            //this.undoMgr().add(value, newRange, newRange, key, result[0], result[0]);
            //modelTA.insertText(value, result[0], 'replace');
            this.editor().setText(value, result[0].start, result[0].end);
        }
        return result;
    },
    selectText: function(range) {
        //var pos = this.model().toTextPosition(range.start);
        //this.view().setSelection(range.start,range.end);
        this.editor().setSelection(range.start,range.end);
        //this.modelTA().setSelection(range);
        //this.view().scrollTo(pos.line, pos.column);
    }
};
//static property
SAS.EditorController.activeCtrl = null;

return SAS.EditorController;
});
