// (c) 2016, SAS Institute Inc.

define([
    "./LanguageService",
    "./Utils",
    //'sas/hc/ui/core/Icon',
    "./orion/code_edit/built-codeEdit",
    "./Mode",
    "./EditorModel",
    "./FoldingMgr",
    "./EventMgr",
    "./CoreRange"
], function(LanguageService, Utils, /*Icon,*/ OrionRequire, Mode, EditorModel, FoldingMgr, EventMgr, CoreRange) {
    "use strict";

    var codeEdit = new (OrionRequire('embeddedEditor/builder/embeddedEditor')),
        Annotations = OrionRequire('orion/editor/annotations'),
        TextView = OrionRequire("orion/editor/textView").TextView,
        LineNumberRuler = OrionRequire("orion/editor/rulers").LineNumberRuler,
        FoldingRuler = OrionRequire("orion/editor/rulers").FoldingRuler,
        KeyBinding = OrionRequire("orion/keyBinding"),
        //rb = sap.ui.getCore().getLibraryResourceBundle("sas.hc.ui.commons"),
        CRLF_REGEXP = /\r\n|\r/g;

    return function(id, content, langMode, settings) {
        var editor, view, viewModel, baseModel, annotationModel, langServ, prompt, self = this,
            rowHighlightDiv, _removedBlocks = [], _internalFolding = false, mode = new Mode(langMode),model,
            foldingMgr, evtMgr = new EventMgr;
        //this._frm = jQuery('<div id="' + id + '-frm"></div>');
        this._frm = jQuery('#'+id).css({width: '100%', height: '100%'}).attr('dir', 'ltr');

        function connectModel() {
            var changeEvt = {type: "textChanged"}, addFunc = langServ.add.bind(langServ);
            baseModel.addEventListener("Changing", function(e) {
                jQuery(self).triggerHandler({type: 'editAttempted'});
                evtMgr.notify('editAttempted');
                checkPrompt();
                changeEvt.removedText = baseModel.getText(e.start, e.start + e.removedCharCount);
                changeEvt.text = e.text;
                changeEvt.addedCharCount = e.addedCharCount;
                changeEvt.oldRange = self.model().toTextRange({start: e.start, end: e.start + e.removedCharCount});
            });

            baseModel.addEventListener('Changed', function(e) {
                changeEvt.newRange = self.model().toTextRange({start: e.start, end: e.start + e.addedCharCount});
                //window.setTimeout(addFunc, 0, jQuery.extend({}, changeEvt));
                langServ.add(changeEvt);
                jQuery(self).triggerHandler(changeEvt);
                evtMgr.notify('textChanged', changeEvt);
            });
        }

        function getFoldingIcon(isCollapsed) {
            return;
            var rm = sap.ui.getCore().createRenderManager(),
                foldingIcon = isCollapsed ? sas.icons.HC.EXPANDSECTION : sas.icons.HC.COLLAPSESECTION,
                tooltip = rb.getText(isCollapsed ? 'ce_menu_expand_txt' : 'ce_menu_collapse_txt'),
                div = jQuery('<div></div>'),
                oIcon = new Icon({
                    src: foldingIcon,
                    tooltip: tooltip
                });
            rm.render(oIcon, div[0]);
            var result = div.html();
            oIcon.destroy();
            return result;
        }

        function updateFolding() {
            var sasMode = self.isSasMode();
            /*view.setRedraw(false);
            _internalFolding = true;

            var remove = [], add = [];
            for (var i = 0; i < _removedBlocks.length; i++) {
                var anno = _removedBlocks[i].anno;
                if (anno) {
                    remove.push(anno);
                    anno.expand();
                }
            }
            _removedBlocks = [];

            if (!sasMode) {
                var iter = annotationModel.getAnnotations();
                while (iter.hasNext()) {
                    var anno = iter.next();
                    if (anno.type === Annotations.AnnotationType.ANNOTATION_FOLDING) {
                        remove.push(anno);
                        anno.expand();
                    }
                }
            }

            self.foldingMgr().iterateBlocks(0, baseModel.getLineCount() - 1, function(){return true;}, function(block) {
                if (!block.anno || !sasMode) {
                    var start = baseModel.getLineStart(block.startLine),
                        end = baseModel.getLineEnd(block.endFoldingLine),
                        anno = new (Annotations.AnnotationType.getType(Annotations.AnnotationType.ANNOTATION_FOLDING))(start, end, viewModel);
                    add.push(anno);
                    anno.html = anno._expandedHTML = getFoldingIcon();
                    anno._collapsedHTML = getFoldingIcon(true);
                    block.anno = anno;
                    anno.block = block;
                }
            });

            annotationModel.replaceAnnotations(remove, add);

            if (!sasMode) {
                add.forEach(function(anno) {
                    if (self.foldingMgr().isCollapsed(anno.block)) {
                        anno.collapse();
                    }
                });
            }

            _internalFolding = false;
            view.setRedraw(true);*/
            view.redraw();

            if (sasMode) {
                view.showSelection();
            }
        }

        function onAnnotationChanged(e) {
            var changed = false;
            if (e.changed && e.changed.length && !_internalFolding && self.isSasMode()) {
                e.changed.forEach(function(anno) {
                    if (anno.block) {
                        if (anno.expanded && self.foldingMgr().isCollapsed(anno.block)) {
                            self.foldingMgr().expand(anno.block);
                            changed = true;
                        } else if (!anno.expanded && !self.foldingMgr().isCollapsed(anno.block)) {
                            self.foldingMgr().collapse(anno.block);
                            changed = true;
                        }
                    }
                });
                if (changed && mode.getLanguage() !== 'sas') {
                    window.setTimeout(updateFolding, 0);
                }
            }
        }

        function onParsingState(e) {
            if (e.state === LanguageService.ParsingState.STARTING) {
                if (e.parsedRange && e.parsedRange.removedBlocks && e.parsedRange.removedBlocks.blocks) {
                    _removedBlocks = e.parsedRange.removedBlocks.blocks;
                }
            } else if (e.state === LanguageService.ParsingState.ENDED) {
                window.setTimeout(updateFolding, 0);
            }
        }

        function regEvents() {
            view.addEventListener('Selection', function() {
                var pos = self.model().toTextPosition(editor.getCaretOffset());
                pos.type = 'caretMoved';
                jQuery(self).triggerHandler(pos);
                evtMgr.notify('caretMoved', pos);
                var range = self.model().toTextRange(editor.getSelection());
                range.type = 'selectionChanged';
                jQuery(self).triggerHandler(range);
                evtMgr.notify('selectionChanged', range);
                annotationModel.removeAnnotations(Annotations.AnnotationType.ANNOTATION_MATCHING_SEARCH);
                self.updateRowHighlight();
            });
            view.addEventListener('Drop', function(e) {
                evtMgr.notify('drop', e);
            });
            view.addEventListener('Scroll', self.updateRowHighlight);
            view.addEventListener('KeyDown', function(e) {
                if (e.event.keyCode === SAS.Utils.KEY.SPACE) {
                    // See dojo/dijit/a11yclick.js
                    // dojo will prevent space key to be input
                    e.event.stopPropagation();
                }
            });
            // S1312557
            view._rootDiv.dojoClick = false;
        }

        this.sidebarSelection = function() {
            function onMouseMove(e) {
                endY = e.clientY;
                valid = true;
            }
            function onInterVal() {
                if (valid) {
                    var rect = view._rootDiv.getBoundingClientRect(),
                        endLine = view.getLineIndex(endY - rect.top + view.getTopPixel()),
                        start = view.getLineStart(startLine < endLine ? startLine : endLine),
                        end = view.getModel().getLineEnd(startLine < endLine ? endLine: startLine, true);
                    view.setSelection(startLine < endLine ? start : end, startLine < endLine ? end : start);
                    view.focus();
                    valid = false;
                }
            }
            var sidebar = editor.getLineNumberRuler(),
                startLine = 0, endY = 0, valid = false, timerId;
            sidebar.onClick = function(){};
            jQuery(sidebar.node).on('mousedown', function(e) {
                var rect = view._rootDiv.getBoundingClientRect();
                endY = e.clientY;
                startLine = view.getLineIndex(e.clientY - rect.top + view.getTopPixel());
                jQuery(document).on('mousemove', onMouseMove)
                    .one('mouseup', function(e) {
                        jQuery(document).off('mousemove', onMouseMove);
                        window.clearInterval(timerId);
                        valid = true;
                        onInterVal();
                    });
                timerId = window.setInterval(onInterVal, 20);
            });

            editor.getFoldingRuler().onClick = function(lineIndex) {
                FoldingRuler.prototype.onClick.apply(this, arguments);
                view.setCaretOffset(view.getLineStart(lineIndex));
                view.focus();
            };
        };

        function checkPrompt() {
            if (prompt) {
                prompt.remove();
                prompt = null;
            }
        }

        codeEdit.create({parent: self._frm[0], contentType: 'text/plain'}).then(function(editorViewer) {

            editor = editorViewer.editor;
            editor._highlightCurrentLine = function() {};
            /*if (editor.getSourceCodeActions()) {
                if (mode.getLanguage() === 'sas') {
                    editor.getSourceCodeActions().autoIndent = function() {};
                }
            }*/
            view = editor.getTextView();
            /*view._doTab = function(){
                self._tabAsSpace();
                return true;
            };*/
            viewModel = view.getModel();
            baseModel = viewModel.getBaseModel();
            model = new EditorModel(baseModel);
            langServ = mode.getHighlight(model);
            foldingMgr = new FoldingMgr(mode, model, settings);
            annotationModel = editor.getAnnotationModel();

            var sourceCodeActions = editor.getSourceCodeActions();
            sourceCodeActions.setAutoCompleteComments(false); // S1384583

            // S1288234
            // Have to prevent Orion's built-in save
            view.setAction('save', function(){return false;});

            if (SAS.Browser.msie()) { // S1320565
                view.setKeyBinding(new KeyBinding.KeyBinding(73, true, true), 'noop'); // Ctrl+Shift+I
                view.setKeyBinding(new KeyBinding.KeyBinding(66, true, true), 'noop'); // Ctrl+Shift+B
            }

            // S1346968
            view._getClipboardText = function (evt, handler) {
                var myHandler = function(text) {
                    if (text.length) {
                        handler(text.map(function(str) {
                            return self.cleanText(str);
                        }));
                    }
                };
                TextView.prototype._getClipboardText.apply(this, [evt, myHandler]);
            };

            view.removeEventListener('MouseMove', editor._listener.onMouseMove);
            view.removeEventListener('MouseOver', editor._listener.onMouseOver);

            // S1271974
            //prompt = jQuery('<pre class="sce-prompt-text"></pre>').html(settings.getPrompt()).appendTo(view._clipDiv);
            rowHighlightDiv = jQuery('<div class="rowHighlight"></div>').prependTo(view._rootDiv);

            view.onLineStyle = self._onLineStyle;
            view.setOptions({themeClass: "sce sce-" + mode.getLanguage()});
            //annotationModel.addEventListener('Changed', onAnnotationChanged);
            jQuery(langServ).on('parsingStateEvent', onParsingState);

            connectModel();
            regEvents();
            self.sidebarSelection();

            content = self.cleanText(content);
            editorViewer.setContents(content).then(function(){
                window.setTimeout(function() {
                    viewModel.setLineDelimiter('\n', true);
                    editorViewer.syntaxHighlighter.styler.setFoldingEnabled(false);
                    jQuery(editorViewer._editorContextMenuNode).remove();

                    self.updateFoldingBar();

                    // Have to call resize, otherwise performance will be bad on Chrome
                    view.resize();

                    // Attn: This will tell controller to init
                    jQuery(self).triggerHandler({type: 'coreReady'});
                    // Attn: controller inited
                }, 0);
            });
        });

        this.settings = function() {
            return settings;
        };

        this.editor = function() {
            return editor;
        };

        this.view = function() {
            return view;
        };

        this.form = function() {
            return this._frm;
        };

        this.baseModel = function() {
            return baseModel;
        };

        this.mode = function() {
            return mode;
        };

        this.model = function() {
            return model;
        };

        this.evtMgr = function() {
            return evtMgr;
        };

        this.foldingMgr = function() {
            return foldingMgr;
        };

        this.langService = function() {
            return langServ;
        };

        this.getDomRef = function() {
            return this._frm[0];
        };

        this.resize = function(wid, hei) {
            this._frm.css({width: wid, height: hei});
            if (view) {
                view.resize();
            }
        };

        this.getInitContent = function() {
            return content;
        };

        /*
         * pos : the line text before pos
         */
        this._calculateTab = function (pos, preText) { //TODO: optimize this
                var distance, allText, lineStart;
            if (!preText) {
                allText = this.editor().getText();
                lineStart = allText.lastIndexOf('\n', pos - 1) + 1;
                preText = allText.substring(lineStart, pos);
                pos = pos - lineStart;
            }
            distance = pos - preText.lastIndexOf('\t', pos) - 1;
            return settings.getTabText().slice(distance % settings.getTabSize());
        };

        /*
         * text : will be inserted into editor
         * start : the start offset which the 'text' will be inserted from
         */
        this._translateTab = function (text, start) {//TODO: optimize this
            var lines, i, tabIndex, lineText,
                allText, offset, offsetText;
            if (text.indexOf('\t') > -1) {
                allText = this.editor().getText();
                offset = allText.lastIndexOf('\n', start - 1) + 1;
                offsetText = allText.substring(offset, start);
                lines = text.split('\n');
                for (i = 0; i < lines.length; i++) {
                    if (i > 0) {
                        offsetText = "";
                    }
                    lineText = offsetText + lines[i];
                    tabIndex = lines[i].indexOf('\t') + offsetText.length;
                    while (tabIndex >= offsetText.length) {
                        lines[i] = lines[i].replace(/\t/, this._calculateTab(tabIndex, lineText.substring(0, tabIndex)));
                        lineText = offsetText + lines[i];
                        tabIndex = lineText.indexOf('\t', tabIndex);
                    }
                }
                return lines.join('\n');
            } else {
                return text;
            }
        };

        this.cleanText = function(text) { //TODO:
            var result = text.replace(CRLF_REGEXP,'\n');
            if(settings.getTabAsSpace()) {
                var selection = this.editor().getSelection();
                result = this._translateTab(result, selection.start);
            }
            return result;
        };

        this._tabAsSpace = function() {
            var selection = this.editor().getSelection();
            var newText = null;
            if (selection.start === selection.end) {
                newText = settings.getTabAsSpace() ? this._calculateTab(selection.start) : '\t';
                this.editor().setText(newText, selection.start, selection.end);
            } else {
                var range, regExp = new RegExp('^()', 'gm'), oldText;
                // TODO: whether there is performance problem.
                oldText = this.baseModel().getText();
                if (oldText[selection.end-1] === '\n') {
                    selection.end--;
                }
                range = new CoreRange(oldText.substring(0, selection.start).lastIndexOf('\n') + 1, selection.end);
                var tab = settings.getTabAsSpace() ? settings.getTabText() : '\t';
                oldText = this.getText('selected');
                newText = oldText.replace(regExp, tab);
                var selectedRange = this.model().toTextRange(selection);
                selectedRange.start.column += tab.length;
                selectedRange.end.column += tab.length;
                this.editor().setText(newText, range.start, range.end);
                var start = this.model().toCoreRange(selectedRange);
                this.editor().setSelection(start, start);
            }
        };

        this.isSasMode = function() {// Remove folding function to the non-sas language. This block will be removed if the ace module is ready.
            return langMode === 'sas';
        };

        this.refresh = function() {
            view.resize();
            view.redraw();
            this.updateFoldingBar();
            this.updateRowHighlight();
            view.setTopIndex(view.getTopIndex(true));
        };

        this.getKeyBinding = function() {
            return KeyBinding;
        };

        this.getAnnotations = function() {
            return Annotations;
        };

        this.readOnly = function(readOnly) {
            view.setOptions({"readonly": readOnly});
            if (readOnly) {
                jQuery(view._clientDiv).attr({'aria-readonly': true, 'role': 'textbox'});
            } else {
                jQuery(view._clientDiv).removeAttr('aria-readonly role');
            }
        };

        this.updateFoldingBar = function() {
            if (settings.getFoldingEnabled() && this.isSasMode()) {
                var mateFoldingBar = jQuery('<div class="ruler folding" style="float: left;"></div>').appendTo(view._leftDiv);
                mateFoldingBar.html(getFoldingIcon());
                var width = mateFoldingBar.css('width');
                mateFoldingBar.remove();
                var ruler = editor.getFoldingRuler();
                ruler.node.style.width = width;
            }
        };

        this.updateRowHighlight = function() {
            var selection = view.getSelection();
            if (settings.getLineHighlightingVisible() && selection.start === selection.end) {
                var line = view.getLineAtOffset(selection.start);
                rowHighlightDiv.css({top: view.getLinePixel(line) - view.getTopPixel() + 4, height: view.getLineHeight(line), display: 'block'});
            } else {
                rowHighlightDiv.css('display', 'none');
            }
        };

        this.getHTML = function(option) {
            var printFrame = jQuery('<div></div>').appendTo(document.body);
            var options = view.getOptions();
            options.parent = printFrame[0];
            options.model = undefined;
            var printView = new TextView(options);
            printView.onLineStyle = this._onLineStyle;

            var size = {};
            if (option && option.onlyVisible) {
                printView.setModel(viewModel);
                printView.setTopIndex(view.getTopIndex(true));
                size.width = Utils.removePX(this._frm.css('width'));
                size.height = Utils.removePX(this._frm.css('height'));
            } else {
                printView.setModel(baseModel);
                size = printView.computeSize();
            }
            printFrame.css({width: size.width + 8, height: size.height + 8});
            printView.resize();
            if (option && option.lineNumber) {
                printView.addRuler(editor.getLineNumberRuler());
                size.width += printView._leftDiv.clientWidth;
                printFrame.css({width: size.width + 8, height: size.height + 8});
                printView.resize();
            }
            var re = printFrame[0].outerHTML;
            printFrame.remove();
            return re;
        };

        this._onLineStyle = function(e) {
            var syntaxHighlightingEnable = settings.getSyntaxHighlighting();
            TextView.prototype.onLineStyle.apply(view, arguments);
            if (false === syntaxHighlightingEnable) {
                return;
            }

            var baseOffset = e.textView.getModel().mapOffset ? e.textView.getModel().mapOffset(e.lineStart) : e.lineStart,
                baseLine = baseModel.getLineAtOffset(baseOffset),
                styles = langServ.getSyntax(baseLine), re = [], j = 0;

            function _mergeStyles() {
                if (e.ranges && e.ranges[j]) {
                    var olds = e.ranges[j];
                    if (olds.start < news.end) {
                        if (olds.start > news.start) {
                            re.push({
                                start: news.start,
                                style: news.style,
                                end: olds.start
                            });
                        }

                        if (olds.end < news.end) {
                            re.push({
                                start: olds.start < news.start ? news.start : olds.start,
                                style: {styleClass: olds.style.styleClass + ' ' + news.style.styleClass},
                                end: olds.end
                            });
                            news.start = olds.end;
                            j++;
                            _mergeStyles();
                        } else {
                            if (olds.start > news.start) news.start = olds.start;
                            news.style = {styleClass: olds.style.styleClass + ' ' + news.style.styleClass};

                            if (olds.end === news.end) {
                                j++;
                            }
                        }
                    }
                }
            }

            for (var i = 0; i < styles.length; i++) {
                var s = styles[i], news = {};
                var colCount = self.model().getColumnCountWithoutDelimiter(baseLine);
                if (s.start === colCount) {
                    break;
                }
                news.start = s.start + e.lineStart;
                news.style = {styleClass: s.style};
                news.end = i === styles.length - 1 ? colCount : styles[i+1].start;
                news.end += e.lineStart;

                _mergeStyles();
                re.push(news);
            }
            e.ranges = re;

            var block = self.foldingMgr().getFoldingBlock(baseLine);
            if (block && block.startLine === baseLine) {
                e.style = e.style || {};
                e.style.styleClass = e.style.styleClass || '';
                e.style.styleClass += ' block-mark';
            }
        };
    };
});
