// (c) 2014, SAS Institute Inc.

define([
    "./CoreRange",
    "./TextPosition",
    "./TextRange",
    "./Basic"
], function(CoreRange, TextPosition, TextRange) {
    "use strict";

var EditorModel = function (baseModel) {
    this.getText = function (textRange) {
        var coreRange = this.toCoreRange(textRange);
        return baseModel.getText(coreRange.start, coreRange.end);
    };
    this.getAllText = function () {
        return baseModel.getText();
    };
    /*this.getAllRange = function () {
        return new TextRange(new TextPosition(0, 0), new TextPosition(_lines.length - 1, _lines[_lines.length - 1].length));
    };*/
    this.getLineCount = function () {
        return baseModel.getLineCount();
    };
    this.getColumnCount = function (line) {
        return baseModel.getLineEnd(line, true) - baseModel.getLineStart(line);
    };
    this.getColumnCountWithoutDelimiter = function (line) {
        return baseModel.getLineEnd(line, false) - baseModel.getLineStart(line);
    };
    this.getLine = function (line) {
        return baseModel.getLine(line);
    };
    /*this.getLengthWithTabAsSpaces = function (text) {
        var tabLength = 0,
            bufferLength = text.length,
            i;
        for (i = 0; i < bufferLength; i++) {
            tabLength += (text[i] === '\t') ? this.tabSize() : 1;
        }
        return tabLength;
    };
    this.getLastPosition = function (line) {
        if (isNaN(line)) {
            line = this.getLineCount() - 1;
        }
        return new TextPosition(line, this.getColumnCount(line));
    };*/
    this.getIndent = function (position) { // query the current indentation of specific line
        var lineText = this.getLine(position.line), indent,
            index = 0;
        while ((index < lineText.length) && (lineText[index] === "\t" || lineText[index] === ' ')) {
            index++;
        }
        indent = lineText.substring(0, index);

        return position.column >= indent.length ? indent : indent.substring(0, position.column);
    };
    /*this.getSyntax = function (line) {
        return _langService.getSyntax(line);
    }*/
    this.isEmpty = function () {
        return baseModel.getCharCount() === 0;
    };
    this.toTextPosition = function (offset) {
        var line = baseModel.getLineAtOffset(offset),
            col = offset - baseModel.getLineStart(line);
        return new TextPosition(line, col);
    };
    this.toTextOffset = function(line,col) {
        return baseModel.getLineStart(line) + col;
    };
    this.toTextRange = function (range) { //CoreRange to TextRange
        return new TextRange(this.toTextPosition(range.start), this.toTextPosition(range.end));
    };
    this.toCoreRange = function (range) { //TextRange to CoreRange
        return new CoreRange(this.toTextOffset(range.start.line, range.start.column), this.toTextOffset(range.end.line, range.end.column));
    };
};
    return EditorModel;
});
