// (c) 2014, SAS Institute Inc.

define([], function() {
    "use strict";

    return function(mode, model, settings) {
        var langService = mode.getHighlight();
        this.iterateBlocks = function(start, end, cond, callback) {
            var block, startVal = start, endVal = end;
            if (!(start instanceof Function)) {
                start = function() {
                    return startVal;
                };
            }
            if (!(end instanceof Function)) {
                end = function() {
                    return endVal;
                };
            }
            for(var i = start(); i <= end(); i++) {
                block = langService.getFoldingBlock(i);
                if (block && block.startLine === i) {
                    if (cond(block)) {
                        callback(block);
                    }
                    if (block.collapsed) i = block.endFoldingLine;
                }
            }
        };
        // iterate all collapsed blocks between start line and end line
        this.iterateCollapsed = function(start, end, callback) {
            this.iterateBlocks(start, end, function(block){
                return block.collapsed; //this.isCollapsed(block)
            }, callback);
        };
        this.iterateExpanded = function(start, end, callback) {
            this.iterateBlocks(start, end, function(block){
                return !block.collapsed;
            }, callback);
        };
        // iterate all visible lines between start line and end line
        this.iterateVisibleLines = function(start, end, callback) {
            var block;
            for (var i = start; i <= end; i++) {
                block = langService.getFoldingBlock(i);
                if (block && block.collapsed && i > block.startLine) {
                    i = block.endFoldingLine;
                    continue;
                }
                callback(i);
            }
        };
        this.getFoldingBlock = function(line) {
            if (settings.getFoldingEnabled()) {
                return langService.getFoldingBlock(line);
            }
            return false;
        };
        this.collapse = function(block) {
            block.collapsed = true;
            if (langService.toggleFoldWidget) {
                langService.toggleFoldWidget(block.startLine);
            }
            jQuery(this).triggerHandler('foldingChanged');
        };
        this.expand = function(block) {
            block.collapsed = false;
            if (langService.toggleFoldWidget) {
                langService.toggleFoldWidget(block.startLine);
            }
            jQuery(this).triggerHandler('foldingChanged');
        };
        this.isCollapsed = function(block) {
            return block.collapsed;
        };

        this.canCollapseAll = function () {
            if (settings.getFoldingEnabled()) {
                var count = model.getLineCount();
                for (var i = 0; i < count; i++) {
                    var block = this.getFoldingBlock(i);
                    if (block) {
                        if (!this.isCollapsed(block)) {
                            return true;
                        }
                        i = block.endFoldingLine;
                    }
                }
            }
            return false;
        };

        this.canExpandAll = function () {
            if (settings.getFoldingEnabled()) {
                var count = model.getLineCount();
                for (var i = 0; i < count; i++) {
                    var block = this.getFoldingBlock(i);
                    if (block) {
                        if (this.isCollapsed(block)) {
                            return true;
                        }
                        //i = block.endFoldingLine;
                    }
                }
            }
            return false;
        };
    };
});
