/**
 * @author SCNBIZ
 */
define([], function() {
SAS = window.SAS || {};

SAS.TextPool = function(){
    this.init();
};

SAS.TextPool.prototype = {
    init : function(){
        this.pools = [];
        this.poolIdx = 0;
        this.offset = 0;
        this.lineNo = -1;
        this.cacheLine = "";
    },
    isEmpty : function(){
        return (this.pools.length === 0);
    },
    add : function(text){
        this.pools.push(text);
    },
    end : function() {
        //return ((this.cacheLine.length === 0) && (this.poolIdx >= this.pools.length));
        return (this.poolIdx >= this.pools.length);
    },
    clear : function() {
        this.init();
    },
    /* 
     * return {
     *  lineNo: 
     *  text: 
     * };
     * if not meet '\n'
     */ 
    readLine : function(){
        var text, len, i, ch, line; 
        
        if (this.poolIdx < this.pools.length){
            text = this.pools[this.poolIdx];
            len = text.length;
            i = this.offset;
            while(i < len) {
                ch = text[i];
                if(ch === '\n') {// end one line
                    line = this.cacheLine + text.substring(this.offset, i-1);
                    this.cacheLine = "";
                    this.offset = i + 1;
                    this.lineNo++;
                    if (this.offset === len){
                        this.poolIdx++;
                        this.offset = 0;
                    }
                    return {
                        lineNo : this.lineNo,
                        text : line
                    }
                } else {
                    i++;
                    if(i === len) {//meet the end of current pool, and move to the next pool
                        this.poolIdx++;
                        this.cacheLine = text.substring(this.offset, i);
                        this.offset = 0;
                        i = 0;
                        if(this.poolIdx >= this.pools.length) { // no enough data
                            //break;
                            return {
                                lineNo : this.lineNo + 1, //not increase this.lineNo
                                text : this.cacheLine
                            }
                        }
                    }
                }// !else
            }// !while
        }// !if
        return null;
    }
}

SAS.LogListSyntaxColor = function(model){
    // public properties
    this.parsingState = new SAS.Event();
    
    // private properties
    var LineType = {
        'B':'llbyline',
        'D':'lldata',
        'E':'llerror',
        'F':'llfootnote',
        'H':'llheader',
        'M':'llmessage',
        'N':'llnote',
        'S':'llsource',
        'T':'lltitle',
        'W':'llwarn'
        },
        CC = {
        'OverPrint': '+',// continue
        'Skip1Line': '0', // 
        'Skip2Line': '-', // 
        'NewPage': '1'
        },
        model = model,
        syntaxTable = [],
        textPool = new SAS.TextPool(),
        log = [],
        currLineNo = 0,
        blocks = [],
        currentBlock,
        self = this;
        
    currentBlock = new SAS.CollapsibleRange();//this is ref to CollapsibleRange obj
    /* ***************************************************************
     *   Private Methods
     *****************************************************************/
    // code folding related 
    function startCollapsibleBlock(startLine, type){
        currentBlock.startLine = startLine;
    };
    function endCollapsibleBlock(endLine){
        if (currentBlock.startLine>=0){
            currentBlock.endLine = endLine;        
            blocks.push(new SAS.CollapsibleRange(currentBlock));
            //console.log('block('+currentBlock.startLine + ',' + currentBlock.endLine + ')');
            currentBlock.startLine = -1;
        }
    };
    function hasCollapsibleBlock(){
        return (currentBlock.startLine >= 0);
    };
    function resetCollapsibleBlockCache(){
        blocks=[];
    };
    // parsing related
    function addLine(lineNo, lt, line){
        syntaxTable[lineNo] = syntaxTable[lineNo] || [];
        syntaxTable[lineNo].push({
            style : (lt?lt:'text'),
            state : null,
            start : 0
        });
        syntaxTable[lineNo].push({
            style : 'text',
            state : null,
            start : line.length
        });
        //log.push(line);
        log[lineNo] = line;
    }
    function parse1Line(){
        var lineObj = textPool.readLine(), ret = false;//assume no content
        while (lineObj){
            var line = lineObj.text, 
                len = line.length, cc='X', lt='X';
            
            if (len >=2){
                lt = line[0];
                cc = line[1];
            }
            lt = lt.toUpperCase();
            if (cc == CC.Skip1Line){
                addLine(currLineNo, 'text', "");                
                currLineNo += 1;
            }else if(cc == CC.Skip2Line){
                addLine(currLineNo, 'text', "");
                currLineNo += 1;
                addLine(currLineNo, 'text', "");
                currLineNo += 1;
            }else if(cc == CC.OverPrint){
                lineObj = textPool.readLine();
                continue;
            }            
            // 
            if (len>=2){
                addLine(currLineNo, LineType[lt], line.substr(2, len-2));
            }else{
                addLine(currLineNo, 'text', "");
            }
            if (cc == CC.NewPage){
                endCollapsibleBlock(log.length-2);
                startCollapsibleBlock(log.length-1);
            }
            //
            //line = textPool.readLine();            
            currLineNo += 1;
            ret = true;
            break;
        }
        return ret;
    }
    function parse(){
        var timeout = new Date().getTime() + 20;
        while (parse1Line()) {
            if (new Date().getTime() > timeout) {
                break;
            }
        }
        if (!textPool.end()) {
            window.setTimeout(parse.bind(this), 100);
        }else{
            if (hasCollapsibleBlock()){
                endCollapsibleBlock(log.length-1);                
            }
            //model.store.setLines(log);
            self.parsingState.raise({state:SAS.LanguageService.ParsingState.ENDED});
        }
    }
    /* ***************************************************************
     *   Public interfaces
     *****************************************************************/
    this.log = function(){
        return log;
    }
    this.reset = function(){
        syntaxTable = [];
        textPool.clear();
        log = [];
        currLineNo = 0;
        blocks = [];
        //resetCollapsibleBlockCache();
    }
    this.getCollapsibleBlockRange = function(line){
        //line--;
        var i = 0;
        for (; i< blocks.length; i++){
            if ( (line >= blocks[i].startLine) && (line <= blocks[i].endLine)){
                return blocks[i];
            }
        }
        return null;
    }
    this.getSyntax = function(line){
        return (syntaxTable[line]) ? syntaxTable[line] : [];
    }
    this.add = function(e){
        if (textPool.isEmpty()){
            this.parsingState.raise({state:SAS.LanguageService.ParsingState.STARTING});
        }
        textPool.add(e.text);
        parse();
    }
    
};
return SAS.LogListSyntaxColor;
});