// (c) 2014, SAS Institute Inc.

define(["dojo/request/xhr",
        './Utils',
        "./mode/sas/SyntaxColor",
        //"./mode/sas/Formatter",
        //"./mode/eel/eel",
        "./mode/SyntaxColorerAdapter"],
    function(xhr,
             Utils,
             SasSyntaxColor,
             //SasFormatter,
             //EELMode,
             SyntaxColorerAdapter) {
        "use strict";

        SAS.codeEditor = {};
        /**
         * @class Defines the language mode supported by Editor
         * @deprecated
         * @static
         * @public
         */
        var langMode = SAS.codeEditor.LanguageMode = {
            Unset: 'none',
            Text: 'text',
            Html: 'html',
            SclCode: 'scl',
            SasCode: 'sas',
            EEL: 'eel',
            Xml: 'xml',
            Log: 'log',
            Lst: 'lst',
            Mdx: 'mdx',
            R: 'r',
            Custom: 'custom'
        };

        var mapping = SAS.codeEditor.LanguageModeClassesMapping = {};
        mapping['sas'] = {"highlight": SasSyntaxColor/*, "formatter": SasFormatter*/};
        //mapping[langMode.EEL] = {"mode": EELMode};


        var Mode = function(language, customConfig) {
            // private
            this._language = language || SAS.codeEditor.LanguageMode.SasCode;
            this._customConfig = customConfig;
            this._highlight = null;
            this._formatter = null;

            this._langModeClasses = SAS.codeEditor.LanguageModeClassesMapping[this._language] || {};
            //if (language === 'sas') {
            //    this._langModeClasses['sas'] = {"highlight": SasSyntaxColor, "formatter": SasFormatter};
            //}

            this._applyHighlightCustomConfig = function(highlight, customConfig) {
                //custom config pipeline operation
                //set custom config to langModeManager
                // TODO
            };

            this._getMode = function () {
                if (this._langModeClasses.mode) {
                    return new this._langModeClasses.mode();
                }
                var mode = 'ace/mode/' + this._language;
                if (this._customConfig) {
                    mode = this._getCustomMode();
                } else {
                    try {
                        xhr.get(Utils.getBaseUrl() + "/ace/mode-" + this._language + '.js', {sync: true});
                    } catch (e) {
                        console.error('Unsupported language!', this._language, 'CodeEditor');
                        return null;
                    }
                }
                return mode;
            };

            this._getCustomMode = function() {
                var customRules = this._customConfig.highlightRules;

                var oop = ace.require('ace/lib/oop'),
                    TextMode = ace.require('ace/mode/text').Mode,
                    TextHighlightRules = ace.require('ace/mode/text_highlight_rules').TextHighlightRules;

                var CustomMode = function() {
                    TextMode.call(this);
                    if (customRules) {
                        var CustomRule = function () {
                            TextHighlightRules.call(this);
                            this.$rules = customRules;
                        };
                        oop.inherits(CustomRule, TextHighlightRules);
                        this.HighlightRules = CustomRule;
                    }
                };
                oop.inherits(CustomMode, TextMode);

                return new CustomMode();
            };

            // private end

            // public
            this.getLanguage = function(){
                return this._language;
            };

            this.getHighlight = function(model){
                if (!this._highlight) {
                    if (this._langModeClasses.highlight) {
                        this._highlight = new this._langModeClasses.highlight(model);
                    } else {
                        this._highlight = new SyntaxColorerAdapter(this._getMode());
                    }
                }
                return this._highlight;
            };

            this.getFormatter = function(controller){
                if (!this._formatter && this._langModeClasses.formatter) {
                    this._formatter = new this._langModeClasses.formatter(controller);
                }
                return this._formatter;
            };

            // public end
        };
        return Mode;
    });
