define([], function() {
SAS = window.SAS || {}; // SAS namespace.

/**
* Settings of the system
*/
SAS.Settings = function () {
    var theme = {},
        themeName = 'default',
        lineNumber = true,
        tabSize = 4,
        tabAsSpace = false,
        tabText = '    ',
        syntaxHighlighting = true,
        lineWrapped = false,
        lineHighlightingVisible = true,
        resourceBundle = SAS.Utils.getResourceBundle("Settings_nls"),
        prompt = resourceBundle.prompt,
        readOnly = false,
        foldingEnabled = true,
        //AutoCompleteSettings
        acCommitKeys = [],
        acCommitKeysWithSpace = [],
        acCommitBySpacebar = true,
        acCommitByEnterOrTab = true,
        acCommitValueWithSpace = false,
        acPrefixLength = 2,
        hintEnabled = true,
        acEnabled = true;

    function _getRuleFromStyleSheet(styleSheet, selectorText) {
        var rules = null, i = 0, len = 0, rule = null;
        try {
            rules = styleSheet.cssRules || styleSheet.rules;
        } catch(err) { // In Firefox, there is an error saying: The operation is insecure. (CORS)
            //ignore, currently, maybe there exists the same error for the css of the code editor which is deployed on a different domain.
            rules = null;
        }
        if (rules === null) return rule;
        len = rules.length;
        for (; i < len && rule === null; i++) {
            if (rules[i].styleSheet) {
                rule = _getRuleFromStyleSheet(rules[i].styleSheet, selectorText);
            } else if (rules[i].selectorText === selectorText) {
                rule = rules[i];
            }
        }
        return rule;
    }
    function _getRule(selectorText){
        var styleSheets = document.styleSheets, i = 0, len = styleSheets.length, rule = null;
        //rules = window.getMatchedCSSRules(dojo.query(".sce")[0]);  //onley available in webkit
        for (; i < len && rule === null; i++) {
            if (styleSheets[i]) {
                rule = _getRuleFromStyleSheet(styleSheets[i], selectorText);				
            }
        }
        return rule;
    };

    this.fontSize = function(size) {
        var rule = _getRule(".sce, .sce textarea, .sce span, .sce pre, .sce label"), ret=16;
        if (rule) {
            ret = parseInt(rule.style.fontSize);
            if (!isNaN(size) && size != ret) {
                rule.style.fontSize = size + 'px';
                jQuery(this).triggerHandler({type: 'settingChanged', fontSize: size});
            }
        }
        return ret;
    };

    // Properties
    this.getTabText = function () {
        return tabText;
    };
    this.getTabAsSpace = function () {
        return this.tabAsSpaces();
    };
    this.getFoldingEnabled = function() {
        return foldingEnabled;
    };
    this.getReadOnly = function() {
        return this.readOnly();
    };
    this.getPrompt = function() {
        return this.promptText();
    };
    this.getIndentEnabled = function() {
        //return this.getIndentEnabled();
    };
    this.getLineNumberVisible = function() {
        return this.lineNumber();
    };
    this.setLineNumberVisible = function() {
        this.lineNumber.apply(this, arguments);
    };
    this.getLineHighlightingVisible = function() {
        return lineHighlightingVisible;
    };
    this.getTabSize = function() {
        return this.tabSize();
    };
    this.getFontSize = function() {
        return this.fontSize();
    };
    this.getSyntaxHighlighting = function() {
        return this.syntaxHighlighting();
    };
    this.getHintEnabled = function() {
        return this.hintEnabled();
    };
    this.getAutoCompleteEnabled = function() {
        return this.autoCompleteEnabled();
    };

    this.tabSize = function (ts) {
        var i;
        if (!isNaN(ts)) { //setter
            if (ts <= 0) { return tabSize; }
            if (ts > 25) { ts = 25; }
            if (tabSize !== ts) {
                tabSize = ts;
                tabText = "";
                for (i = 0; i < tabSize; i++) {
                    tabText += " ";
                }
                jQuery(this).triggerHandler({type: 'settingChanged', tabSize: ts});
            }
        }
        return tabSize;
    };

    this.syntaxHighlighting = function (set) {
        if (arguments.length > 0 && typeof (set) === 'boolean') {
            if (syntaxHighlighting !== set) {
                syntaxHighlighting = set;
                jQuery(this).triggerHandler({type: 'settingChanged', syntaxHighlighting: set});
            }
        }
        return syntaxHighlighting;
    };
    
    this.tabAsSpaces = function (set) {
        if (arguments.length > 0 && typeof (set) === 'boolean') {
            tabAsSpace = set;
            jQuery(this).triggerHandler({type: 'settingChanged', tabAsSpace: set});
        }
        return tabAsSpace;
    };

    this.lineNumber = function (set) {
        if (arguments.length > 0 && typeof (set) === 'boolean') {
            if (lineNumber !== set) {
                lineNumber = set;
                jQuery(this).triggerHandler({type: 'settingChanged', lineNumberVisible: set});
            }
        }
        return lineNumber;
    };

    this.lineWrapped = function (set) {
        if (arguments.length > 0 && typeof (set) === 'boolean') {
            if (lineWrapped !== set) {
                lineWrapped = set;
                jQuery(this).triggerHandler({type: 'settingChanged', lineWrapped: set});
            }
        }
        return lineWrapped;
    };

    this.promptText = function (text) {
        if (arguments.length > 0 && prompt !== text) {
            prompt = text;
            jQuery(this).triggerHandler({type: 'settingChanged', promptText: prompt});
        }
        return prompt;
    };

    this.readOnly = function(_readOnly) {
        if (typeof(_readOnly) === 'boolean' && readOnly !== _readOnly) {
            readOnly = _readOnly;
            jQuery(this).triggerHandler({type: 'settingChanged', readOnly: readOnly});
        }
        return readOnly;
    };
    //AutoCompleteSettings
    this.hintEnabled = function (option) {
        if (typeof(option) === 'boolean' && hintEnabled !== option) {
            hintEnabled = option;
            jQuery(this).triggerHandler({type: 'settingChanged', hintEnabled: hintEnabled});
        }
        return hintEnabled;
    };
    
    this.autoCompleteEnabled = function (option) {
        if (typeof(option) === 'boolean' && acEnabled !== option) {
            acEnabled = option;
            jQuery(this).triggerHandler({type: 'settingChanged', acEnabled: acEnabled});
        }
        return acEnabled;
    };
    this.autoCompletePrefixLength = function() {
        return acPrefixLength;
    }
};
return SAS.Settings;
});