define(["./LanguageService","./ace/tokenizer", "./ace/background_tokenizer", "./ace/mode/xml_highlight_rules"], function(LanguageService, Tokenizer, BackgroundTokenizer, XmlHighlightRules) {

var SyntaxColorerAdapter = function (model) {
    var defaultSyntax = [];
    defaultSyntax.push({style:'text',start:0});
    this.parsingStateEvent = new SAS.Event();
    if (!this.createHighlightRulesObj)  {
        this.createHighlightRulesObj = function() {
            return new XmlHighlightRules();
        }
    }
    var tokenizer = new Tokenizer(this.createHighlightRulesObj().getRules());
    var DocAdapter = function(model) {
        this.model = model;
    };
    (function(){
        this.getLength = function() {
            return this.model.getLineCount();
        };
        this.getLines = function(start, end) {
            var lines = [], i = start;
            for (;i<=end;i++) {
                lines.push(this.model.getLine(i));
            }
            return lines;
        };
        this.getLine = function(row) {
            return this.model.getLine(row);
        };
    }).call(DocAdapter.prototype);
    var bgTokenizer = new BackgroundTokenizer(tokenizer);
    bgTokenizer.setDocument(new DocAdapter(model));
    function onUpdate(data) {
        parsingState = LanguageService.ParsingState.ENDED;
        this.parsingStateEvent.raise({state:parsingState});
    };
    bgTokenizer.parsingUpdate.add(onUpdate.bind(this));
    //TODO: implement in the future
    function handleStyle(type) {
        var types = type.split('.');
        var style = '';
        types.forEach(function(item){
            style += ' ' + item;
        });
        return style;
    }
    this.getSyntax = function(line){
        //var lines = bgTokenizer.getTokens(line,line);
        //var state = lines[0].state, tokens = lines[0].tokens, i = 0, len = tokens.length, start = 0;
        var tokens = bgTokenizer.getTokens(line), i = 0, len = tokens.length, start = 0;
        var syntax = [];
        for(;i<len;i++) {
            syntax.push({
            style: handleStyle(tokens[i].type),
            start: start
            });
            start += tokens[i].value.length;
        }
        //return defaultSyntax;
        return syntax;
    };
    this.getFoldingBlock = function(line, col){
        return null;
    };
    this.add = function(change){
        parsingState = LanguageService.ParsingState.STARTING;
        this.parsingStateEvent.raise({state:parsingState});
        
        bgTokenizer.start(change.oldRange.start.line);
        
        parsingState = LanguageService.ParsingState.ENDED;
        this.parsingStateEvent.raise({state:parsingState});
    };
    this.type = function(line,col) {
        return 'text';
    };
    //TODO: This is ugly, to be improved in the future.
    this.getCurrentZone = function(line, col) {
        return null;
    };
    this.getProcName = function(){ return null;};
    this.getStmtName = function(){return null;};
};
return SyntaxColorerAdapter;
});