define(["dojo/_base/declare", "./CoreRange", "./TransBase", "./TransFactory", "./TransDriver", "./EditContext"], 
function(declare, CoreRange, TransBase, TransFactory, TransDriver, EditContext) {
    var TransControl = declare("", null, {
        MODIFY_TRANS: {
            "basic":1,"tab":1,"deltab":1,"backspace":1,"del":1,"return":1,
            "undo":1,"redo":1,"deleteword":1,"cut":1,"paste":1,"cutsim":1,"pastesim":1,
            "comment":1,"ime":1
        },
        PATTERN_TO_TRANS_NAME: { //general
            "-mousemove-focusin-input":"basic",  // Defect S1062879, Issue #23, 'input' event occurs after select trans (select from line number bar)
            "-focusin-input":"basic",  // Defect S1062879, Issue #23, 'input' event occurs after select trans (select from line number bar)
            "-input":"basic",  // Defect S1062879, Issue #23, 'input' event occurs after select trans (select from line number bar)
            "-keydown-keypress-input":"basic", //special
            "-keydown-input":"basic", //IME mode
            "-mousemove-mouseup-input":"basic",//In Chrome, there exists event sequence "mousemove mouseup input keydown keyup input" after typing two characters continuously
            "-keydown-keyup-input":"basic", //IME mode
            "-keyup-input":"basic", //IME mode
            "-keypress-input": "basic", //english input with IME
            "-focusout-focusin-input": "basic",// This is related to 'scrollbymouse'
            //"-ctrlzdoc-focusin-input": "ugly-undo",  //Defect S1062879, Issue #13,#16, but not fix #17
            "-ctrlzdoc":"disabledocundo",//Defect S1062879, Issue #13, #16, #17
            "-keydown-compend-input":"basic", // Windows safari, space key
            //commands
            "-tab":     "tab",     
            "-shifttab":"deltab",
            "-backspace":"backspace",
            "-del":     "del",
            "-enter":  "return",
            // ctrl+d 
            "-pageup":  "pageup",//page up
            "-pagedown":  "pagedown",//pagedown
            "-ctrlup": "scrollbydelta", //ctrl + up
            "-ctrldown": "scrollbydelta", //ctrl + down
            "-end":     "setcursor",//end
            "-home":    "home",     //home
            "-ctrlend": "setcursortoend",//ctrl+end
            "-ctrlhome":"setcursortostart",//ctrl+home
            "-ctrlshifthome":"selectthetextbeforecaret",
            "-ctrlshiftend":"selectthetextaftercaret",
            "-left":    "setcursor",//left
            "-ctrlleft":"setcursor",//ctrl+left
            "-up":      "setcursor",//up
            "-right":   "setcursor",  //right
            "-ctrlright": "setcursor",//ctrl+right
            "-down":    "setcursor",//down
            "-ctrlshiftleft": "selectbyword",//ctrl+shift+left
            "-ctrlshiftright": "selectbyword",//ctrl+shift+right
            "-ctrlz": "undo",//ctrl+z
            "-ctrlshiftz": "redo",//ctrl+shift+z
            "-ctrly": "redo",//ctrl+y
            "-ctrldel": "deleteword",//ctrl+delete
            "-ctrlbackspace": "deleteword",//ctrl+backspace
            "-ctrlshiftu":"cyclecase",//ctrl+shift+u
            "-ctrlu":"uppercase",//ctrl+u
            "-ctrll":"lowercase",//ctrl+l
            //shift+delete
            //shift+backspace
            //ctrl+shift+backspace
            //ctrl+alt+backspace
            //alt+backspace
            "-mousedown":"setcursorbymouse",
            "-mousedown1":"scrollbymouse",
            "-mousedown2":"setcursorbymouse", 
            "-select": "select",
            "-mousedown-mousemove":"select",
            "-cutsim":"cutsim",
            "-copysim":"copysim",
            "-pastesim":"pastesim",
            "-keydown-paste":"paste",
            "-keypress-paste":"paste",
            "-keydown-cut":"cut",
            "-keypress-cut":"cut", //FireFox
            "-keydown-copy":"copy",
            "-keypress-copy":"copy",
            "-paste":"paste",
            "-cut":"cut",
            "-copy":"copy",
            "-compstart":"ime",
            "-sidebarselect": "sidebarselect",
            "-touchstart":"selectbytouch"
        },
        PATTERN_TO_TRANS_NAME_MAC: { //for mac
            "-altleft":    "setcursor",//alt+left
            "-altright":   "setcursor",//alt+right
            "-metaslash":  "comment",//meta+slash
            "-metaz":"undo",
            "-metashfitz":"redo",
            "-metay":"redo",
            "-metashiftu":"cyclecase",
            "-altup": "scrollbydelta",
            "-altdown": "scrollbydelta"
        },
        // FIXID S0889394: Strange behavior by pressing "ctrl+shift" key in Chrome21 in CE.
        // The reason is that "ctrl+shift" brings up the input event of textarea when Arabic input method(s) being enabled, 
        // but the data in texarea has not be changed.
        // preventDefault() and stopPropagation() cannot stop the input event, 
        PATTERN_TO_TRANS_NAME_CHROME: { //for chrome
            "-ctrlshift":    ""//ctrl+shift, it will be igored after the improvement
        },
        PATTERN_TO_TRANS_NAME_IE: { //for ie
            //FIXID: 0888235. Shift+END is not supposed to select the last return character
            "-shiftend":    "not-select-return",//shift+end
            "-esc":   "esc",//esc, this should be disabled
            //"-dblclick": "selectword"
        },
        PATTERN_TO_TRANS_NAME_MSAFARI: { //for msafari
            "-input":"undoredo-msafari", //undo, redo
            "-scroll-input":"undoredo-msafari", //undo, redo
            "-mousemove-input":"undoredo-msafari", //undo, redo
            "-keyup-input":"undoredo-msafari", //undo, redo. This will override the item in PATTERN_TO_TRANS_NAME on mobile safari.
            "-ctrlw":"",//ctrl+w  ?
            "-metam":"return",//meta+m
            "-metaj":"return",//meta+j
            "-metah":"del",//meta+h
            "-ctrltab":"tab",//ctrl+tab
            "-ctrlenter":"return",//ctrl+enter
            "-ctrlbackspace":"del",
            "-metaslash": "comment",//meta+slash
            "-metaz":"undo",
            "-metashfitz":"redo",
            "-metay":"redo",
            "-metashiftu":"cyclecase",
        },
        PATTERN_TO_TRANS_NAME_COMMENT: { //comment function for ie, ff, chrome and safari on windows, no msafari and mac
            "-ctrlslash":    "comment"//ctrl+slash
        },
        PATTERN_DISABLED: {
            "-shiftdelete":1,
            "-shiftbackspace":1,
            "-ctrlshiftbackspace":1,
            "-ctrlaltbackspace":1,
            "-altbackspace":1
        },
        PATTERN_ALLOWING_ONE: {
            "-mousemove":1,
            "-scroll":1,
            "-keyup":1
        },
        START_PATTERN: {
            'paste':1,
            'copy':1,
            'cut':1,
            'keydown':1,
            'keypress':1,
            'mousedown':1,
            'MSPointerDown':1,  // //FIXID S1118940
            'pointerdown':1,
            //'focusin':1,
            'blur':1,
            //'mousemove':1,
            'compositionstart':1,
            'select':1,
            'touchstart':1
        },
        SAVE_CONTEXT: {
            'mouseup':1,
            'MSPointerUp':1, //FIXID S1118940
            'pointerup':1,
            'keyup':1
        },
        constructor: function(control, modelTA, model) {
            this.pattern = "";
            this.trans = null;
            this.lastTrans = null;
            this.control = control;
            this.modelTA = modelTA;
            this.textarea = this.modelTA.getTextarea();
            this.model = model;
            //this.model = this.modelTA.model();
            this.context = new EditContext();
            //this.state = TransBase.TS.END;
            if(SAS.Browser.mac()) {
                declare.safeMixin(this.PATTERN_TO_TRANS_NAME, this.PATTERN_TO_TRANS_NAME_MAC);
            } else if(SAS.Browser.msie()) {
                declare.safeMixin(this.PATTERN_TO_TRANS_NAME, this.PATTERN_TO_TRANS_NAME_IE);
            } else if(SAS.Browser.msafari()) {
                declare.safeMixin(this.PATTERN_TO_TRANS_NAME, this.PATTERN_TO_TRANS_NAME_MSAFARI);
            } else if(SAS.Browser.chrome()) {
                declare.safeMixin(this.PATTERN_TO_TRANS_NAME, this.PATTERN_TO_TRANS_NAME_CHROME);
            }
            if (!SAS.Browser.msafari() && !SAS.Browser.mac()) {
                declare.safeMixin(this.PATTERN_TO_TRANS_NAME, this.PATTERN_TO_TRANS_NAME_COMMENT);
            }
            if (SAS.Browser.mac()) {
                declare.safeMixin(this.PATTERN_DISABLED, {
                    '-ctrld':1
                });
            }
        }, 
        end: function() {
            //if(this.trans)console.log('--------------trans end: '+ this.trans.name);
            this.lastTrans = this.trans;
            this.trans = null;
            this.pattern = "";
            this.guard = window.setTimeout(this.guardCheck.bind(this), 1000);
        },
        saveContext: function() {//console.log('======saveContext');
            this.modelTA.updateSelectionCache();
            this.preContext = this.context;
            this.context = new EditContext();
            this.context.selection = this.modelTA.getSelectionRange();
            this.context.selectedText = this.modelTA.getSelectedText();
            //this.context.length = this.textarea.value.length;
            this.context.length = this.model.length(); //this.modelTA.getLength();
            this.context.lastChar = this.modelTA.getSelectedText(this.context.selection.start - 1, this.context.selection.start);
        },
        config: function() {
            var self = this;
            return {modelTA: this.modelTA,
                model: this.model,
                control: this.control,
                view: this.control.view(),
                context: this.context,
                preContext: this.preContext,
                onEnd: function(e){self.end(e);}};
        },
        createTrans: function(pattern) {
            var i, name, trans = null;
            // filter
            i = pattern.lastIndexOf("-keydown");
            if (i>0) {
                pattern = this.pattern.substring(i);
            }
            name = this.PATTERN_TO_TRANS_NAME[pattern];
            if (name) {
                if (!this.modelTA.readOnly() || !this.MODIFY_TRANS[name]) {
                    trans = TransFactory.createTrans(name, this.config());
                }
            }
            if (trans && this.guard) {
                window.clearTimeout(this.guard);
                this.guard = null;
            }
            return trans;
        },
        tryStop: function(e) {
            var name = this.PATTERN_TO_TRANS_NAME[this.pattern], r = true;
            if (this.PATTERN_DISABLED[this.pattern] ||
                (this.modelTA.readOnly() && name && this.MODIFY_TRANS[name])) {
                //console.log('-------------------------------------disabled');
                TransDriver.stop(e);
                r = false;
            }
            return r;
        },
        dispatch: function(name, e) {
            var r = true, pattern;
            //if (e.type !== 'mousemove' && e.type !== 'MSPointerMove') console.log('event:'+name+',type:'+e.type+',keyCode:'+e.keyCode);
            if (this.trans === null || this.trans.allow(e)) { //(1) enter when no trans, (2) enter when event is allowed for the trans
                // call the event handler
                if (this[name]) {
                    this[name].apply(this, [e]);
                }
                // try to create a trans
                if (this.trans === null) {
                    // create pattern
                    pattern = name.replace(/^on/,'-').toLowerCase();
                    if (this.START_PATTERN[e.type]) {
                        this.pattern = '';
                        //this.saveContext();
                    }
                    if (this.pattern === '' || this.PATTERN_ALLOWING_ONE[pattern]===undefined) { //allow the first '-mousemove' to add
                        if (pattern === '-mousedown') {
                            if (e.button === 1 || e.button === 2) {
                                pattern += e.button
                            }
                        }
                        this.pattern += pattern;
                    }
                    if (this.START_PATTERN[e.type] || this.SAVE_CONTEXT[e.type]) {
                        this.saveContext();
                    }
                    //if (e.type !== 'mousemove')console.log("PATTERN:"+this.pattern);
                    this.trans = this.createTrans(this.pattern);
                    //if(this.trans)console.log('--------------trans start: '+ this.trans.name);
                }
                // execute the trans
                if (this.trans) {
                    r = this.trans.exec(e);
                } else {
                    r = this.tryStop(e);
                }
            } else {
                //console.log('-------------------------------------'+ e.type +' disabled');
                TransDriver.stop(e);
                r = false;
            }
        },
        tryCreateAndExecTrans: function(pattern) {
            var trans = this.createTrans(pattern);
            if (trans) {
                trans.exec();
            }
        },
        undo: function() {
            this.tryCreateAndExecTrans('-ctrlz');
        },
        redo: function() {
            this.tryCreateAndExecTrans('-ctrly');
        },
        cut: function() {
            this.tryCreateAndExecTrans('-cutsim');
        },
        copy: function() {
            this.tryCreateAndExecTrans('-copysim');
        },
        paste: function() {
            this.tryCreateAndExecTrans('-pastesim');
        },
        guardCheck: function() {
            var modelText = this.model.getAllText();
            if (this.textarea.value !== modelText) {
                this.textarea.value = modelText;
                this.modelTA._updateLines();
                //console.log('oh something wrong! recovered...');
            }
        }
    });
    return TransControl;
});