define(["dojo/on", "dojo/has", "./CmdMgr"], function(on, has, CmdMgr) {
    var TransDriver = function(textarea, control) {
        this.KEY = SAS.Utils.KEY;
        this.listeners = {};
        this.cmdMgr = new CmdMgr();
        this.textarea = textarea;
        this.control = control;
        //this.on = on;
        var events = [
            'input', 'keydown', 'keyup', 'keypress', 'select', 
            'compositionstart', 'compositionend', 
            'touchstart', 'touchend', 
            'mousedown', 'mousemove', 'mouseup',
            'dragover', 'drop',
            'focusin', 'focusout',
            'paste', 'copy', 'cut',
            'contextmenu',
            'scroll', 'dblclick'
            ], 
            handlers = [
            this.onInput, this.onKeyDown, this.onKeyUp, this.onKeyPress, this.onSelect, //1
            this.onCompStart, this.onCompEnd,//2
            this.onTouchStart, this.onTouchEnd, //3
            this.onMouseDown, this.onMouseMove, this.onMouseUp,//4
            this.onDragOver, this.onDrop,//5
            this.onFocusIn, this.onFocusOut,//6
            this.onPaste, this.onCopy, this.onCut, //7
            this.onContextMenu,
            this.onScroll, this.onDblClick
        ];
        if (has("MSPointer")){ //FIXID S1118940
            if (has("pointer-events")){ // IE10
                events[9] = 'pointerdown';
                events[10] = 'pointermove';
                events[11] = 'pointerup';
            } else { // IE11
                events[9] = 'MSPointerDown';
                events[10] = 'MSPointerMove';
                events[11] = 'MSPointerUp';
            }
        }
        this.regListeners(events, handlers);
        this.regCmds();
        if (SAS.Browser.webkit()) {
            this.keyDownListener = on(document, "keydown", this.onKeyDownDoc.bind(this))
        }
    };
    TransDriver.stop = function (e) {
        e.preventDefault();
        e.stopPropagation();
        return false;
    };
    TransDriver.prototype = {
        destroy: function() {
        },
        /*dispatch: function(name, e) {
            this.control.dispatch(name, e);
        },*/
        regCmds: function() {
            var self = this;
            function _h(fn) {
                return function(e){
                fn.apply(self,arguments);};
            }
            this.cmdMgr.regCommands([
                //Special Keyboard Event Process: tab, shift+tab, backspace, del, return
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.TAB,  handle:_h(this.onTab)},           //Tab
                {ctrlKey:false, altKey:false, shiftKey:true, keyCode: this.KEY.TAB,  handle:_h(this.onShiftTab)},   //Shift+Tab
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onBackSpace)},     //Backspace
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.DEL,  handle:_h(this.onDel)},          //Delete
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.ENTER,  handle:_h(this.onEnter)},       //Return
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.D,  handle:_h(this.onCtrlD)},     //Ctrl+D
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.INSERT,  handle:_h(this.onInsert)},     //Ctrl+D
                
                //Cursor Event Process: pgup, pgdown, end, home, left, up, right, down
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.PAGE_UP,  handle:_h(this.onPageUp)},  //Page Up
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.PAGE_DOWN,  handle:_h(this.onPageDown)},  //Page Down
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.END,  handle:_h(this.onEnd)},  //End
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.HOME,  handle:_h(this.onHome)},  //Home
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.END,  handle:_h(this.onCtrlEnd)},   //Ctrl+End
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.HOME,  handle:_h(this.onCtrlHome)},   //Ctrl+Home
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.END,  handle:_h(this.onCtrlShiftEnd)},   //Ctrl+Shift+Home
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.HOME,  handle:_h(this.onCtrlShiftHome)},   //Ctrl+Shift+Home
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.LEFT,  handle:_h(this.onLeft)},  //Left
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.LEFT,  handle:_h(this.onCtrlLeft)},   //Ctrl+Left
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.UP,  handle:_h(this.onUp)},  //Up
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.UP,  handle:_h(this.onCtrlUp)},  //Up
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.RIGHT,  handle:_h(this.onRight)},  //Right
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.RIGHT,  handle:_h(this.onCtrlRight)},   //Ctrl+Right
                {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.DOWN,  handle:_h(this.onDown)},  //Down
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.DOWN,  handle:_h(this.onCtrlDown)},  //Down
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 37,  handle:enableFunction},  //Shift+Left
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 38,  handle:enableFunction},  //Shift+Up
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 39,  handle:enableFunction},  //Shift+Right
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 40,  handle:enableFunction},  //Shift+Down 
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 33,  handle:enableFunction},  //Shift+Page Up
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 34,  handle:enableFunction},  //Shift+Page Down
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 35,  handle:enableFunction},  //Shift+End
                //{ctrlKey:false, altKey:false, shiftKey:true, keyCode: 36,  handle:enableFunction},  //Shift+Home
                
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.LEFT,  handle:_h(this.onCtrlShiftLeft)},  //Ctrl+Shift+Left
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.RIGHT,  handle:_h(this.onCtrlShiftRight)},  //Ctrl+Shift+Right
                //{ctrlKey:true, altKey:false, shiftKey:true, keyCode: 35,  handle:enableFunction},  //Ctrl+Shift+End
                //{ctrlKey:true, altKey:false, shiftKey:true, keyCode: 36,  handle:enableFunction},  //Ctrl+Shift+Home
                //Special Shortcut Key Event Process: Ctrl+Z, Ctrl+Shift+Z, Ctrl+Y
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.Z,  handle:_h(this.onCtrlZ)},       //Ctrl+Z
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.Z,  handle:_h(this.onCtrlShiftZ)},        //Ctrl+Shift+Z
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.Y,  handle:_h(this.onCtrlY)},       //Ctrl+Y
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.U,  handle:_h(this.onCtrlShiftU)},   //Ctrl+Shift+U
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.U,  handle:_h(this.onCtrlU)},  //Ctrl+U
		{ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.L,  handle:_h(this.onCtrlL)},  //Ctrl+L
                // for editing
                //{ctrlKey:true, altKey:false, shiftKey:false, keyCode: 67,  handle:enableFunction},  //Ctrl+C        
                //{ctrlKey:true, altKey:false, shiftKey:false, keyCode: 86,  handle:enableFunction},  //Ctrl+V
                //{ctrlKey:true, altKey:false, shiftKey:false, keyCode: 88,  handle:enableFunction},  //Ctrl+X
                //{ctrlKey:true, altKey:false, shiftKey:false, keyCode: 65,  handle:enableFunction},  //Ctrl+A
                // for debug
                //{ctrlKey:true, altKey:false, shiftKey:true, keyCode: 73,  handle:enableFunction},  //Ctrl+Shift+I
                // disable IME,Ctrl+Delete,Shift+Delete, Ctrl+Backspace, Shift+Backspace,etc
                //{ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.IME,  handle:disableFunction},     //IME
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.DEL,  handle:_h(this.onCtrlDel)},        //Ctrl+Delete
                {ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onCtrlBackspace)},       //Ctrl+Backspace
                {ctrlKey:false, altKey:false, shiftKey:true, keyCode: this.KEY.DEL,  handle:_h(this.onShiftDel)},        //Shift+Delete
                {ctrlKey:false, altKey:false, shiftKey:true, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onShiftBackspace)},       //Shift+Backspace
                {ctrlKey:true, altKey:false, shiftKey:true, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onCtrlShiftBackspace)},       //Ctrl+Shift+Backspace
                {ctrlKey:true, altKey:true, shiftKey:false, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onCtrlAltBackspace)},       //Ctrl+Alt+Backspace
                {ctrlKey:false, altKey:true, shiftKey:false, keyCode: this.KEY.BACKSPACE,  handle:_h(this.onAltBackspace)}       //Alt+Backspace
            ]);
            if (SAS.Browser.mac()) {
                this.cmdMgr.regCommands([
                    {ctrlKey:false, altKey:true, shiftKey:false, keyCode: this.KEY.LEFT,  handle:_h(this.onAltLeft)},
                    {ctrlKey:false, altKey:true, shiftKey:false, keyCode: this.KEY.RIGHT,  handle:_h(this.onAltRight)}
                ]);//defect S0890541
            }
            // FIXID S0889394: Strange behavior by pressing "ctrl+shift" key in Chrome21 in CE.
            // The reason is that "ctrl+shift" brings up the input event of textarea when Arabic input method(s) being enabled, 
            // but the data in texarea has not be changed.
            // preventDefault() and stopPropagation() cannot stop the input event, 
            // so add action type 'ignore' that will be handled. 
            if (SAS.Browser.chrome()) {
                this.cmdMgr.regCommands([{
                    ctrlKey: true, altKey: false, shiftKey: true, keyCode: this.KEY.SHIFT, handle: _h(this.onCtrlShift)
                }]);
            }
            /*
             * FIXID: 0888235. Shift+END is not supposed to select the last return character.
             * Author: Boyce.
             */
            if (SAS.Browser.msie()) {
                this.cmdMgr.regCommands([
                    {ctrlKey:false, altKey:false, shiftKey:true, keyCode: this.KEY.END,  handle:_h(this.onShiftEnd)},
                    {ctrlKey:false, altKey:false, shiftKey:false, keyCode: this.KEY.ESC,  handle:_h(this.onEsc)}
                ]);
            }
            
            if (SAS.Browser.msafari()) {
                /*
                 * CapsLock is considered as Control in iPad. So CapsLock+W is considered as Ctrl+W and close the current dojo tab.
                 * But we need 'W' instead. So register this shortcut key to stop propagation inside regCommands.
                */
                this.cmdMgr.regCommands([
                    { ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.W,  handle: _h(this.onCtrlW )},
                    { ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.M,  handle: _h(this.onMetaM), forced: true },
                    { ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.J,  handle: _h(this.onMetaJ), forced: true },
                    { ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.H,  handle: _h(this.onMetaH), forced: true },
                    { ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.TAB,  handle: _h(this.onCtrlTab), forced: true},
                    { ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.ENTER,  handle: _h(this.onCtrlEnter), forced: true},
                    { ctrlKey:true, altKey:false, shiftKey:false, keyCode: this.KEY.BACKSPACE,  handle: _h(this.onCtrlBackspace), forced: true}]);
            }
            
            if(SAS.Browser.msafari() || SAS.Browser.mac()) {
                this.cmdMgr.regCommands([
                    {ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.SLASH, handle: _h(this.onMetaSlash)},
                    {ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.Z, handle: _h(this.onMetaZ)},
                    {ctrlKey:false, altKey:false, shiftKey:true, metaKey: true, keyCode: this.KEY.Z, handle: _h(this.onMetaShiftZ)},
                    {ctrlKey:false, altKey:false, shiftKey:true, metaKey: true, keyCode: this.KEY.U, handle: _h(this.onMetaShiftU)},
                    //{ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.U, handle: _h(this.onMetaU)},
                    {ctrlKey:false, altKey:false, shiftKey:false, metaKey: true, keyCode: this.KEY.Y, handle: _h(this.onMetaY)}
                ]);
            }else{
                this.cmdMgr.regCommands([
                    {ctrlKey:true, altKey:false, shiftKey:false, metaKey: false, keyCode: this.KEY.SLASH, handle: _h(this.onCtrlSlash)}]);
            }
        },
        regListeners: function (events, callbacks) {
            /*var self = this;
            var i, callback, Listener = function(name, callback){
                return function(e){
                    return callback.apply(self,arguments);
                };
            };*/
            for (var i = 0; i < events.length; i++) {
                //callback = SAS.Browser.msie() ? new Listener(events[i], callbacks[i]) : callbacks[i];
                //callback = new Listener(events[i], callbacks[i]);
                this.listeners[events[i]] = on(this.textarea, events[i], callbacks[i].bind(this));
                //this.listeners[events[i]] = this.on(this.textarea, events[i], callbacks[i]);
            }
        },
        onKeyDown: function(e) {
            if (this.control.model.readOnly() && (SAS.Browser.chrome() || SAS.Browser.safari())) {
                if (e.keyCode === this.KEY.DOWN || e.keyCode === this.KEY.UP || e.keyCode === this.KEY.LEFT || e.keyCode === this.KEY.RIGHT
                    || e.keyCode === this.KEY.HOME || e.keyCode === this.KEY.END || e.keyCode === this.KEY.PAGE_UP || e.keyCode === this.KEY.PAGE_DOWN) {
                    this.textarea.readOnly = false;
                } else if (!this.textarea.readOnly) {
                    this.textarea.readOnly = true;
                }
            }
            if (!this[TransDriver.getHandleName(e)])
                this.control.dispatch('onKeyDown', e);
            return this.cmdMgr.execCommands(e);
        },
        onCompStart: function(e) {
            if (this.control.model.readOnly() && (SAS.Browser.chrome() || SAS.Browser.safari())) {
                if (!this.textarea.readOnly) {
                    this.textarea.readOnly = true;
                }
            }
            this.control.dispatch('onCompStart',e);
        },
        onKeyDownDoc: function(e) {
            if ((e.ctrlKey || e.metaKey) && e.keyCode === this.KEY.Z) {
                this.control.dispatch('onCtrlZDoc',e);
            }
        },
        onInsert: function(e) {
            return TransDriver.stop(e);
        },
        onCtrlD: function(e) {
            if(SAS.Browser.mac()) {
                return TransDriver.stop(e);
            }
        },
        onShiftDel: function(e) {
            return TransDriver.stop(e);
        },
        onShiftBackspace: function(e) {
            return TransDriver.stop(e);
        },
        onCtrlShiftBackspace: function(e) {
            return TransDriver.stop(e);
        },
        onCtrlAltBackspace: function(e) {
            return TransDriver.stop(e);
        },
        onAltBackspace: function(e) {
            //FIXID S0894029:  The text will be in mess after pressing "alt + backspace" key on keyboard in IE
            // We cannot disable Alt+Backspace in IE, but this strange code can work. orz.
            if (SAS.Browser.msie()) {
                //var uglyCode4IE = 0;
                //while (uglyCode4IE < 50000000) {uglyCode4IE++;}
                alert('\'Alt+Backspace\' is not supported!');
            }
            return TransDriver.stop(e);
        },
        onDragOver: function(e) {
            return TransDriver.stop(e);
        },
        onDrop: function(e) {
            //TODO: This is a temp implementation, improve in the future. 
            this.control.modelTA._evtMgr.notify("drop",e);
            return this.control.dispatch("onDrop", e);
            //return TransDriver.stop(e);
        }
    };
    TransDriver.KEY_NAMES = {
8:"Backspace",
9:"Tab",
//12:"Clear",
13:"Enter",//"Enter",
//16:"Shift_L",
//17:"Control_L",
//18:"Alt_L",
//19:"Pause",
//20:"Caps_Lock",
27:"Esc",//"Escape ",
//32:"space ",
33:"PageUp",//"Prior",
34:"PageDown",//"Next",
35:"End",
36:"Home",
37:"Left",
38:"Up",
39:"Right",
40:"Down",
//41:"Select",
//42:"Print",
//43:"Execute",
//45:"Insert",
46:"Del",
//47:"Help",
//48://"0 equal braceright",
//49://"1 exclam onesuperior",
//50://"2 quotedbl twosuperior",
//51://"3 section threesuperior",
//52://"4 dollar",
//53://"5 percent",
//54://"6 ampersand",
//55://"7 slash braceleft",
//56://"8 parenleft bracketleft",
//57://"9 parenright bracketright",
65:"A",
66:"B",
67:"C",
68:"D",
69:"E",
70:"F",
71:"G",
72:"H",
73:"I",
74:"J",
75:"K",
76:"L",
77:"M",
78:"N",
79:"O",
80:"P",
81:"Q",
82:"R",
83:"S",
84:"T",
85:"U",
86:"V",
87:"W",
88:"X",
89:"Y",
90:"Z",
191:"Slash",
229:"Ime"
    };
    TransDriver.getHandleName = function(e) {
        var name = "on", keyName = TransDriver.KEY_NAMES[e.keyCode];
        if (e.metaKey){name += 'Meta';}
        if (e.ctrlKey){name += 'Ctrl';}
        if (e.altKey){name += 'Alt';}
        if (e.shiftKey){name += 'Shift';}
        if (keyName) name += keyName;
        return name;
    };
    TransDriver.HANDLERS = [
        "onInput","onKeyDown","onKeyUp","onKeyPress","onSelect",//1
        "onCompStart","onCompEnd",//2
        "onTouchStart","onTouchEnd",//3
        "onMouseDown","onMouseMove","onMouseUp",//4
        "onDragOver","onDrop",//5
        "onFocusIn","onFocusOut",//6
        "onPaste","onCut","onCopy", //7
        "onContextMenu",//8
        "onScroll", "onDblClick", //9
        "onTab","onShiftTab","onBackSpace","onDel","onEnter","onInsert",//commands
        "onCtrlD","onPageUp","onPageDown",
        "onEnd","onCtrlEnd","onCtrlShiftEnd","onHome","onCtrlHome","onCtrlShiftHome",
        "onLeft","onCtrlLeft","onUp","onCtrlUp","onRight","onCtrlRight","onDown", "onCtrlDown",
        "onCtrlShiftLeft","onCtrlShiftRight","onCtrlZ","onCtrlShiftU","onCtrlU","onCtrlL","onCtrlShiftZ",
        "onCtrlY","onCtrlDel","onCtrlBackspace","onShiftDel","onShiftBackspace",
        "onCtrlShiftBackspace","onCtrlAltBackspace","onAltBackspace",
        "onAltLeft","onAltRight","onCtrlShift","onShiftEnd","onEsc",
        "onCtrlW","onMetaM","onMetaJ","onMetaH",
        "onCtrlTab","onCtrlEnter","onCtrlBackspace","onMetaSlash","onMetaZ","onMetaShiftZ","onMetaY",
        "onCtrlSlash","onMetaShiftU"
    ];
    (function initPrototype () {
        TransDriver.HANDLERS.forEach(function(item){
            if(TransDriver.prototype[item] === undefined){
                TransDriver.prototype[item] = function(e){
                    //return this.dispatch(item, e)
                    return this.control.dispatch(item, e);
					
                };
            }
        });
    })();
    return TransDriver;
});