define(["dojo/_base/declare", "dojo/on", "dojo/_base/lang", "dojo/has", "dojo/dom-geometry", "./CoreRange", "./TransBase", "./Clipboard"],
    function(declare, on, lang, has, domGeo, CoreRange, TransBase, Clipboard) {

var BasicTrans = declare("", [TransBase], {
    constructor: function() {
        //this.allowedEvents = {
        //};
        this.name = "basic";
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE &&
            e.type === 'input') {
            var oldStart = this.preContext.selection.start,
                newCursor = new CoreRange(this.textarea.selectionStart, this.textarea.selectionEnd),
                text = this.modelTA.getSelectedText(oldStart, newCursor.end),
                newRange = new CoreRange(oldStart, newCursor.end);
            this.modelTA.insertTextToModel(text, this.preContext.selection, 'regular');
            this.control.undoMgr().add(text, newRange, newCursor,
                this.preContext.selectedText, this.preContext.selection, this.preContext.selection
            );
            /* NOTE: We cannot use 'keyup' event to mark the end of the transaction,
             * or user cannot input continuously!
             */ 
            this.end(e);
        }
    }    
});

var BasicTransSafari = declare("", [TransBase], {
    constructor: function() {
        this.name = "BasicTransSafari"
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE &&
            e.type === "input") {
            var oldStart = this.preContext.selection.start,
                newCursor = new CoreRange(oldStart+1, oldStart+1),
                text = this.modelTA.getSelectedText(oldStart, newCursor.end),
                newRange = new CoreRange(oldStart, newCursor.end);
            this.modelTA.insertTextToModel(text, this.preContext.selection, "regular");
            this.control.undoMgr().add(text, newRange, newCursor,
                this.preContext.selectedText, this.preContext.selection, this.preContext.selection
            );
            this.end(e);
        }
    }
});

var BasicTransiPad = declare("", [BasicTrans], {
    constructor: function() {
        //this.allowedEvents = {
        //};
        this.name = "BasicTransiPad";
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE &&
            e.type === 'input') {
            var oldStart = this.preContext.selection.start,
                newCursor = new CoreRange(this.preContext.selection.start+1, this.preContext.selection.start+1),
                newRange = new CoreRange(oldStart, newCursor.end),
                text = this.modelTA.getSelectedText(oldStart, newCursor.end);
            if (this.preContext.length - this.textarea.value.length === 1) {
                // For Korean IME on iPad
                var oldRange = new CoreRange(oldStart - 1, oldStart),
                    oldText = this.preContext.lastChar;
                newCursor = new CoreRange(oldStart - 1, oldStart - 1);
                newRange = newCursor;
                newText = "";
                --this.preContext.length;
                this.state = TransBase.TS.COMP_END;
                this.preContext.selection = newCursor;
                //this.core.modifyText(newText, newRange, newCursor, oldText, oldRange, this.preContext.selection);
                this.modelTA.insertTextToModel(newText, oldRange, 'regular');
                this.control.undoMgr().add(newText, newRange, newCursor,
                    oldText, oldRange, newCursor
                );
                return;
            }
            this.modelTA.insertTextToModel(text, this.preContext.selection, 'regular');
            this.control.undoMgr().add(text, newRange, newCursor,
                this.preContext.selectedText, this.preContext.selection, this.preContext.selection
            );
            /*
             * NOTE: We cannot use 'keyup' event to mark the end of the transaction,
             * or user cannot input continuously!
             */
            this.end(e);
        }
        if (this.state === TransBase.TS.COMP_END &&
            e.type === 'input') {
            var step, start;

            step = this.textarea.value.length - this.preContext.length;
            step += this.preContext.selection.end - this.preContext.selection.start;
            start = this.preContext.selection.start;

            if (step) {
                var newCursor = new CoreRange(start + step, start + step),
                    oldText = this.preContext.selectedText,
                    oldRange = this.preContext.selection,
                    newText = this.modelTA.getSelectedText(start, newCursor.end),
                    newRange = new CoreRange(start, start + step);

                //this.core.modifyText(newText, newRange, newCursor, oldText, oldRange, this.preContext.selection);
                this.modelTA.insertTextToModel(newText, oldRange, 'regular');
                this.control.undoMgr().add(newText, newRange, newCursor,
                    oldText, oldRange, this.preContext.selection
                );
                this.end(e);
            }
        }
    }
});

var UndoRedoTransMsafari = declare("",[TransBase],{
    constructor: function() {
        this.name = "undoredo-msafari";
    },
    exec: function(e) {
        var newCursor = this.modelTA.getSelectionRange(),
            oldRange = new CoreRange(0, this.model.length());

        this.modelTA.insertTextToModel(this.textarea.value, oldRange ,'undo');
        this.modelTA.setSelectionRange(newCursor);
        this.end(e);
    }
});

var ImeTransBase= declare("", [TransBase], {
    constructor: function() {
        this.name = "ImeTrans";
        this.allowedEvents = {
            'compositionend':true,
            'keypress': true,
            'keydown': true,
            'keyup': true,
            'input': true
        };
    },
    exec: function(e) {
        switch(this.state){
            case TransBase.TS.IDLE:
                //console.log('    STATE:IDLE');
                if (e.type === 'compositionstart') {//console.log('        evt:compositionstart');
                    this.onCompStart(e);
                }
            break;
            case TransBase.TS.IN_COMP://console.log('    STATE:IN_COMP');
                switch(e.type) {
                    case 'compositionend'://  console.log('        evt:compositionend');
                        this.onCompEndInComp(e);
                    break;
                    case 'keypress'://console.log('        evt:keypress');
                        this.onKeyPressInComp(e);
                    break;
                    case 'input'://console.log('        evt:input');
                        this.onInputInComp(e);
                    break;
                    case 'keydown'://console.log('        evt:keydown');
                        this.onKeyDownInComp(e);
                    break;
                    case 'keyup'://console.log('        evt:keyup');
                        this.onKeyUpInComp(e);
                    break;
                }
            break;
            case TransBase.TS.COMP_END: //console.log('    STATE:COMP_END');
                switch(e.type) {
                    case 'keydown':
                        this.onKeyDownCompEnd(e);
                    break;
                    case 'input'://console.log('        evt:input');
                        this.onInputCompEnd(e);
                    break;
                }
            break;
        }
    },
    onCompStart: function(e) {
        this.state = TransBase.TS.IN_COMP; //1
        this.modelTA.publish('onCompStart');
        this.timerId = null;
    },
    onCompEndInComp: function(e) {
        this.state = TransBase.TS.COMP_END;
        this.modelTA.publish('onCompEnd');
    },
    onKeyDownInComp: function(e) {
        if (e.ctrlKey && e.keyCode === this.KEY.A) {
            this.stop(e);
        }
    },
    onKeyPressInComp: function(e) {
        if (e.ctrlKey && (e.keyCode === this.KEY.END || e.keyCode === this.KEY.HOME)) {
            this.stop(e);
        }
        this.modelTA.updateSelectionCache();//TODO:
        this.preContext.selection = this.modelTA.getSelectionRange();
        this.preContext.selectedText = this.modelTA.getSelectedText();
    },
    onKeyUpInComp: function(e) {
        //blank
    },
    onInputInComp: function(e) {
        var step, start;
        
        step = this.textarea.value.length - this.preContext.length;
        step += this.preContext.selection.end - this.preContext.selection.start;
        start = this.preContext.selection.start;
        
        this.modelTA.publish('onCompUpdate', {text: this.modelTA.getSelectedText(start, start + step), pos: null});
    },
    onKeyDownCompEnd: function(e) {
        if (e.keyCode === this.KEY.DEL) {
            var end = this.modelTA.getSelectionRange().end;
            if (end < this.textarea.value.length - 1) {
                this.preContext.lastChar = this.modelTA.getSelectedText(end, end+1);
                this._delete = true;
            }
        }
    },
    onInputCompEnd: function(e) {
        var step = this.textarea.value.length - this.preContext.length;
        step += this.preContext.selection.end - this.preContext.selection.start;

        var self = this;
        function _insertTextThenEnd() {
            var step = self.textarea.value.length - self.preContext.length;
            step += self.preContext.selection.end - self.preContext.selection.start;

            if (step < 0) {
                if (self._delete) {
                    self.preContext.selection.end++;
                    self.preContext.selectedText += self.preContext.lastChar;
                } else { // backspace
                    self.preContext.selection.start--;
                    self.preContext.selectedText = self.preContext.lastChar + self.preContext.selectedText;
                }
                step = 0;
            }

            var start = self.preContext.selection.start,
                newCursor = new CoreRange(start + step, start + step),
                oldText = self.preContext.selectedText,
                oldRange = self.preContext.selection,
                newText = self.modelTA.getSelectedText(start, newCursor.end),
                newRange = new CoreRange(start, start + step);

            //this.modelTA.modifyText(newText, newRange, newCursor, oldText, oldRange, this.preContext.selection);
            self.control.undoMgr().add(newText, newRange, newCursor, oldText, oldRange, self.preContext.selection);
            self.modelTA.insertTextToModel(newText, oldRange, 'regular');
            self.end(e);
        }

        if (this.timerId) {
            window.clearTimeout(this.timerId);
            this.timerId = null;
        }
        if (step || !SAS.Browser.chrome()) {
            _insertTextThenEnd();
        } else { // For Korean on Chrome
            this.timerId = window.setTimeout(_insertTextThenEnd, 100);
        }
    }
});

var ImeTrans = declare("", [ImeTransBase], {
});

var ImeTransChrome = declare("", [ImeTransBase], {
    end: function() {
        this.endTimestamp = new Date().getTime();
        this.inherited(arguments);
    },
    onKeyDownInComp: function(e) {
        if (e.ctrlKey && (e.keyCode === this.KEY.Z || e.keyCode === this.KEY.Y)) {
            this.stop(e);
        } else {
            this.inherited(arguments);
        }
    }
});

var ImeTransSafari = declare("", [ImeTransChrome], {
    onKeyDownInComp: function(e) {
        if ((/*e.ctrlKey || */e.metaKey) && (e.keyCode === this.KEY.Z || e.keyCode === this.KEY.Y || e.keyCode === this.KEY.A)) {
            this.stop(e);
        } else {
            this.inherited(arguments);
        }
    }
});

var ImeTransIE10= declare("", [ImeTransBase], {
    constructor: function(config) {
        this.name = "ImeTransIE10";
        this.preContext = config.preContext;
    }
});

var ImeTransIE9= declare("", [ImeTransIE10], {
    constructor: function() {
        this.name = "ImeTransIE9";
        this.allowedEvents = {
            'compositionend':true,
            'keypress': true,
            'keydown': true,
            'keyup': true,
            'input': true,
            'mousedown': true
        };
    },
    exec: function(e) {
        if (e.type=='mousedown') {
            //this.stop(e);
            if (this.deferredInput == null)
                this.onCompEndInComp();
            return;
        }
        this.inherited(arguments);
    },
    onCompEndInComp: function() {
        var self = this;
        if (this.deferredInput) {
            window.clearTimeout(this.deferredInput);
        }
        this.deferredInput = window.setTimeout(function() {
            self.state = TransBase.TS.COMP_END;
            self.exec({type: "input"});
            self.deferredInput = null;
        }, 0);
        this.modelTA.publish('onCompEnd');
    },
    onKeyPressInComp: function(e) {}
});

var ImeTransIE11= declare("", [ImeTransBase], {
    constructor: function(config) {
        this.name = "ImeTransIE11";
        this.deferredInput = null;
        this.korean = false; 
        this.allowedEvents = {
            'compositionstart':true,
            'compositionend':true,
            'keypress': true,
            'keydown': true,
            'keyup': true,
            'input': true,
            'select': true,
            'blur': true
        };
        this.preContext = config.preContext;
    },
    exec: function(e) {
        switch(this.state){
            case TransBase.TS.IDLE:
                //console.log('    STATE:IDLE');
                if (e.type === 'compositionstart') {//console.log('        evt:compositionstart');
                    this.onCompStart(e);
                }
                break;
            case TransBase.TS.IN_COMP://console.log('    STATE:IN_COMP');
                switch(e.type) {
                    case 'compositionstart':  //console.log('        evt:compositionstart');
                        this.onCompStart(e);
                        break;
                    case 'compositionend':  //console.log('        evt:compositionend');
                        this.onCompEndInComp(e);
                        break;
                    case 'keypress'://console.log('        evt:keypress');
                        this.onKeyPressInComp(e);
                        break;
                    case 'input'://console.log('        evt:input');
                        this.onInputInComp(e);
                        break;
                    case 'keydown'://console.log('        evt:keydown');
                        this.onKeyDownInComp(e);
                        break;
                    case 'keyup'://console.log('        evt:keyup');
                        this.onKeyUpInComp(e);
                        break;
                    case 'blur'://console.log('        evt:focusout');
                        this.onFocusOutInComp(e);
                        break;
                    case 'select'://console.log('        evt:select');
                        this.modelTA.updateSelectionCache();
                        break;
                }
                break;
            case TransBase.TS.COMP_END: //console.log('    STATE:COMP_END');
                switch(e.type) {
                    case 'input'://console.log('        evt:input');
                        this.onInputCompEnd(e);
                        break;
                }
                break;
        }
    },
    onCompStart: function(e) {
        if (this.deferredInput) {
            window.clearTimeout(this.deferredInput);
            this.deferredInput = null;
        } else if (this.preContext.length !== this.textarea.value.length && /ko/.test(e.locale)) {
            // Korean IME
            if (this.preContext.length > this.textarea.value.length) {
                this.preContext.selection.end = this.preContext.selection.start 
                + this.preContext.length - this.textarea.value.length + 1;
            } else {
                this.korean = true;
            }
        }
        this.inherited(arguments);
    },
    onCompEndInComp: function(e) {
        var self = this;
        this.deferredInput = window.setTimeout(function() {
            self.state = TransBase.TS.COMP_END;
            self.exec({type: "input"});
            self.deferredInput = null;
        }, 100);
        this.modelTA.publish('onCompEnd');
    },
    onInputInComp: function(e) {
        var step, start, pos;
        if (this.korean) {
            pos = this.modelTA.getSelectionRange().start;
            this.preContext.selection.start = pos;
            this.preContext.selection.end = pos;
            this.korean = false;
        }
        //this.inherited(arguments);
        
        step = this.textarea.value.length - this.preContext.length;
        step += this.preContext.selection.end - this.preContext.selection.start;
        start = this.preContext.selection.start;
        pos = this.model.toTextRange(this.preContext.selection).start;
        this.modelTA.publish('onCompUpdate', {text: this.modelTA.getSelectedText(start, start + step), pos: pos});
    },
    onFocusOutInComp: function() {
        if (this.deferredInput === null) {
            this.onCompEndInComp();
        }
    },
    onKeyDownInComp: function(e) {
        // S1160638
        if (e.keyCode === this.KEY.RIGHT || e.keyCode === this.KEY.LEFT || e.keyCode === this.KEY.UP || e.keyCode === this.KEY.DOWN) {
            this.stop(e);
        }
        if (this.deferredInput) {
            this.onKeyDownCompEnd(e);
        }
    },
    onKeyPressInComp: function(e) {}
});

var ImeTransiPad= declare("", [ImeTransBase], {
    constructor: function() {
        this.name = "ImeTransiPad";
    },
    onKeyPressInComp: function() {}
});

var SetCursorTrans = declare("", [TransBase], {
    constructor: function() {
        this.needEnd = false;
        this.allowedEvents = {
            'keypress':true,
            'keyup':true
        };
        this.name = "setcursor";
    },
    exec: function(e) {
        switch(e.type){
        case 'keydown': {
            var selection, self = this;
            selection = this.modelTA.getSelectionRange();
            /*
             * FIXID: S0893515.
             * Author: Boyce
             * Description: Move the cursor to the start/end of the selected range if pressing up/down when selected.
             */
            //TODO: Combine selection and onCursorMoved() to moveCursor /Boyce
            //Special situations for direction arrow keys.
            if (!this.modelTA.noSelected()) {
                if (e.keyCode === this.KEY.UP || e.keyCode === this.KEY.LEFT) {
                    this.stop(e);
                    this.modelTA.setCaretPosition(selection.start);
                    this.end(e);
                    return false;
                } else if (e.keyCode === this.KEY.DOWN || e.keyCode === this.KEY.RIGHT ) {
                    this.stop(e);
                    this.modelTA.setCaretPosition(selection.end);
                    this.end(e);
                    return false;
                } else if (e.keyCode === 0) {
                    if (this.preContext.selection.start === selection.start && this.preContext.selection.end !== selection.end) {
                        this.modelTA.setCaretPosition(selection.end);
                    } else if (this.preContext.selection.end === selection.end && this.preContext.selection.start !== selection.start) {
                        this.modelTA.setCaretPosition(selection.start);
                    }
                }
            }
            // prevent cursor move up at the first line and move down at the last line
            var preTextPos = this.model.toTextRange(selection);
            /**
             * FIXID S0893583: The sample code in tab2 and tab3 could not be selected for the first time in IE9 and IE10
             * Author: Wei
             */
            if (SAS.Browser.msie() && e.type === 'mousedown') {
                this.modelTA.selecting = true;
            }
            if ((e.keyCode === this.KEY.UP && preTextPos.start.line === 0) || (e.keyCode === this.KEY.DOWN && preTextPos.start.line === this.model.getLineCount() - 1)) {
                this.stop(e);
                this.end(e);
                return;
            }
            window.setTimeout(function () {
                //selection = self.modelTA.getSelectionRange();
                //TODO:
                selection = new CoreRange(self.textarea.selectionStart, self.textarea.selectionEnd);
                if (selection.start !== selection.end) {
                    self.end(e);
                    return true;
                }
                self.modelTA.setCaretPosition(selection.start);
                //if (!self.modelTA.isFocused) { self.modelTA.focus();} //(1)
                self.end(e);

            }, 0);
            }
        break;
        case 'keyup':
        break;
        }
    }
});
    var SetCursorToStartTrans = declare("", [SetCursorTrans],{
        exec: function(e) {
            this.modelTA.setCaretPosition(0);
            this.stop(e);
            this.end(e);
        }
    });
    var SetCursorToEndTrans = declare("", [SetCursorTrans],{
        exec: function(e) {
            this.modelTA.setCaretPosition(this.modelTA.getLength());
            this.stop(e);
            this.end(e);
        }
    });
    var SetCursorByMouseTrans = declare("", [SetCursorTrans], {
        constructor: function() {
            this.allowedEvents = {
                'mousemove':true,
                'mouseup':true,
                'drop': true
            };
            this.mouseUpListener = on(document, "mouseup", lang.hitch(this, this.onMouseUp));
        },
        onMouseUp: function(e) {
            var self = this;
            this.mouseUpListener.remove();
            //if (this.modelTA.noSelected() || e.button !== 0) {
            if (this.textarea.selectionStart === this.textarea.selectionEnd || e.button !== 0) {
                if (e.button === 0) {
                    self.modelTA.setCaretPosition(this.textarea.selectionStart);
                }
                this.end(e);
                return;
            }
            this.modelTA.setSelectionRange(this.textarea.selectionStart, this.textarea.selectionEnd);
            window.setTimeout(function(){
                var pos, offset;
                if (SAS.Browser.msie()) {
                    pos = domGeo.position(self.textarea, true);
                    offset = self.view.getTextOffset(
                            e.pageX - pos.x + self.textarea.scrollLeft,
                            e.pageY - pos.y + self.textarea.scrollTop);
                }
                if (self.preContext.selection.start !== self.preContext.selection.end &&
                    e.pageX === self._pageX && e.pageY === self._pageY) { // user clicks the selected text
                    if (SAS.Browser.msie()) self.modelTA.setCaretPosition(offset);
                    else self.modelTA.setSelectionRange(self.textarea.selectionStart, self.textarea.selectionEnd);
                }
                self.end(e);
            },0);
        },
        onMouseDown: function(e) {
            this._pageX = e.pageX;
            this._pageY = e.pageY;
        },
        exec: function(e) {
            if (e.type === 'mousedown') {
                this.onMouseDown(e);
            } else if (e.type === 'mouseup') {
                this.onMouseUp(e);
            } else if (e.type === 'drop') {
                this.stop(e);
                this.end();
            }
        }
    });
    //FIXID S1118940
	
    var SetCursorByMouseTransIE = declare("", [SetCursorByMouseTrans], {
        constructor: function() {
            this.mouseUpListener.remove();
            if (has("MSPointer")){
                if (has("pointer-events")){ // IE11
                    this.mouseUpListener = on(document, "pointerup", lang.hitch(this, this.onMouseUp));		
                } else { // IE10
                    this.mouseUpListener = on(document, "MSPointerUp", lang.hitch(this, this.onMouseUp));
                }
            }
            this.allowedEvents = {
                'MSPointerMove':true,
                'pointermove':true,
                'MSPointerUp':true,
                'pointerup':true,
                'drop': true
            };

        },
        exec: function(e) {
            if (e.type === 'pointerdown' || e.type === 'MSPointerDown') {
                this.onMouseDown(e);
            } else if (e.type === 'pointerup' || e.type === 'MSPointerUp') {
                this.onMouseUp(e);
            }
            this.inherited(arguments);
        }
    });
    var SetCursorByMouseTransChrome = declare("", [SetCursorByMouseTrans], {
        constructor: function() {
            this.allowedEvents = {
                'mousemove':true,
                'keydown':true,  //we cannot disable IME through not including keydown and compositionstart
                'compositionstart':true,
                'mouseup':true,
                'drop': true
            };
            this.readOnly = this.modelTA.readOnly();
            this.readOnlyEntered = false;
            this.caretPosition = 0;
            this.shiftLMB = false;
            //this.enterReadOnly();
        },
        enterReadOnly: function() {
            if (!this.readOnly && !this.readOnlyEntered) {
                this.modelTA.readOnly(true);
                this.caretPosition = this.textarea.selectionStart;
                if (!this.shiftLMB) {
                    this.modelTA.setCaretPosition(this.caretPosition);
                }
                this.readOnlyEntered = true;
            }
        },
        exec: function(e) {
            if ((e.type === 'keydown' &&
                    (e.keyCode === 229 ||
                        e.keyCode === this.KEY.TAB ||
                     e.keyCode === this.KEY.BACKSPACE ||
                     e.keyCode === this.KEY.DEL))
                || e.type === 'compositionstart') {
                this.enterReadOnly();
                //if (e.keyCode === this.KEY.BACKSPACE || e.keyCode === this.KEY.TAB) {
                    this.stop(e);  //we need to stop all.
                //}
            } else if (e.type === 'mousedown' && e.shiftKey) {
                this.shiftLMB = true;
                this.enterReadOnly();
            }
            //NOTE: When using IME in Chrome, it will bring up a messy issue to input and click quickly and continiously (e.g. press 'a') at the same time.
            var lastTrans = this.control.transControl().lastTrans;
            if (e.type === 'mousedown' && lastTrans && lastTrans.name === 'ImeTrans') {//NOTE: 'ImeTrans' is the trans name for Chrome.
                var interval = (new Date()).getTime() - lastTrans.endTimestamp;
                if (interval < 200) {
                    this.stop(e);
                    this.end();
                }
            }
            //console.log('exec:'+ e.type + ',selection');
            this.inherited(arguments);
        },
        exitReadOnly: function() {
            if (!this.readOnly && this.readOnlyEntered) {
                this.modelTA.readOnly(false);
                if (!this.shiftLMB) {
                    this.modelTA.setCaretPosition(this.caretPosition);
                }
                this.readOnlyEntered = false;
            }
        },
        end: function(e) {
            var self = this;
            self.inherited(arguments);
            self.exitReadOnly();
        }
    });
    var SetCursorByMouseTransFF = declare("",[SetCursorByMouseTransChrome],{});
    var SetCursorByMouseTransSafari = declare("",[SetCursorByMouseTransChrome],{});

var ScrollByMouseTransChrome = declare("",[TransBase],{
    constructor: function() {
        this.allowedEvents = {
            'mousemove':true,
            'mouseup':true
        };
        this.name = "scrollbymouse";
        this.readOnly = this.modelTA.readOnly();
        this.mouseUpListener = on(document, "mouseup", lang.hitch(this, this.onMouseUp));
        this.enterReadOnly();
    },
    onMouseUp: function(e) {
        this.exec(e);
    },
    exec: function(e) {
        if (e.type === 'mouseup') {
            //var self = this;
            this.mouseUpListener.remove();
            //this.exitReadOnly();
            this.view.blur();
            this.delay(function() {
                this.end(e);
                this.exitReadOnly();
                this.view.focus();
                this.modelTA.setCaretPosition(this.textarea.selectionStart);
            },100);
        }
    },
    enterReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(true);
        }
    },
    exitReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(false);
        }
    }
});


var HomeTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "home";
    },
    exec: function(e) {
        if(e.type !== 'keydown') {
            //preSelection = this.modelTA.getSelectionRange();
            var preStart = this.preContext.selection.start,
                allText = this.modelTA.getAllText(),
                start = allText.lastIndexOf('\n', preStart - 1) + 1,
                end = allText.indexOf('\n', preStart),
                target = allText.slice(start, end).search(/[^\t\s]/);
            if (start + target === preStart || target === -1) {
                this.modelTA.setCaretPosition(start);
            } else {
                this.modelTA.setCaretPosition(start + target);
            }
            this.stop(e);
            this.end();
            return false;
        }
    }
});

var SelectTrans = declare("", [TransBase], {
    constructor: function(){
        this.allowedEvents = {
            "mouseup":1
        };
        this.name = "select";
    },
    exec: function(e) {
        //if (e.type === "mouseup") {
            this.end(e);
          //  return true;
        //}
        //return true;
    }
});
var SelectTheTextBeforeCaretTrans = declare("",[TransBase],{
    constructor: function(){
        this.name = "selectthetextbeforecaret";
    },
    exec: function(e){
        this.modelTA.setSelectionRange(0,this.preContext.selection.end);
        this.view.scrollTo(0,0);
        this.stop(e);
        this.end(e);
    }
});
var SelectTheTextAfterCaretTrans = declare("",[TransBase],{
    constructor: function(){
        this.name = "selectthetextaftercaret";
    },
    exec: function(e){
        this.modelTA.setSelectionRange(this.preContext.selection.end, this.modelTA.getLength());
        this.view.scrollTo(this.model.getLineCount()-1,0);
        this.stop(e);
        this.end(e);
    }
});

var TabTrans = declare("", [TransBase], {
    constructor: function(){
        this.name = "tab";
    },
    insertTab: function() {
            var settings = SAS.Settings.getInstance(),
                selection = this.modelTA.getSelectionRange(),
                newText = null, newRange;

            if (selection.start === selection.end) {
                newText = settings.tabAsSpaces() ? this.modelTA._calculateTab(selection.start) : '\t';
                newRange = new CoreRange(selection.start, selection.start + newText.length);
                this.modelTA.insertText(newText, selection, 'tab');
                this.control.undoMgr().add(newText, newRange, newRange, "", selection, selection);
            } else {
                var range, tab, result, regExp = new RegExp('^()', 'gm'), oldText;
                // TODO: whether there is performance problem.
                range = new CoreRange(this.modelTA.getAllText().substring(0, selection.start).lastIndexOf(this.RETURN_CHAR) + 1, selection.end);
                tab = settings.tabAsSpaces() ? settings.getTabText() : '\t';
                oldText = this.modelTA.getSelectedText(range);
                newText = oldText.replace(regExp, tab);
                result = newText.match(regExp);
                newRange = new CoreRange(range.start, selection.end + result.length * tab.length);
                var selectedRange = new CoreRange(selection.start + tab.length, newRange.end);
                this.modelTA.insertText(newText, range, 'tab');
                this.control.undoMgr().add(newText, newRange, newRange, oldText, range, range);
                this.modelTA.setSelectionRange(selectedRange);
            }
    },
    exec: function(e) {
        if (e.type === 'keydown') {
            this.stop(e);
            this.insertTab();
            this.end(e);
        }
    }
});

var DelTabTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "deltab";
    },
    deleteTabs: function () {
        var allText = this.modelTA.getAllText(),
            selection = this.modelTA.getSelectionRange(),
            range = new CoreRange(allText.substring(0, selection.start).lastIndexOf(this.RETURN_CHAR) + 1, selection.end),
            oldText = this.modelTA.getSelectedText(range),
            newText = null,
            tabSize = SAS.Settings.getInstance().tabSize(),
            regExp = new RegExp('^([\t]| {1,' + tabSize + '})', 'gm'),
            result = oldText.match(regExp),
            beginLineTest = this.modelTA.getSelectedText(range.start, selection.start).match(regExp),
            i = 0, changedCount = 0;
        if (!this.model.readOnly() && !!result) {
            for (i; i < result.length; i++) {
                changedCount += result[i].length;
            }
            newText = oldText.replace(regExp, '');
            var newRange = new CoreRange(range.start, range.start + newText.length);
            this.modelTA.insertText(newText, range, 'deletetab');
            this.control.undoMgr().add(newText, newRange, newRange, oldText, range, range);
            this.modelTA.setSelectionRange(selection.start - (beginLineTest?beginLineTest[0].length:0), selection.end-changedCount);
        }
        return false;
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE) {
            this.stop(e);
            this.deleteTabs();
            this.end(e);
            return false;
        }
    }
});

var BackspaceTrans = declare("", [TransBase], {
    constructor: function() {
        this.allowedEvents = {
            'input':true,
            'keypress':true,
            'keyup':true
        };
        this.deletedRange = null;
        this.deletedText = '';
        this.name = "backspace";
    },
    checkValidity: function() {
        var selection = this.preContext.selection;
        if (selection.start === selection.end &&
            selection.end === 0) {
            return false;
        }
        return true;
    },
    saveContext: function(start, end) {
        if (this.modelTA.noSelected()) {
            this.deletedRange = new CoreRange(start, end);
            this.deletedText = this.modelTA.getSelectedText(start, end);
            if (this.deletedText === "") {
                //TODO:
            }
        } else {
            this.deletedRange = new CoreRange(this.preContext.selection);
            this.deletedText = this.modelTA.getSelectedText();
        }
    },
    changeText: function() {
        var oldRange = this.deletedRange,
            newRange = new CoreRange(oldRange.start, oldRange.start);
        this.modelTA.insertTextToModel("", oldRange, 'backspace');
        this.modelTA.setCaretPosition(oldRange.start);
        this.control.undoMgr().add("",//newText
            newRange, //newRange
            newRange, //newCursor
            this.deletedText, //oldText
            oldRange, //oldRange
            this.preContext.selection //oldCursor
        );
    },
    exec: function(e) {
        var end;
        switch(e.type){
        case 'keydown':
            if (!this.checkValidity()) {
                this.end();
                return;
            }
            end = this.preContext.selection.start;
            this.saveContext(end-1, end);
        break;
        case 'input':
            this.changeText();
            this.end(e);
        break;
        }
    }
});

var BackspaceTransIE9 = declare("", [BackspaceTrans], {
    changeText: function() {
        var oldRange = this.deletedRange,
            newRange = new CoreRange(oldRange.start, oldRange.start);
        this.modelTA.insertText("",oldRange, 'backspace');
        this.modelTA.setCaretPosition(oldRange.start);
        this.control.undoMgr().add("",
            newRange,
            newRange,
            this.deletedText,
            oldRange,
            this.preContext.selection
        );
    },
    exec: function(e) {
        if (e.type === 'keydown') {
            this.stop(e);
            var end = this.preContext.selection.start;
            if (this.checkValidity()) {
                this.saveContext(end - 1, end);
                this.changeText();
            }
            this.end(e);
        }
    }
});

var BackspaceTransSafari = declare("", [BackspaceTransIE9], {

});

var DelTrans = declare("", [BackspaceTrans], {
    constructor: function() {
        this.name = "del";
    },
    checkValidity: function() {
        var selection = this.preContext.selection;
        if (selection.start === selection.end &&
            selection.end === this.modelTA.getLength()) {
            return false;
        }
        return true;
    },
    exec: function(e) { 
        var start;
        switch(e.type){
        case 'keydown':
            if (!this.checkValidity()) {
                this.end(e);
                return;
            }
            start = this.preContext.selection.start;
            this.saveContext(start,start+1);
            if (SAS.Browser.msie() || SAS.Browser.ff()) {
                this.startWatchDog(200);
            }
        break;
        case 'input':
            this.changeText();
            this.end(e);
        break;
        }
    }
});
var DelTransIE9 = declare("", [BackspaceTrans], {
    exec: function(e) {
        var start = this.preContext.selection.start;
        if (start < this.modelTA.getLength()) {
            this.saveContext(start,start+1);
            this.delay(this.changeText);
        }
        this.end(e);
    }
});

var DelTransMSafari = declare("", [BackspaceTransIE9], {
    constructor: function(){    
        this.name = "del";
    },
    exec: function(e) {
        var start = this.preContext.selection.start;
        this.saveContext(start,start+1);
        this.changeText();
        this.stop(e);
        this.end(e);
        return false;
    }
});

var DelWordTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = 'delword';
    },
    deleteWord: function (direction) { //'left' or 'left' on the cursor
        var selection, wordRange = this.modelTA.getSelectionRange(), oldText, newRange = null, oldCursor = null;
        if (this.modelTA.noSelected()) {
            selection = this.modelTA.getSelectionRange();
            if (direction === 'right') {
                wordRange = this.modelTA.getRightWordRange(selection.end);
            } else {
                wordRange = this.modelTA.getLeftWordRange(selection.start);
            }
            if (wordRange) {
                this.modelTA.setSelectionRange(wordRange);
            }
        }
        selection = this.modelTA.getSelectionRange(),
        oldText = this.modelTA.getSelectedText();
        newRange = new CoreRange(selection.start, selection.start);
        oldCursor = new CoreRange(wordRange.end, wordRange.end);
        this.modelTA.insertText("", selection, 'delword');
        this.control.undoMgr().add("",newRange, newRange, oldText, wordRange, oldCursor);
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE ) {
            this.stop(e);
            this.deleteWord((e.keyCode === this.KEY.DEL)? 'right' : 'left');//BACKSPACE -> left
            this.end(e);
        }
    }
});

var ReturnTrans = declare("", [TransBase], {
    constructor: function() {
        this.indent = "";
        this.name = "return";
        /*this.allowedEvents = {
            'input':true,
            'keypress':true,
            'keyup':true
        };*/
    },
    exec: function(e) {
        switch(e.type){
        case 'keydown':
            this.indent = this.model.getIndent(this.model.toTextPosition(this.preContext.selection.start));
            break;
        case 'input':
            if (SAS.Browser.chrome() || SAS.Browser.ff()) {
                this.handleReturn(e);
            } else {
                this.delay(this.handleReturn);
            }
            break;
        }
    },
    handleReturn: function(e) {
        var preStart = this.preContext.selection.start,
            oldCursor = this.preContext.selection,
            newCursor = new CoreRange(preStart + 1, preStart + 1),
            oldText = this.preContext.selectedText,
            oldRange = this.preContext.selection,
            newText = this.RETURN_CHAR + this.indent,
            newRange = new CoreRange(preStart, preStart + newText.length);
            
        if (this.indent.length > 0) {
            // Add indent and restore cursor without impacting preSelection. (preSelection will be changed via getSelectionRange() method)
            newCursor.start = newRange.end;
            newCursor.end = newRange.end;
            /*this.modelTA._setTextToTextarea(this.textarea.value.substring(0, preStart + 1)
                     + this.indent 
                     + this.textarea.value.substring(preStart + 1));//TODO:
            //this.modelTA.restoreSelection(newCursor);
            this.modelTA.setSelectionRange(newCursor);*/
            this.modelTA.setTextToModelTA(this.textarea.value.substring(0, preStart + 1)
                + this.indent
                + this.textarea.value.substring(preStart + 1));
            this.modelTA.restoreSelectionRange(newRange.end,newRange.end);
        }
        //preSelection = newCursor;
        //this.modelTA.modifyText(newText, newRange, newCursor, oldText, oldRange, oldCursor);
        this.modelTA.insertTextToModel(newText, oldRange, 'return');
        this.control.undoMgr().add(newText, newRange, newCursor, oldText, oldRange, oldCursor);
        this.modelTA.setSelectionRange(newRange.end, newRange.end);
        this.end(e);
    }
});

var ReturnTransSafari = declare("", [TransBase], {
    constructor: function() {
        this.name = "return";
    },
    exec: function(e) {
        if (e.type === 'keydown') {
            var range = this.modelTA.getSelectionRange(),
                oldText = this.modelTA.getSelectedText(),
                newText = this.RETURN_CHAR + this.model.getIndent(this.model.toTextPosition(range.start)),
                caretPos = range.start + newText.length,
                newRange = new CoreRange(range.start,caretPos),
                newCursor= new CoreRange(caretPos);
            this.modelTA.insertText(newText, range, 'return');
            this.control.undoMgr().add(newText, newRange, newCursor, oldText, range, range);
            this.end(e);
            this.stop(e);
            return false;
        }        
    }
});

var EscTransIE = declare("",[TransBase],{
    constructor: function() {
        this.name = "esc";
    },
    exec: function(e) {
        this.stop(e);
        this.end();
    }
});

var UndoTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "undo";
        this.allowedEvents = { };// disable all events
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE ) {
            //this.control.undo();
            var unit = this.control.undoMgr().undo();
            if (unit) {
                //this.changeText(unit.oldText, unit.newRange, unit.oldCursor, true);
                this.view.clearFound();/* NOTE: In insertText, insertTextToModelTA will restore the range,
                                        * then insertTextToModel will not find the caret changed.
                                        * This is not very good.
                                        */
                this.modelTA.insertText(unit.oldText, unit.newRange, 'undo');
                this.modelTA.setSelectionRange(unit.oldCursor);
            }
            this.end(e);
            this.stop(e);
        }
    }
});

var UglyUndoTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "undo";
        this.allowedEvents = {};
    },
    exec: function(e) {
        var unit = this.control.undoMgr().undo();
        if (unit) {
            this.view.clearFound();
            this.modelTA.insertTextToModel(unit.oldText, unit.newRange, 'undo');
            this.modelTA.setSelectionRange(unit.oldCursor);
        }
        this.end(e);
    }
});

var DisableDocUndo = declare("", [TransBase], {
    constructor: function() {
        this.name = "disabledocundo";
        this.allowedEvents = {
            "keydown": true,
            "keyup": true,
            "input": true
        };
        this.readOnly = this.modelTA.readOnly();
        this.enterReadOnly();
        this.startWatchDog(500);
    },
    enterReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(true);
        }
    },
    end: function() {
        this.inherited(arguments);
        this.exitReadOnly();
        this.stopWatchDog();
    },
    exitReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(false);
        }
    },
    exec: function(e) {
        if (e.type === 'keyup' || e.type === 'input') {
            this.end();
        }
    }
});

var RedoTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "redo";
        this.allowedEvents = {};//disable all events
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE ) {
            var unit = this.control.undoMgr().redo();
            if (unit) {
                //this.changeText(unit.newText, unit.oldRange, unit.newCursor, true);
                this.modelTA.insertText(unit.newText, unit.oldRange, 'redo');
                this.modelTA.setSelectionRange(unit.newCursor);
            }
            this.end(e);
            this.stop(e);
        }
    }
});
var UppercaseTrans = declare("", [TransBase],{
    constructor: function() {
        this.name = "uppercase";
    },
    exec: function(e) {
        var selectedText = this.modelTA.getSelectedText();
        var selectedRange = this.modelTA.getSelectionRange();
        var newText = selectedText.toUpperCase();
        this.modelTA.insertText(newText,selectedRange,'uppercase');
        this.control.undoMgr().add(newText, selectedRange, selectedRange, selectedText, selectedRange, selectedRange);
        this.modelTA.setSelectionRange(selectedRange);
        this.stop(e);
        this.end(e);
    }
});
var LowercaseTrans = declare("", [TransBase],{
    constructor: function() {
        this.name = "lowercase";
    },
    exec: function(e) {
        var selectedText = this.modelTA.getSelectedText();
        var selectedRange = this.modelTA.getSelectionRange();
        var newText = selectedText.toLowerCase();
        this.modelTA.insertText(newText,selectedRange,'lowercase');
        this.control.undoMgr().add(newText, selectedRange, selectedRange, selectedText, selectedRange, selectedRange);
        this.modelTA.setSelectionRange(selectedRange);
        this.stop(e);
        this.end(e);
    }    
});
var CamelcaseTrans = declare("", [TransBase],{
    constructor: function() {
        this.name = "camelcase";
    },
    exec: function(e) {
        var selectedText = this.modelTA.getSelectedText();
        var selectedRange = this.modelTA.getSelectionRange();
        var newText = selectedText.toLowerCase();
        var reg = /\b(\w)|\s(\w)/g;
        
        newText = newText.replace(reg,function(m){
            return m.toUpperCase();    
        });
        this.modelTA.insertText(newText,selectedRange,'camelcase');
        this.control.undoMgr().add(newText, selectedRange, selectedRange, selectedText, selectedRange, selectedRange);
        this.modelTA.setSelectionRange(selectedRange);
        this.stop(e);
        this.end(e);
    }
});
var CyclecaseTrans = declare("", [TransBase],{
    constructor: function() {
        this.name = "cyclecase";
    },
    exec: function(e) {
        var selectedText = this.modelTA.getSelectedText();
        var lowercaseExisted = /[a-z]/.test(selectedText);
        var uppercaseExisted = /[A-Z]/.test(selectedText);
        var transName = "uppercase";
        
        if (lowercaseExisted && !uppercaseExisted){
            transName = "camelcase";
        } else if (!lowercaseExisted && uppercaseExisted){
            transName = "lowercase";
        }        
        TransFactory.createTransAndExec(transName,this.config);
        this.stop(e);
        this.end(e);
    }
});
var PasteSimTrans = declare("", [TransBase], {
    CRLF_REGEXP: /\r\n|\r/g,
    constructor: function() {
        this.name = 'pastesim';
    },
    exec: function(e) {
        var selectedText = this.modelTA.getSelectedText(),
            selectedRange = this.modelTA.getSelectionRange(),
            data = null,
            newRange = null;
        data = Clipboard.data().replace(this.CRLF_REGEXP,'\n'); //TODO: please remove the duplication, that Controller has this kind of data cleaning
        if (SAS.Settings.getInstance().tabAsSpaces()) {
            data = this.modelTA._translateTab(data, selectedRange.start);
        }
        newRange = new CoreRange(selectedRange.start, selectedRange.start + data.length);
        if (data.length > 0) {
            this.modelTA.insertText(data, selectedRange, 'pastesim');
            this.control.undoMgr().add(data, newRange, newRange, selectedText, selectedRange, selectedRange);
            this.modelTA.setCaretPosition(newRange.end);
        }
        this.end(e);
    }
});
var CutSimTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = 'cutsim';
    },
    exec: function(e) {
        if (!this.modelTA.noSelected()) {
            var selectedText = this.modelTA.getSelectedText(),
                selectedRange = this.modelTA.getSelectionRange(),
                newRange = new CoreRange(selectedRange.start, selectedRange.start);
            Clipboard.data(selectedText);
            this.modelTA.insertText("", selectedRange, 'cutsim');
            this.control.undoMgr().add("", newRange, newRange, selectedText, selectedRange, selectedRange);
            this.modelTA.setCaretPosition(newRange.start);
            this.end(e);
        }
    }
});

var CopyTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "copy";
    },
    exec: function(e) {
        Clipboard.data(this.modelTA.getSelectedText());
        this.end(e);
    }
});

var CopySimTrans = declare("", [CopyTrans], {
    constructor: function() {
        this.name = 'copysim';
    }
});

var CutTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "cut";
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE ) {
            if (e.type === 'cut') {
                if (this.modelTA.noSelected()) {
                    this.end(e);
                }
                if (SAS.Browser.msie() === 9) {
                    this.delay(this.handleDelete);
                } else {
                    this.state = TransBase.TS.INPUT;
                }
            }
        } else if(this.state === TransBase.TS.INPUT) {
            if (e.type === 'input') {
                if (SAS.Browser.msie()) {
                    var self = this;
                    this.delay(function(){
                        self.handleDelete(e); 
                    });
                } else {
                    this.handleDelete(e);
                }
            }
        }
    },
    handleDelete: function(e) {
        var start = this.preContext.selection.start,
            selection = new CoreRange(start, start);
        /*this.modelTA.modifyText("",selection, selection,
            this.preContext.selectedText,
            this.preContext.selection,
            this.preContext.selection
        );*/
        this.modelTA.insertTextToModel("",this.preContext.selection, 'cut');
        this.modelTA.setSelectionRange(selection);
        this.control.undoMgr().add("",selection,selection,
            this.preContext.selectedText,
            this.preContext.selection,
            this.preContext.selection
        );
        this.end(e);
    }
});


var PasteTransBase = declare("", [TransBase], {
    constructor: function(){
        //this.oldCursor = null;
        this.name = "paste";
        this.startWatchDog(5000);
    },
    exec: function(e) {
        //this.oldCursor = this.preContext.selection;
    },
    getClipContent: function() {
        var selection = this.modelTA.getSelectionRange();
        return this.modelTA.getSelectedText(this.preContext.selection.start, selection.end);
    },
    handlePaste: function(e) {
        var selection = this.modelTA.getSelectionRange(),
            pastedText, 
            allText, 
            preStart = this.preContext.selection.start,
            clipContent = this.getClipContent();
        
        pastedText = SAS.Settings.getInstance().tabAsSpaces() 
                   ? this.modelTA._translateTab(clipContent, preStart)
                   : clipContent;
        if (pastedText !== clipContent) {//console.log('change textarea\'s value');
            allText = this.modelTA.getAllText();
            this.modelTA.setTextToModelTA(allText.substring(0, preStart) + pastedText + allText.substring(selection.end)); //TODO:
        }
        var newCursor = new CoreRange(preStart + pastedText.length, preStart + pastedText.length);
        this.modelTA.insertTextToModel(pastedText, this.preContext.selection, 'paste');
        var newRange = new CoreRange(preStart, preStart + pastedText.length);
        this.control.undoMgr().add(pastedText, newRange, newCursor,
            this.preContext.selectedText,
            this.preContext.selection,
            this.preContext.selection
        );
        //this.modelTA.setSelectionRange(newRange);
        this.modelTA.setCaretPosition(newRange.end);
        this.end(e);
    }
});

var PasteTransGeneral = declare("", [PasteTransBase], {
    exec: function(e) {
        if (e.type !== 'input') return false;
        
        this.inherited(arguments);
        this.handlePaste();
    }
});
var PasteTransIE = declare("", [PasteTransBase], {
    constructor: function(){
        this.state = 0;
        this.allowedEvents = {
            'input':true
        };
        this.stopWatchDog();
    },

    exec: function(e) {
        var self = this;
        this.startWatchDog(200,function(){return (self.state === 0);})
        if (e.type !== 'input') return;
        
        this.inherited(arguments);
        this.state++;
        if (this.state >= 2) {
            return;
        }
        this.delay(function(){
            //sync the selection for IE, Defect S0891809. 
            //Or pasted content will not be set to model when a range is selected.
            this.modelTA.updateSelectionCache();
            this.handlePaste();
        });
    }
});
var PasteTransWebKit = declare("", [PasteTransBase], {
    exec: function(e) {
        if (e.type !== 'input') return false;
        
        this.inherited(arguments);
        this.delay(this.handlePaste);
    }
});

var PasteTransMacSafari = declare("",[PasteTransWebKit], {
    getClipContent: function() {
        var clipContent = this.inherited(arguments);
        //if (this.textarea.isFirstPaste === undefined) {
        if (PasteTransMacSafari.isFirstPaste) {
            if (clipContent[clipContent.length-1] === '\n') {
                clipContent = clipContent.substring(0,clipContent.length-1);
            }
            //this.textarea.isFirstPaste = false;
            PasteTransMacSafari.isFirstPaste = false;
        }
        return clipContent;
    }
});
PasteTransMacSafari.isFirstPaste = true;

var PasteTransMobileSafari = declare("",[PasteTransMacSafari],{});//NOTE: use also isFirstPaste

var SasCommentTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "comment";
    },
    comment: function() {
        var selection = this.modelTA.getSelectionRange(),
            allText = this.modelTA.getAllText(),
            start, end, range, text, commented, endWithEnter = false;

        start = selection.start === 0 ? 0 : (allText.lastIndexOf(this.RETURN_CHAR, selection.start - 1) + 1);
        if (selection.start < selection.end && this.modelTA.getSelectedText(selection.end - 1, selection.end) === this.RETURN_CHAR) {
            endWithEnter = true;
        }
        end = allText.indexOf(this.RETURN_CHAR, selection.end - (endWithEnter ? 1 : 0)); // Omit return character.
        if (end < 0) {
            end = allText.length;
        }
        range = new CoreRange(start, end);
        text = this.modelTA.getSelectedText(range);
        commented = (/^\n*\s*\/\*.*(\*\/\s*(\n)*\/\*.*)*.*\*\/\s*\n*$/).test(text);
        //commented = (/^\/\*.*\*\/(\n.*)*$|^(\s)*\/\*.*(\*\/\n\/\*.*)*.*\*\/$/).test(text);
        //commented = (/^\/\*.*\*\/\n+(\n.*)*$|^\/\*.*(\*\/\n\/\*.*)*.*\*\/$/).test(text);
        //console.log(commented)
        var newText = null;
        if (!commented) {
            newText = text.replace(/^(\s*)\/\* ?| ?\*\/(\s*)$/gm, '$1$2');
            newText = newText.replace(/^(.*)$/gm, '/* $1 */');
        } else {
            newText = text.replace(/^(\s*)\/\* ?| ?\*\/(\s*)$/gm, '$1$2');
        }
        this.modelTA.insertText(newText, range, 'comment');
        var newRange = new CoreRange(start, start + newText.length),
            newCursor = new CoreRange(newRange.end, newRange.end);
            
        this.control.undoMgr().add(newText,newRange, newCursor,
            text, range, range
        );
    },
    exec: function(e) {
        this.stop(e);
        this.comment();
        this.end(e);
    }
});
var SelectWordTrans = declare("",[TransBase], {
    constructor: function() {
        this.name = 'selectword';
    },
    exec: function(e){
        var word = this.view.getWordInfoByXY(e.layerX, e.layerY), offset = null;

        if (word) {
            offset = this.model.toTextOffset(word.line, word.column);
            if (offset) {
                this.modelTA.setSelectionRange(offset, offset + word.text.length);
            }
        }
        this.end();
    }
});
var SelectByWordTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "selectbyword";
        this.direction = this.modelTA.getSelectionDirection();
        if (this.direction===undefined) {
            this.direction = this.textarea.selectionDirection;
        }
    },
    setDirection: function(direction) {
        this.modelTA.setSelectionDirection(direction);
        this.direction = direction;
    },
    exec: function(e) {
        if (this.state === TransBase.TS.IDLE ) {
            var range, selection = this.preContext.selection, newSelection;
            if (e.keyCode === this.KEY.LEFT) {
                if (selection.start === selection.end){
                    this.setDirection('forward');
                }

                if (this.direction === 'forward') {
                    range = this.modelTA.getLeftWordRange(selection.start);
                    newSelection = new CoreRange(range.start, selection.end);
                } else if (this.direction === 'backward') {
                    range = this.modelTA.getLeftWordRange(selection.end);
                    newSelection = new CoreRange(selection.start, range.start);
                }
            } else if(e.keyCode === this.KEY.RIGHT) {
                if (selection.start === selection.end) {
                    this.setDirection('backward');
                }
                if (this.direction === 'forward') {
                    range = this.modelTA.getRightWordRange(selection.start);
                    newSelection = new CoreRange(range.end, selection.end);
                } else if(this.direction === 'backward') {
                    range = this.modelTA.getRightWordRange(selection.end);
                    newSelection = new CoreRange(selection.start, range.end);
                }
            }
            this.modelTA.setSelectionRange(newSelection);//TODO: not good
            //if (validTextAreaSelection) { //TODO: This is only for FF6, FF7, the code editor will not support FF6 and FF7
            //this.modelTA.publishCursorMove(newSelection);
            //}
            this.stop(e);
            this.end(e);
        }
    }
});
    var ScrollByDeltaTrans = declare("", [TransBase], {
        constructor: function(){
            this.name = 'scrollbydelta';
        },
        exec: function(e){
            this.stop(e);
            if (e.keyCode === this.KEY.UP) {
                this.view.scrollByDelta(0,-15);
            } else {
                this.view.scrollByDelta(0,15);
            }
            this.end(e);
        }
    });
    var ScrollByPageTrans = declare("", [TransBase], {
        constructor: function(){
            this.name = "scrollbypage";
        },
        scroll: function(direction) {
            var selectedRange = this.modelTA.getSelectionRange(),
                lineHei = this.view.getLineHeight(),
                viewLineCount = Math.round(this.view.getHeight()/lineHei),
                startLine = 0;
            selectedRange = this.model.toTextRange(selectedRange);
            startLine = selectedRange.start.line;
            if (direction === 'backward') {
                startLine = startLine - viewLineCount;
                if (startLine < 0) {
                    startLine = 0;
                }
            } else {
                startLine = startLine + viewLineCount;
                if (startLine > this.model.getLineCount()) {
                    startLine = this.model.getLineCount()-1;
                }
            }
            var column = this.model.getColumnCount(startLine);
            if (selectedRange.start.column < column) {
                column = selectedRange.start.column;
            }
            var offset = this.model.toTextOffset(startLine, column);
            this.modelTA.setCaretPosition(offset);
        },
        onScroll: function() {
            this.scroll('forward');
        },
        exec: function(e){
            this.stop(e);
            this.onScroll();
            this.end(e);
        }
    });
    var PageUpTrans = declare("", [ScrollByPageTrans], {
        constructor: function(){
            this.name = "pageup";
        },
        onScroll: function(){
            this.scroll('backward');
        }
    });
    var PageDownTrans = declare("", [ScrollByPageTrans], {
        constructor: function(){
            this.name = "pagedown";
        },
        onScroll: function(){
            this.scroll('forward');
        }
    });

var SidebarSelectTrans = declare("", [TransBase], {
    constructor: function(){
        this.name = "sidebarselect";
        this.allowedEvents = {
            "scroll": true,
            "mousemove": true,
            "mouseup": true
        };
        this.readOnly = this.modelTA.readOnly();
        this.enterReadOnly();
    },
    exec: function(e) {
        var selectingContext = {}, self = this;

        function _onCaptureInterval() {
            var offsetY,
                start,
                end;

            offsetY = Math.floor((self.textarea.scrollTop - selectingContext.scrolledY + selectingContext.y - selectingContext.startY) / self.view.getViewAttr('lineHei'));

            if (offsetY < 0) {
                start = Math.max(selectingContext.startLine + offsetY, 1);
                end = selectingContext.startLine;

                if ((start < (self.view.getViewAttr('startLine') + 1)) || (start > (self.view.getViewAttr('endLine') + 1))) {
                    self.view.scrollTo(start - 1, 0);
                }
            } else {
                start = selectingContext.startLine;
                end = Math.min(selectingContext.startLine + offsetY, self.model.getLineCount());

                if ((end > (self.view.getViewAttr('endLine') + 1)) || (end < (self.view.getViewAttr('startLine') + 1))) {
                    self.view.scrollTo(end - 1, 0);
                }
            }
            var selectedRange = new SAS.TextRange(new SAS.TextPosition(start - 1, 0), new SAS.TextPosition(end, 0));
            self.modelTA.setSelectionRange(self.model.toCoreRange(selectedRange));
        }

        function _onMouseMove(e) {
            selectingContext.y = e.clientY;
            self.stop(e);
            return false;
        }

        function _onMouseUp() {
            clearInterval(selectingContext.timerId);
            if (selectingContext.mouseMoveListener) {
                selectingContext.mouseMoveListener.remove();
                selectingContext.mouseMoveListener = null;
            }
            if (selectingContext.mouseUpListener) {
                selectingContext.mouseUpListener.remove();
                selectingContext.mouseUpListener = null;
            }
            self.end();
            self.exitReadOnly();
        }
        
        if (e.type === "mousedown" || e.type === "pointerdown" || e.type === "MSPointerDown") {
            var target = e.originalTarget || e.srcElement;          // Across browser.
            if ((target.className == "line-nbr") && (e.button == 0)) {                   // Required for valid innerHTML.
                selectingContext.startY = e.clientY;
                selectingContext.scrolledY = this.textarea.scrollTop;
                selectingContext.y = e.clientY;
                selectingContext.startLine = parseInt(target.innerHTML);             // Set line.
                if (has("MSPointer")){ //FIXID S1118940
                    if (has("pointer-events")){ // IE10
                        selectingContext.mouseMoveListener = on(document, "pointermove", _onMouseMove);
                        selectingContext.mouseUpListener = on(document, "pointerup", _onMouseUp);        
                    } else { // IE11
                        selectingContext.mouseMoveListener = on(document, "MSPointerMove", _onMouseMove);
                        selectingContext.mouseUpListener = on(document, "MSPointerUp", _onMouseUp);
                    }
                } else {
                    selectingContext.mouseMoveListener = on(document, "mousemove", _onMouseMove);
                    selectingContext.mouseUpListener = on(document, "mouseup", _onMouseUp);
                }
                selectingContext.timerId = setInterval(_onCaptureInterval, 20);
            } else {
                self.end();
                self.exitReadOnly();
            }
            e.preventDefault(); //S1124137
            //this.stop(e);
        }
    },
    enterReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(true);
        }
    },
    exitReadOnly: function() {
        if (!this.readOnly) {
            this.modelTA.readOnly(false);
        }
    }
});

var SidebarSelectTransIE = declare("", [SidebarSelectTrans], {
    constructor: function() {
        this.allowedEvents = {
            "scroll": true,
            "mousemove": true,
            "MSPointerMove": true,
            "pointermove": true,
            "mouseup": true,
            "MSPointerUp": true,
            "pointerup": true
        };
    },
    enterReadOnly: function(){},
    exitReadOnly: function(){}
});

var SelectByTouchTrans = declare("", [TransBase], {
    constructor: function() {
        this.name = "selectbytouch";
        this.allowedEvents = {
            'touchend':true,
            'scroll':true // NOTE: allow this
        };
    },
    exec: function(e) {
        if (e.type === 'touchend') {
            var self = this;
            window.setTimeout(function() {
                if (self.preContext.selection.start !== self.textarea.selectionStart ||
                    self.preContext.selection.end !== self.textarea.selectionEnd) {
                    self.modelTA.setSelectionRange(self.textarea.selectionStart, self.textarea.selectionEnd);
                }
                self.end(e);
            },0);
        }
    }
});

var TransFactory = declare("", null, {
    constructor: function() {
        this.cache = {};
    },
    getTrans: function(name) {
    }
    
});
/*
TransFactory.NAME_TO_TRANS = {
    "basic": TransFactory.createBasicTrans,
    "setcursor": TransFactory.createSetCursorTrans,
    "select": TransFactory.createSelectTrans,
    "paste": TransFactory.createPasteTrans,
    "backspace": TransFactory.createBackspaceTrans,
    "return": TransFactory.createReturnTrans,
    "del": TransFactory.createDelTrans,
    "tab": TransFactory.createTabTrans
};

TransFactory.createBasicTrans = function(config) {
}
TransFactory.createSetCursorTrans = function(config) {
}
TransFactory.createSelectTrans = function(config) {
}
TransFactory.createPasteTrans = function(config) {
}
TransFactory.createBackspaceTrans = function(config) {
}
TransFactory.createReturnTrans = function(config) {
}
TransFactory.createDelTrans = function(config) {
}
TransFactory.createTabTrans = function(config) {
}

*/

TransFactory.createTrans = function(name, config) {
    switch(name){
        case 'basic':
            if (SAS.Browser.ipad()) {
                return new BasicTransiPad(config);
            } else if (SAS.Browser.safari()) {
                return new BasicTransSafari(config);
            } else return new BasicTrans(config);
        case 'setcursor': return new SetCursorTrans(config);
        case 'setcursortostart': return new SetCursorToStartTrans(config);
        case 'setcursortoend': return new SetCursorToEndTrans(config);
        case 'setcursorbymouse':
            if (SAS.Browser.chrome()) {
                return new SetCursorByMouseTransChrome(config);
            } else if (SAS.Browser.ff()) {
                return new SetCursorByMouseTransFF(config);
            } else if (SAS.Browser.safari()) {
                return new SetCursorByMouseTransSafari(config);
            } else if (SAS.Browser.msie() === 10 || SAS.Browser.msie() === 11) {//FIXID S1118940
                return new SetCursorByMouseTransIE(config);
            } else {
                return new SetCursorByMouseTrans(config);
            }
        case 'scrollbymouse':
            if (SAS.Browser.chrome() || SAS.Browser.safari()) {
                return new ScrollByMouseTransChrome(config);
            } else {
                return null;
            }
        case 'selectbyword': return new SelectByWordTrans(config);
        case 'selectword': return new SelectWordTrans(config);
        case 'select': return new SelectTrans(config);
        case 'selectthetextbeforecaret': return new SelectTheTextBeforeCaretTrans(config);
        case 'selectthetextaftercaret': return new SelectTheTextAfterCaretTrans(config);
        case 'undo': return new UndoTrans(config);
        case 'ugly-undo': return new UglyUndoTrans(config);
        case 'disabledocundo': return new DisableDocUndo(config);
        case 'undoredo-msafari': return new UndoRedoTransMsafari(config);
        case 'redo': return new RedoTrans(config);
        case 'cutsim': return new CutSimTrans(config);
        case 'copysim': return new CopySimTrans(config);
        case 'pastesim': return new PasteSimTrans(config);
        case 'cut': return new CutTrans(config);
        case 'copy': return new CopyTrans(config);
        case 'uppercase': return new UppercaseTrans(config);
        case 'lowercase': return new LowercaseTrans(config);
        case 'cyclecase': return new CyclecaseTrans(config);
        case 'camelcase': return new CamelcaseTrans(config);
        case 'paste':
            if (SAS.Browser.mac() && SAS.Browser.safari()) {
                return new PasteTransMacSafari(config);
            } else if (SAS.Browser.msafari()) {
                return new PasteTransMobileSafari(config);
            } else if(SAS.Browser.webkit()) {
                return new PasteTransWebKit(config);
            } else if(SAS.Browser.msie()) {
                return new PasteTransIE(config);
            } else {
                return new PasteTransGeneral(config);
            }
        case 'backspace':
            if (SAS.Browser.msafari()) {
                return new BackspaceTransSafari(config);
            } else if (SAS.Browser.msie() === 9) {
                return new BackspaceTransIE9(config);
            } else
            return new BackspaceTrans(config);
        case 'return':
            if (SAS.Browser.msafari() || SAS.Browser.safari()) {
                return new ReturnTransSafari(config);
            } else return new ReturnTrans(config);
        case 'esc':
            if (SAS.Browser.msie()) {
                return new EscTransIE(config);
            }
            return null;
        case 'del':
            if (SAS.Browser.msafari()) {
                return new DelTransMSafari(config);
            } else if(SAS.Browser.msie() === 9) {
                return new DelTransIE9(config);
            } else {
                return new DelTrans(config);
            }
        case 'deleteword':return DelWordTrans(config);
        case 'tab':return new TabTrans(config);
        case 'deltab':return new DelTabTrans(config);
        case 'home':return new HomeTrans(config);
        case 'pageup': return new PageUpTrans(config);
        case 'pagedown': return new PageDownTrans(config);
        case 'scrollbydelta': return new ScrollByDeltaTrans(config);
        case 'ime':
        if (SAS.Browser.msie() === 11) {
            return new ImeTransIE11(config);
        } else if (SAS.Browser.msie() === 9) {
            return new ImeTransIE9(config);
        } else if (SAS.Browser.msie() === 10) {
            return new ImeTransIE10(config);
        } else if (SAS.Browser.ipad()) {
            return new ImeTransiPad(config);
        } else if (SAS.Browser.chrome()) {
            return new ImeTransChrome(config);
        } else if (SAS.Browser.safari()) {
            return new ImeTransSafari(config);
        } else {
            return new ImeTrans(config);
        }
    case 'comment':
        if (config.model.langMode() === SAS.Editor.LanguageMode.SasCode) {
            return SasCommentTrans(config);
        }
        return null;
        case 'sidebarselect':
            if (SAS.Browser.msie()) return new SidebarSelectTransIE(config);
            return new SidebarSelectTrans(config);
        case 'selectbytouch': return new SelectByTouchTrans(config);
    }
    
    return null;
};

TransFactory.createTransAndExec = function(name, config, arg) {
    config.onEnd = function(){};
    var trans = TransFactory.createTrans(name, config);
    if (trans) {
        return trans.exec(arg);
    }
};

return TransFactory;

});