define(["./CoreRange"], function(CoreRange) {
SAS.UndoManager = function () {
    var undoContainer = [],
        redoContainer = [];
    
    this.add = function (newText, newRange, newCursor, oldText, oldRange, oldCursor) {
        var unit = {
            newText:   newText,
            oldText:   oldText,
            oldRange:  oldRange,
            newRange:  newRange,
            newCursor: newCursor,
            oldCursor: oldCursor
        }
        // undoContainer.push(merge(unit));
        undoContainer.push(unit);
        if (redoContainer.length > 0) {
            redoContainer = [];
        }
    };

    this.undo = function () {
        var unit;
        if (undoContainer.length > 0) {
            unit = undoContainer.pop();
            redoContainer.push(unit);
        }
        return unit;
    };
    
    this.redo = function () {
        var unit;
        if (redoContainer.length > 0) {
            unit = redoContainer.pop();
            undoContainer.push(unit);
        }
        return unit;
    };
    
    this.canUndo = function() {
        return (undoContainer.length > 0);
    };
    this.canRedo = function() {
        return (redoContainer.length > 0);
    };

    this.clear = function () {
        undoContainer = [];
        redoContainer = [];
    };

    function merge (unit) {
        if (undoContainer.length === 0) { return unit; }
        var top = undoContainer.pop(),
            result = {};
        if (top.newRange.end === unit.oldRange.start) { //ADD / DEL
            result.newText  = top.newText + unit.newText;
            result.oldText  = top.oldText + unit.oldText;
            result.oldRange = new SAS.CoreRange(top.oldRange.start, top.oldRange.start + unit.oldText.length);
            result.newRange = new SAS.CoreRange(top.newRange.start, top.newRange.start + unit.newRange.length);
            result.newCursor= unit.newCursor;
            result.oldCursor= top.oldCursor;
        } else if (top.newRange.end === unit.oldRange.end) { // BACK
            if (top.newRange.start > unit.oldRange.start) { // delete more than added last time.
                result.newText  = top.newText.substr();
                result.oldText  = top.oldText;
                result.oldRange = top.oldRange;
                result.newRange = new SAS.CoreRange(top.newRange.start, top.newRange.start + unit.newRange.length);
                result.newCursor= unit.newCursor;
                result.oldCursor= top.oldCursor;
            } else {
                result.newText  = top.newText.substr(0, top.newText.length - unit.oldText.length).concat(unit.newText);
                result.oldText  = top.oldText;
                result.oldRange = top.oldRange;
                result.newRange = new SAS.CoreRange(top.newRange.start, top.newRange.start + unit.newRange.length);
                result.newCursor= unit.newCursor;
                result.oldCursor= top.oldCursor;
            }
        } else { // CANNOT MERGE.
            undoContainer.push(top);
            result = unit;
        }
        return result;
    }
};
return SAS.UndoManager;
});