// (c) 2014, SAS Institute Inc.

define(["../FormatterBase", "./SasLexerEx"], function(FormatterBase, SasLexerEx) {
    "use strict";

    var SasFormatter = FormatterBase.extend("sas.hc.ui.commons.codeEditor.mode.sas.Formatter", {
        BLOCK_END_EXP: /^\s*(RUN|RUN\s*CANCEL|%MEND.*)\s*;\s*$/gi,
        constructor: function() {
            FormatterBase.apply(this, arguments);
            this.allowedActions = {"regular":true, "return":true, "api":true};
            this.textChangedHandler = this.onTextChanged.bind(this);
            //TODO: Improve the chain access
            jQuery(this.control.core().baseModel()).on('textChanged', this.textChangedHandler);
        },
        onTextChanged: function(e) {
            if (!this.allowedActions[e.actionType]) {
                return;
            }
            if(this.settings.getIndentEnabled()) {
                if(e.text === ";" && e.removedText === "") {//Only handle to type ";" normally
                    this.tryToOutdent(e);
                } else if (e.text === "\n") {
                    this.tryToIndent(e);
                }
            } else {
                if (e.text === "\n") {
                    this.tryToDefaultIndent(e);
                }
            }

        },
        setCaretPosition: function(offset, delayed) {
            if (delayed) {
                jQuery.sap.delayedCall(0, this, function(){
                    this.control.selection(offset, offset);
                });
            } else {
                this.control.selection(offset, offset);
            }
        },
        getTabText: function() {
            return this.settings.getTabAsSpace() ? this.settings.getTabText() : '\t';
        },
        getIndentByLen: function(len) {
            var count = Math.floor(len / this.control.tabSize()),
                tabText = this.getTabText(), indent = "";

            while(count--) {
                indent += tabText;
            }
            count = this.control.getLengthWithTabAsSpaces(indent);
            if (count < len) {
                count = len - count;
                while(count--) {
                    indent += ' ';
                }
            }

            return indent;
        },
        getIndent: function(line) {
            var preLine = line - 1, preCol,
                lineText, indentLen,
                block, indent = ""; // default value

            if (line > 0) {
                preCol = this.model.getColumnCount(preLine);
                indent = this.model.getIndent({line: preLine, column: preCol});// same to the previous line
                if (!this.settings.getIndentEnabled()) {
                    return indent;
                }
                block = this.langServ.getFoldingBlock(preLine, preCol);// block statements
                if (block && block.startLine === preLine &&
                    !(block.endLine === preLine && block.endCol <= preCol && block.explicitEnd)) { //not in block
                    // block start, get indentation length
                    lineText = this.model.getLine(preLine);
                    indent = lineText.substring(0, block.startCol);
                    indent += this.getTabText();
                    indentLen = this.control.getLengthWithTabAsSpaces(indent);
                    // calculate indentation texts
                    indent = this.getIndentByLen(indentLen);
                }
            }

            return indent;
        },
        isBlockEnd: function (text) {
            this.BLOCK_END_EXP.lastIndex = 0;
            return this.BLOCK_END_EXP.test(text);
        },
        tryToIndent: function(editAction) {
            var caret = editAction.newRange.end,
                caretOffset = this.model.toTextOffset(caret.line, caret.column),
                indent = this.getIndent(caret.line), self = this;

            if (indent.length > 0) {
                //this.control.selection(caretOffset, caretOffset);
                this.setCaretPosition(caretOffset);
                window.setTimeout(function(){
                    self.control.insertText(indent);
                }, 0);
            }
        },
        tryToDefaultIndent: function(editAction) {
            this.tryToIndent(editAction);
        },
        tryToOutdent: function(editAction) {
            //we cannot use folding information because they are not ready
            var caret = editAction.newRange.end,
                endLineText = this.model.getLine(caret.line).substring(0, caret.column), matched;
            this.BLOCK_END_EXP.lastIndex = 0;
            matched = this.BLOCK_END_EXP.exec(endLineText);

            if (matched && matched.length > 0) {
                var block = this.langServ.getFoldingBlock(caret.line, 0),
                    startPrefix, endPrefix, startPrefixLen, endPrefixLen,
                    start, end, indent;

                switch (block.type) {
                    case SasLexerEx.SEC_TYPE.PROC:
                    case SasLexerEx.SEC_TYPE.DATA:
                        if ((/^%mend.*/gi).test(matched[1])) return;
                        break;
                    case SasLexerEx.SEC_TYPE.MACRO:
                        if (!(/^%mend.*/gi).test(matched[1])) return;
                }
                if (block && block.startLine !== block.endLine) {
                    startPrefix = this.model.getLine(block.startLine).substring(0, block.startCol);
                    matched = endLineText.match(/(^\s*|.*)/g);

                    if (matched && matched.length) {
                        endPrefix = matched[0];

                        startPrefixLen = this.control.getLengthWithTabAsSpaces(startPrefix);
                        endPrefixLen = this.control.getLengthWithTabAsSpaces(endPrefix);
                        if (startPrefixLen < endPrefixLen) { //decrease indentation
                            // replace end prefix
                            start = this.model.toTextOffset(caret.line, 0);
                            end = start + endPrefix.length;


                            indent = this.getIndentByLen(startPrefixLen);//alignment with the start
                            var self = this;
                            window.setTimeout(function(){
                                self.control.replaceText(indent, {start: start, end: end});// replace the indentation
                                    // set caret position
                                start = start + matched[1].length + indent.length;
                                self.setCaretPosition(start);
                            }, 0);
                        }
                    }
                }
            }
        }
    });

    return SasFormatter;
}, /* bExport= */ true);
