// (c) 2014, SAS Institute Inc.

define(["./SasKeywords", "../../TextPosition", "../../Utils"], function(SasKeywords, TextPosition, Utils) {
    "use strict";

    var keywordMap = null;
    var macroKwMap = null;
    var wordReg = null;

    var SasLexer = function(model){
        if (!wordReg) {
            var unicode = ace.require('ace/unicode');
            wordReg = new RegExp("["
                + unicode.packages.L
                + unicode.packages.Mn + unicode.packages.Mc
                + unicode.packages.Nd
                + unicode.packages.Pc + "\\$_]+$");
        }

        //'private' properties
        this.model = model;
        this.start = new TextPosition(0,0);
        this.curr = new TextPosition(0,0);
        this.quoting = -1;
        this.ignoreFormat = false;
        /*if (keywordMap === null) {
            keywordMap = Utils.arrayToMap(SasKeywords);//need to be optimized
        }*/
        this.context = {};
        /*
         this.keywordMap = Utils.arrayToMap([
         "ALL","ANY","CHARACTER","NUMERIC","ADDHIER","ADMINLIBS","AGE","ALL",
         "ALLOCATE","ALTER","AND","APPEND","ASCENDING","ATTRIB","BOUNDS","BY",
         "BYLINE","CALL","CATNAME","CCOPY","CDEF","CDELETE","CDFPLOT","CHART",
         "CHANGE","CLASS","CLASSES","CLASSLEV","CMAP","COLUMN","COMPUTE","CONTENTS",
         "CONTRAST","COORD","COORDINATES","COPY","DATASETS","DEFINE","DESCENDING",
         "DESCRIBE","DEVICE","DISCRIM","DROP","ELSE","ENDCOMP","EQ","ERROR",
         "EXCHANGE","EXCLUDE","FACTORS","FS","GE","GO","GOTO","GOUT","GROUPFORMAT",
         "GT","HIERARCHY","HIGH","HISTOGRAM","ID","IGOUT","IN","INDEX","INFORMAT",
         "INPUT","INSET","INTERVALS","KEEP","KEYLABEL","LE","LENGTH","LINEQS",
         "LINK","LOSTCARD","LOW","LSMEANS","LT","MATINGS","MEAN","MEANS","MERGE",
         "MODEL","MODIFY","MOVE","N","NE","NOBYLINE","NODUPKEY","NOT","NOTE",
         "NOTSORTED","OR","OTHER","OTHERWISE","OUTPUT","PAGEBY","PARMS","PARTIAL",
         "PCT","PCTN","PCTSUM","PICTURE","PPPLOT","PREDICT","PREFIX","PREVIEW",
         "PRISM","PROBPLOT","PROGLIBS","QQPLOT","RANDOM","RANGE","REMOVE",
         "REMOVEHIER","RENAME","REPEATED","REPLACE","REPLAY","REQUEST","RESET",
         "RESPONSE","RETAIN","S2","SAVE","SCATTER","SESSION","SET","SIM","SKIP",
         "SOURCE2","STARTSAS","STD","STRATA","SUM","SUMBY","SUPVAR","TABLE","TABLES",
         "TC","TCOPY","TDEF","TDELETE","TEMPLATE","TEST","TIME","TO","TPLAY",
         "TRANSFORM","TREPLAY","VALUE","VAR","VARIABLES","WEIGHT","WHEN",
         "WINDOW","WITH"
         ]); //from SasKeywordManager.java
         */
        if (macroKwMap === null) {
            macroKwMap = Utils.arrayToMap([
                "BQUOTE","BY","CMS","DISPLAY","DO","ELSE","END","EVAL","GLOBAL","GO",
                "IF","INC","INCLUDE","INDEX","INPUT","KEYDEF","LENGTH","LET","LOCAL",
                "MACRO","MEND","NRBQUOTE","NRQUOTE","NRSTR","PUT","QSCAN","QSUBSTR",
                "QSYSFUNC","QUOTE","QUPCASE","SCAN","STR","SUBSTR","SUPERQ","SYSCALL",
                "SYSEVALF","SYSEXEC","SYSFUNC","SYSGET","SYSPROD","SYSRPUT","THEN",
                "TO","TSO","UNQUOTE","UNTIL","UPCASE","WHILE","WINDOW",//from SasKeywordManager.java
                "DEQUOTE","SYMEXIST","SYMGLOBL","SYMLOCAL","LOWCASE","QLOWCASE",
                "TPLOT","SYSMACEXEC","SYSMACEXIST","SYSMEXECDEPTH","SYSMEXECNAME",
                "CMPRES","QCMPRES","QCMPRES","COMPSTOR","DATATYP","LEFT","QLEFT",
                "SYSRC","TRIM","QTRIM","VERIFY","DQLOAD","DQPUTLOC","DQUNLOAD",
                "KLOWCASE","QKLOWCAS","KTRIM","QKTRIM","KVERIFY", // from SASMacroFunctions.xml
                "ABORT","COPY",/*"DISPLAY","DO","UNTIL","WHILE","END","GLOBAL",*/
                "GOTO",/*"IF","THEN","ELSE","INPUT","LET","LOCAL","MACRO","MEND","PUT",*/
                "RETURN","SYMDEL",/*"SYSCALL","SYSEXEC","WHILE","UNTIL",*/
                "SYSLPUT",/*"SYSRPUT","WINDOW","INCLUDE",*/
                "LIST","RUN","SYSMSTORECLEAR","SYSMACDELETE"]);// from SASMacroStatements.xml
        }
    };
SasLexer.TOKEN_TYPES = {
    SEP: 'sep',
    KEYWORD: 'keyword',
    SKEYWORD: 'sec-keyword',
    COMMENT: 'comment',
    MKEYWORD: 'macro-keyword',
    MCOMMENT: 'macro-comment',
    MREF: 'macro-ref',
    MSKEYWORD: 'macro-sec-keyword',
    CARDSDATA: 'cards-data',
    STR: 'string',
    DATE: 'date',
    TIME: 'time',
    DT: 'dt',
    BM: 'bitmask',
    NL: 'namelit',
    HEX: 'hex',
    NUM: 'numeric',
    WORD: 'text',
    FORMAT: 'format',
    BLANK: 'blank'
};
SasLexer.notSyntaxToken = Utils.arrayToMap(['comment','macro-comment','blank']);
SasLexer.isComment = Utils.arrayToMap(['comment','macro-comment']);
SasLexer.isLiteral = Utils.arrayToMap(['string','date','time','dt','bitmask','namelit','hex','numeric','cards-data']);
SasLexer.isWord = Utils.arrayToMap(['keyword','sec-keyword', 'macro-sec-keyword', 'macro-ref', 'macro-keyword', 'text']);
SasLexer.isConst = Utils.arrayToMap(['string','date','time','dt','bitmask','namelit','hex','numeric','format']);
SasLexer.isCards = Utils.arrayToMap(['CARDS','LINES','DATALINES','DATALINES4','CARDS4','LINES4']);
SasLexer.isCards4 = Utils.arrayToMap(['DATALINES4','CARDS4','LINES4']);
SasLexer.isParmcards = Utils.arrayToMap(['PARMCARDS','PARMCARDS4']);
SasLexer.isParmcards4 = Utils.arrayToMap(['PARMCARDS4']);
SasLexer.isQuoting = Utils.arrayToMap(['%STR', '%NRSTR', '%QUOTE', '%NRQUOTE']);
SasLexer.isBinaryOpr = Utils.arrayToMap([
    '**','+','-','*','/',//arithmetic operators
    '&','|','!','\u00A6','AND','OR',
    //'\u00AC','\u00B0','~','NOT',//boolean operators
    '||',//contatenation operators
    '<>','><', //min and max operators
    '=','EQ',
    '^=','\u00AC=','~=','NE',
    '>','GT','<','LT',
    '>=','GE','<=','LE',
    '=:','>:','<:','^=:','\u00AC=:','~=:','>=:','<=:'//,//comparison
    //'IN' // disable this for HTMLCOMMONS-3847 (height=0.75in ctext=red)
]); // operators
SasLexer.isUnaryOpr   = Utils.arrayToMap(['\u00AC','\u00B0','~','NOT']);//boolean operators

SasLexer.longBiOprs = /^(\^=:|\u00AC=:|~=:|>=:|<=:|\*\*|<>|\^=|\u00AC=|~=|>=|<=|\|\||=:|>:|<:)/;
//'**','<>','><','^=','\u00AC=','~=','>=','<=','||','=:','>:','<:',  //len 2
//'^=:','\u00AC=:','~=:','>=:','<=:'   //len 3
    /**
     * SasLexer public methods
     */

    SasLexer.prototype = {
        getNext : function() {
            var token = this.getNext_();
            if (token) {
                if (SasLexer.isComment[token.type] === undefined) {
                    this.context.lastNoncommentToken = {
                        type:token.type,
                        start:{
                            line:token.start.line,
                            column:token.start.column
                        },
                        end:{
                            line:token.end.line,
                            column:token.end.column
                        }
                    };
                }
                if (SasLexer.isLiteral[token.type]) {
                    token.text = this.getText(token);
                } else {
                    token.text = this.getWord(token).toUpperCase();
                }
                if (SasLexer.isComment[token.type] === undefined) {
                    if (this.quoting === -1 && SasLexer.isQuoting[token.text]) {
                        this.quoting = 0;
                    } else if (this.quoting === 0 && token.text !== '(') {
                        this.quoting = -1;
                    } else if (this.quoting >= 0) {
                        if (token.text === '(') {
                            this.quoting++;
                        } else if (token.text === ')') {
                            this.quoting--;
                            if (this.quoting === 0) {
                                this.quoting = -1;
                            }
                        }
                    }
                    if (this.quoting === -1) {
                        if (!this.ignoreFormat && token.text === '%PUT') {
                            this.ignoreFormat = true;
                        } else if (this.ignoreFormat && token.text === ';') {
                            this.ignoreFormat = false;
                        }
                    }
                }
            }
            return token;
        },
        /**
         * @returns {object}
         * type: 'sep', 'comment', 'keyword', 'date', 'time', 'dt', 'normaltext',
         *       'numeric', 'string',
         *       'macro-sec', 'macro-keyword', 'macro-text', 'macro-def',
         * start: start position
         * end: end position
         *
         */
        getNext_ : function(){
            var j = 0,
                type,
                i,
                text,
                qm,
                word,
                ch,
                totalLines = this.model.getLineCount(),
                len;
            //loop1:
            while(this.curr.line < totalLines){
                this.start.line = this.curr.line;
                this.start.column = this.curr.column;
                i = this.curr.column;
                text = this.model.getLine(this.curr.line);
                len = text.length;
                switch(text[i]){
                    case '/':
                        i++;
                        if (i < len && text[i] === '*'){//this is comment
                            //find the end of this comment
                            i++;
                            do{
                                j = text.indexOf('*/',i);
                                if (j >= 0){
                                    this.curr.column = j+2;
                                    break;
                                }
                                this.curr.line++;
                                if (this.curr.line >= totalLines) {
                                    this.curr.line--;
                                    this.curr.column = text.length;
                                    break;
                                } else {
                                    this.curr.column = 0;
                                    i = 0;
                                    text = this.model.getLine(this.curr.line);
                                }
                            }while(1);
                            //this.curr.column++;
                            return {
                                type: SasLexer.TOKEN_TYPES.COMMENT,
                                start: this.start,
                                end: this.curr
                            };
                        } else {
                            this.curr.column = i;
                            return {type: SasLexer.TOKEN_TYPES.SEP,
                                start: this.start,
                                end: this.curr
                            };//
                        }
                        break;
                    case '%':
                        //macro,
                        i++;
                        if (i < len && /[A-Za-z_]/.test(text[i])){//the first char
                            i++;
                            while (i<len && /[\w_]/.test(text[i])){
                                i++;
                            }
                            this.curr.column = i;
                            //check macro section keyword
                            type = SasLexer.TOKEN_TYPES.WORD;
                            if (this.isMacroKeyword(text.substr(this.start.column + 1, i - this.start.column - 1))){
                                type = SasLexer.TOKEN_TYPES.MKEYWORD;
                            } else {
                                type = SasLexer.TOKEN_TYPES.MREF;
                            }
                            return {
                                type: type,//SasLexer.TOKEN_TYPES.MKEYWORD,
                                start: this.start,
                                end: this.curr
                            };
                        } else if(i< len && text[i] === '*'){ //macro comment
                            do {
                                i++;
                                if (i>=len){
                                    this.curr.line++;
                                    i=0;
                                    if (this.curr.line < totalLines) {
                                        text = this.model.getLine(this.curr.line);
                                        len = text.length;
                                    } else {
                                        this.curr.line--;
                                        this.curr.column = text.length;
                                        break;
                                    }
                                }
                                if (text[i] === ';') {
                                    this.curr.column = i + 1;
                                    break;
                                }
                            } while(1);

                            return {
                                type: SasLexer.TOKEN_TYPES.MCOMMENT,
                                start: this.start,
                                end: this.curr
                            };

                        }else if (this.quoting > 0 && i<len && (text[i] ==='\'' || text[i] ==='\"' || text[i] ==='('|| text[i] ===')')){
                            this.curr.column++;
                        }
                        this.curr.column++;
                        return {
                            type: SasLexer.TOKEN_TYPES.SEP,
                            start: this.start,
                            end: this.curr
                        };
                    case '\"':
                    case '\'': // string constants
                        // string constant, date, time, datetime,
                        qm = text[i];//qm = qutation mark
                        type = SasLexer.TOKEN_TYPES.STR;
                        //over multi lines
                        //escape??
                        i++;
                        while (this.curr.line < totalLines){
                            if (i<len){
                                if (qm === text[i]){
                                    //found end mark
                                    i++;
                                    if (qm === text[i]) {
                                        i++; // for escape, e.g. name='Tom''s'; name="Tom""s";
                                    } else {
                                        //check if this is date, time, datatime
                                        if (i<len){
                                            switch(text[i]) {
                                                case 'd': case 'D':
                                                type = SasLexer.TOKEN_TYPES.DATE;
                                                i++;
                                                if (i<len && (text[i] === 't' || text[i] === 'T')){
                                                    type = SasLexer.TOKEN_TYPES.DT;
                                                    i++;
                                                }
                                                break;
                                                case 't': case 'T': type = SasLexer.TOKEN_TYPES.TIME; i++; break;
                                                case 'b': case 'B': type = SasLexer.TOKEN_TYPES.BM; i++; break;//bit mask
                                                case 'n': case 'N': type = SasLexer.TOKEN_TYPES.NL; i++; break;//name literal
                                                case 'x': case 'X': type = SasLexer.TOKEN_TYPES.HEX; i++; break;//hexadecimal notation
                                            }
                                        }
                                        this.curr.column = i;
                                        return {
                                            type: type,
                                            start: this.start,
                                            end: this.curr
                                        };
                                    }
                                }else{
                                    i++;
                                }
                            }else{
                                this.curr.line++;
                                if (this.curr.line >= totalLines){
                                    this.curr.line--;
                                    this.curr.column = i;
                                    return {
                                        type: type,
                                        start: this.start,
                                        end: this.curr
                                    };
                                }
                                i=0;
                                text = this.model.getLine(this.curr.line);
                                len = text.length;
                            }
                        }
                        break;
                    case '\t':/*
                 this.curr.column++;
                 return {
                 type: SasLexer.TOKEN_TYPES.WORD,
                 start: this.start,
                 end: this.curr
                 };*/
                    case ' '://space
                    case '\u00a0':
                        this.curr.column++;
                        break;
                    case '.':
                        this.curr.column = this.readNum(i,10);
                        if (this.curr.column === i){
                            this.curr.column++;
                            type = SasLexer.TOKEN_TYPES.SEP;
                        } else {
                            type = SasLexer.TOKEN_TYPES.NUM;
                        }
                        return {
                            type: type,
                            start: this.start,
                            end:this.curr
                        };
                    //break;
                    case '$':
                        this.curr.column++;
                        type = SasLexer.TOKEN_TYPES.SEP;
                        i++;
                        if (i < len) {
                            ch = text[i];
                            if (/[a-zA-Z_0-9]/.test(ch)) {
                                i++;
                                while (i < len && /[\w_]/.test(text[i])) {
                                    i++;
                                }
                                if (i<len && text[i] === '.') {//format or informat
                        this.curr.column = this.readNum(i, 10);
                        if (this.curr.column === i){
                            this.curr.column++;
                        }
                                    type = SasLexer.TOKEN_TYPES.FORMAT;
                                }
                            }
                        }
                        return {
                            type: type,
                            start: this.start,
                            end: this.curr
                        };
                    //break;
                    case '*': //comment
                        // this kind of comment must follow ';' , may exists spaces between them
                        // may be multi lines
                        if (this.context.lastNoncommentToken !== null && this.getWord(this.context.lastNoncommentToken) !== ';'){
                            i++;
                            this.curr.column = i;
                            return {
                                type: SasLexer.TOKEN_TYPES.SEP,
                                start: this.start,
                                end: this.curr
                            };
                        }
                        //
                        //this is comment
                        while(/*(this.curr.line < totalLines) &&*/ (text[i] !== ';')) {
                            i++;
                            if(i >= len) {//continue to handle the next line
                                this.curr.line++;
                                if(this.curr.line >= totalLines) {//FIXID S0965730:Commenting and then uncommenting code in code editor causes code editor to be unusable.
                                    i--;
                                    this.curr.line--;
                                    break;
                                }
                                i = 0;
                                text = this.model.getLine(this.curr.line);
                                len = text.length;
                            }
                        }
                        this.curr.column = i + 1;
                        return {
                            type : SasLexer.TOKEN_TYPES.COMMENT,
                            start : this.start,
                            end : this.curr
                        };
                    //break;
                    case '\n':
                    case '\r':
                    case '\u2028':
                    case '\u2029':
                    case undefined:
                        this.curr.line++;
                        this.curr.column = 0;
                        i = 0;
                        //if (this.end()){
                        //    return null;
                        //}
                        //continue loop1;
                        break;
                    //return null;
                    default:{
                        type = SasLexer.TOKEN_TYPES.WORD;
                        word = '';
                        ch = text[i];
                        if (wordReg.test(ch) && !/\d/.test(ch)){//the first char
                            i++;
                            while (i<len && wordReg.test(text[i])){
                                i++;
                            }
                            //format or informat
                            if (this.quoting === -1 && !this.ignoreFormat && text[i] === '.') {
                                this.curr.column = this.readNum(i,10);
                                if (this.curr.column === i) {
                                    this.curr.column++;
                                }
                                i++;
                                if (this.curr.column > i || // has digital after .
                                    (i<len && /[a-zA-Z_]/.test(text[i]) === false) ||  //format like abc.#
                                    (this.curr.column === i && i === len)) { //format is the end of line
                                    return {
                                        type: SasLexer.TOKEN_TYPES.FORMAT,
                                        start: this.start,
                                        end: this.curr
                                    };
                                } else { // aaaa.bbb is token
                                    while(i<len && /\w/.test(text[i])) {
                                        i++;
                                    }
                                }
                            }
                            this.curr.column = i;
                            //TODO: check if it's keyword
                            word = text.substr(this.start.column, i - this.start.column);
                            //if (this.isSectionWord(word)){
                            //    type = SasLexer.TOKEN_TYPES.SKEYWORD;
                            //}else
                //if (this.isKeyword(word)) {
                //    type = SasLexer.TOKEN_TYPES.KEYWORD;
                //}

                            //end
                            return {
                                type: type,
                                start: this.start,
                                end: this.curr
                            }; //TODO: hex
                        }else if ((ch <= '9' && ch >= '0')
                        //|| ch === '+'
                        //|| ch === '-'
                        ){
                            // numeric
                            i++;
                            if (ch === '+' || ch === '-'){
                                if (text[i] <= '9' && text[i] >= '0'){
                                    //this.curr.column = i;
                                    i = this.readNum(i);
                                }else{
                                    //seperator
                                    this.curr.column = i;
                                    return {
                                        type: SasLexer.TOKEN_TYPES.SEP,
                                        start: this.start,
                                        end: this.curr
                                    };
                                }
                            }else{
                                i = this.readNum(this.curr.column);
                            }
                            //i++;//test
                            this.curr.column = i;
                            return {
                                type: SasLexer.TOKEN_TYPES.NUM,
                                start: this.start,
                                end: this.curr
                            };

                        } else {
                            var matches= SasLexer.longBiOprs.exec(text.substring(this.curr.column));
                            if (matches) {
                                this.curr.column += matches[0].length;
                            } else {
                                this.curr.column++;
                            }
                            return {
                                type: SasLexer.TOKEN_TYPES.SEP,
                                start: this.start,
                                end: this.curr
                            };
                        }
                    }//end default
                }//end switch
            }//end while(1)
        },
        end : function(){
            if (this.curr.line < this.model.getLineCount()) {
                return false;
            }
            return true;
        },
        reset : function(){
            this.curr.line = 0;
            this.curr.column = 0;
        },
        startFrom : function(line, col){
            this.curr.line = line;
            this.curr.column = col;
            this.context.lastNoncommentToken = null;
            this.quoting = -1;
            this.ignoreFormat = false;
        },
        readNum : function(col,radix){
            var text = this.model.getLine(this.curr.line), matched, end = col,
                reg = /^\b0?[0-9a-fA-F]+[xX]\b|^(?!0\d)\d+(\.\d{1,})(E([-+])?(\d+)?)?|^\d+\.|^(?!0\d)\d+(\.\d{1,})?(E([-+])?(\d+)?)?|^\d+\.?\d*|\.\d+/gi;
                // Pay attention to the duplication, it is necessary.
                if (radix === 10) {
                reg = /^\d+\.?\d*|^\.\d+/;
                }

            //reg.lastIndex = col;
            //matched = reg.exec(text);
            matched = reg.exec(text.substring(col));
            if (matched) {
                end = matched[0].length + col;
                    }
            return end;
        },/*
        isKeyword : function(word){
            var key = word.toUpperCase();
            if (keywordMap[key]) {
                return true;
            } else if (/^(TITLE|FOOTNOTE|AXIS|LEGEND|PATTERN|SYMBOL)\d{0,}$/.test(key)){
                var results = key.match(/(^(TITLE|FOOTNOTE|AXIS|LEGEND|PATTERN|SYMBOL)|\d{0,}$)/g),
                    nbr = parseInt(results[1],10);
                switch(results[0]){
                    case 'TITLE':
                    case 'FOOTNOTE':
                        return (nbr<11 && nbr>0);
                    case 'AXIS':
                    case 'LEGEND':
                        return (nbr<100 && nbr>0);
                    case 'PATTERN':
                    case 'SYMBOL':
                        return (nbr<256 && nbr>0);
                }
            }
            return false;
    },*/
        isMacroKeyword : function(word){
            return (macroKwMap[word.toUpperCase()] ? true : false);
        },
        getWord : function(token){ //token must be related to a word
            var str = '';
            if(token){
                str = this.model.getLine(token.end.line).substr(token.start.column, token.end.column - token.start.column);
            }
            return str;
        },
        getText : function(token) {
            return this.model.getText(token);
        }
    };
    return SasLexer;
});
